import { O as reportDiagnostic, c as compilerAssert, i as ignoreDiagnostics, S as SyntaxKind, a as createDiagnostic, L as ListenerFlow, d as createDiagnosticCollector, g as getSourceLocation, f as defineCodeFix, N as NoTarget, I as IdentifierKind, R as ResolutionResultFlags, P as typeReferenceToString } from './syntax-utils-DfiQDdeq.js';
import { a as isArrayModelType, c as isTemplateDeclarationOrInstance, i as isTemplateDeclaration, g as getTypeName, C as CustomKeyMap, r as isTemplateInstance, h as isType, s as isValue, q as isRecordModelType, o as isNeverType, m as isErrorType, t as isVoidType, b as isUnknownType, e as getEntityName, p as isNullType, d as getFullyQualifiedSymbolName, u as getParentTemplateNode } from './custom-key-map-CsDDh10S.js';
import { n as createRekeyableMap, i as isDefined, m as mutate, M as MultiKeyMap, b as isArray } from './misc-97rxrklX.js';
import { useStateMap, useStateSet, DuplicateTracker } from './utils.js';
import { visitChildren, getIdentifierContext, hasParseError, getFirstAncestor, exprIsBareIdentifier } from './ast.js';

/**
 * The prototype object for Typekit instances.
 *
 * @see {@link defineKit}
 *
 * @experimental
 * @internal
 */
const TypekitPrototype = {};
const TypekitNamespaceSymbol = Symbol.for("TypekitNamespace");
/**
 * Defines an extension to the Typekit interface.
 *
 * All Typekit instances will inherit the functionality defined by calls to this function.
 *
 * @experimental
 */
function defineKit(source) {
    for (const [name, fnOrNs] of Object.entries(source)) {
        let kits = fnOrNs;
        if (TypekitPrototype[name] !== undefined) {
            kits = { ...TypekitPrototype[name], ...fnOrNs };
        }
        // Tag top-level namespace objects with the symbol
        if (typeof kits === "object" && kits !== null) {
            Object.defineProperty(kits, TypekitNamespaceSymbol, {
                value: true,
                enumerable: false, // Keep the symbol non-enumerable
                configurable: false,
            });
        }
        TypekitPrototype[name] = kits;
    }
}

defineKit({
    array: {
        is(type) {
            return (type.kind === "Model" && isArrayModelType(this.program, type) && type.properties.size === 0);
        },
        getElementType(type) {
            if (!this.array.is(type)) {
                throw new Error("Type is not an array.");
            }
            return type.indexer.value;
        },
        create(elementType) {
            return this.model.create({
                name: "Array",
                properties: {},
                indexer: {
                    key: this.builtin.integer,
                    value: elementType,
                },
            });
        },
    },
});

defineKit({
    builtin: {
        get string() {
            return this.program.checker.getStdType("string");
        },
        get boolean() {
            return this.program.checker.getStdType("boolean");
        },
        get bytes() {
            return this.program.checker.getStdType("bytes");
        },
        get decimal() {
            return this.program.checker.getStdType("decimal");
        },
        get decimal128() {
            return this.program.checker.getStdType("decimal128");
        },
        get duration() {
            return this.program.checker.getStdType("duration");
        },
        get float() {
            return this.program.checker.getStdType("float");
        },
        get float32() {
            return this.program.checker.getStdType("float32");
        },
        get float64() {
            return this.program.checker.getStdType("float64");
        },
        get int8() {
            return this.program.checker.getStdType("int8");
        },
        get int16() {
            return this.program.checker.getStdType("int16");
        },
        get int32() {
            return this.program.checker.getStdType("int32");
        },
        get int64() {
            return this.program.checker.getStdType("int64");
        },
        get integer() {
            return this.program.checker.getStdType("integer");
        },
        get offsetDateTime() {
            return this.program.checker.getStdType("offsetDateTime");
        },
        get plainDate() {
            return this.program.checker.getStdType("plainDate");
        },
        get plainTime() {
            return this.program.checker.getStdType("plainTime");
        },
        get safeInt() {
            return this.program.checker.getStdType("safeint");
        },
        get uint8() {
            return this.program.checker.getStdType("uint8");
        },
        get uint16() {
            return this.program.checker.getStdType("uint16");
        },
        get uint32() {
            return this.program.checker.getStdType("uint32");
        },
        get uint64() {
            return this.program.checker.getStdType("uint64");
        },
        get url() {
            return this.program.checker.getStdType("url");
        },
        get utcDateTime() {
            return this.program.checker.getStdType("utcDateTime");
        },
        get numeric() {
            return this.program.checker.getStdType("numeric");
        },
    },
});

/**
 * Creates a shallow copy of a rekeyable map.
 *
 * @experimental
 */
function copyMap(map) {
    return createRekeyableMap(map.entries());
}
/**
 * Utility function for converting a list of decorators and arguments into structure decorator applications that are
 * compatible with the TypeSpec type graph.
 *
 * Note: Not all JavaScript values can be converted faithfully to TypeSpec values. This function will attempt to convert
 * the value to a literal TypeSpec value if it is a primitive value, otherwise it will be left as a JavaScript value.
 * The `jsValue` property of the argument will always contain the original JavaScript value passed to the decorator.
 *
 * @param typekit - The Typekit instance to use for creating types and values.
 * @param decorators - The list of decorators and arguments to apply.
 *
 * @see {@link DecoratorArgs}
 *
 * @experimental
 */
function decoratorApplication(typekit, decorators) {
    return (decorators?.map((arg) => {
        const [decorator, args] = Array.isArray(arg) ? [arg[0], arg.slice(1)] : [arg, []];
        return {
            decorator,
            args: args.map((rawValue) => ({
                value: typeof rawValue === "object" && rawValue !== null
                    ? rawValue
                    : typekit.literal.create(rawValue),
                jsValue: rawValue,
            })),
        };
    }) ?? []);
}

defineKit({
    enumMember: {
        create(desc) {
            const member = this.program.checker.createType({
                kind: "EnumMember",
                name: desc.name,
                value: desc.value,
                decorators: decoratorApplication(this, desc.decorators),
                node: undefined,
                enum: desc.enum, // initialized in enum.build if not provided here
            });
            this.program.checker.finishType(member);
            return member;
        },
        is(type) {
            return type.kind === "EnumMember";
        },
    },
});

defineKit({
    enum: {
        create(desc) {
            const en = this.program.checker.createType({
                kind: "Enum",
                name: desc.name,
                decorators: decoratorApplication(this, desc.decorators),
                members: createRekeyableMap(),
                node: undefined,
            });
            if (Array.isArray(desc.members)) {
                for (const member of desc.members) {
                    member.enum = en;
                    en.members.set(member.name, member);
                }
            }
            else {
                for (const [name, member] of Object.entries(desc.members ?? {})) {
                    en.members.set(name, this.enumMember.create({ name, value: member, enum: en }));
                }
            }
            this.program.checker.finishType(en);
            return en;
        },
        is(type) {
            return type.kind === "Enum";
        },
        createFromUnion(type) {
            if (!type.name) {
                throw new Error("Cannot create an enum from an anonymous union.");
            }
            const enumMembers = [];
            for (const variant of type.variants.values()) {
                if ((variant.name && typeof variant.name === "symbol") ||
                    (!this.literal.isString(variant.type) && !this.literal.isNumeric(variant.type))) {
                    continue;
                }
                enumMembers.push(this.enumMember.create({ name: variant.name, value: variant.type.value }));
            }
            return this.enum.create({ name: type.name, members: enumMembers });
        },
    },
});

class InvalidNumericError extends Error {
    code = "InvalidNumeric";
}
/** @internal */
const InternalDataSym = Symbol.for("NumericInternalData");
/**
 * Check if the given arg is a Numeric
 */
function isNumeric(arg) {
    return typeof arg === "object" && arg !== null && InternalDataSym in arg;
}
/**
 * Represent any possible numeric value
 */
function Numeric(stringValue) {
    if (new.target) {
        throw new Error("Numeric is not a constructor");
    }
    const data = parse(stringValue);
    const isInteger = data.d === 0;
    const obj = {
        [InternalDataSym]: data,
        isInteger,
    };
    // We are explicitly not using a class here due to version mismatch between the compiler and the runtime that could happen and break instanceof checks.
    const numeric = setTypedProptotype(obj, NumericPrototype);
    return Object.freeze(numeric);
}
function setTypedProptotype(obj, prototype) {
    Object.setPrototypeOf(obj, prototype);
    return obj;
}
function parse(original) {
    let stringValue = original;
    let start = 0;
    let sign = 1;
    let n;
    let exp;
    let decimal = undefined;
    if (stringValue[0] === "-") {
        start = 1;
        sign = -1;
    }
    const second = stringValue[start + 1]?.toLowerCase();
    if (stringValue[start] === "0" && (second === "b" || second === "x" || second === "o")) {
        try {
            n = BigInt(stringValue.slice(start));
            exp = n.toString().length;
            decimal = 0;
        }
        catch {
            throw new InvalidNumericError(`Invalid numeric value: ${original}`);
        }
    }
    else {
        // Skip leading 0.
        while (stringValue[start] === "0") {
            start++;
        }
        const decimalPointIndex = stringValue.indexOf(".");
        const adjustedPointIndex = decimalPointIndex - start;
        // Decimal point?
        if (decimalPointIndex !== -1) {
            exp = adjustedPointIndex;
            stringValue = stringValue.replace(".", "");
        }
        let i;
        if ((i = stringValue.search(/e/i)) > 0) {
            // Determine exponent.
            if (exp === undefined) {
                exp = i - start;
            }
            exp += Number(stringValue.slice(i + 1));
            stringValue = stringValue.slice(start, i);
            decimal = Math.max(stringValue.length - exp, 0);
        }
        else if (exp === undefined) {
            // Integer.
            exp = stringValue.length - start;
            stringValue = stringValue.slice(start);
        }
        else {
            stringValue = stringValue.slice(start);
        }
        let end = stringValue.length;
        while (stringValue[end - 1] === "0" && end > adjustedPointIndex) {
            end--;
        }
        // Only if there is 0 before the decimal point, keeps checking how many 0 there is after it and update the exponent accordingly.
        if (start === adjustedPointIndex + 1) {
            let cur = adjustedPointIndex;
            while (stringValue[cur] === "0" && cur < end) {
                cur++;
                exp--;
            }
        }
        try {
            stringValue = stringValue.slice(0, end);
            stringValue = stringValue + "0".repeat(Math.max(exp - stringValue.length, 0)); // add remaining zeros for cases like 3e30
            n = BigInt(stringValue);
            if (n === 0n) {
                decimal = 0;
            }
            else if (decimal === undefined) {
                decimal = Math.max(stringValue.length - Math.max(exp, 0), 0);
            }
        }
        catch {
            throw new InvalidNumericError(`Invalid numeric value: ${original}`);
        }
    }
    return { n, e: exp, s: sign, d: decimal };
}
function stringify(value) {
    if (value.n === 0n)
        return "0";
    const n = value.n.toString();
    const sign = value.s === -1 ? "-" : "";
    const int = value.e <= 0 ? "0" : n.slice(0, value.e);
    const decimal = value.e < n.length ? "." + n.slice(-value.d).padStart(value.d, "0") : "";
    return sign + int + decimal;
}
const equals = (a, b) => a.n === b.n && a.e === b.e;
const compare = (a, b) => {
    if (a.s < b.s) {
        return -1;
    }
    else if (a.s > b.s) {
        return 1;
    }
    const neg = a.s;
    if (a.e < b.e) {
        return (-1 * neg);
    }
    else if (a.e > b.e) {
        return (1 * neg);
    }
    let aN = a.n;
    let bN = b.n;
    if (a.d < b.d) {
        aN *= 10n ** BigInt(b.d - a.d);
    }
    else {
        bN *= 10n ** BigInt(a.d - b.d);
    }
    if (aN < bN)
        return (-1 * neg);
    if (aN > bN)
        return (1 * neg);
    return 0;
};
const NumericPrototype = {
    toString: function () {
        return stringify(this[InternalDataSym]);
    },
    asNumber: function () {
        const num = Number(stringify(this[InternalDataSym]));
        return equals(this[InternalDataSym], Numeric(num.toString())[InternalDataSym]) ? num : null;
    },
    asBigInt: function () {
        if (!this.isInteger) {
            return null;
        }
        const { s, n } = this[InternalDataSym];
        return BigInt(s) * n;
    },
    equals: function (other) {
        return equals(this[InternalDataSym], other[InternalDataSym]);
    },
    lt: function (other) {
        return compare(this[InternalDataSym], other[InternalDataSym]) === -1;
    },
    lte: function (other) {
        return compare(this[InternalDataSym], other[InternalDataSym]) <= 0;
    },
    gt: function (other) {
        return compare(this[InternalDataSym], other[InternalDataSym]) === 1;
    },
    gte: function (other) {
        return compare(this[InternalDataSym], other[InternalDataSym]) >= 0;
    },
};
NumericPrototype.toString = function () {
    return stringify(this[InternalDataSym]);
};

defineKit({
    literal: {
        create(value) {
            if (typeof value === "string") {
                return this.literal.createString(value);
            }
            else if (typeof value === "number") {
                return this.literal.createNumeric(value);
            }
            else {
                return this.literal.createBoolean(value);
            }
        },
        createString(value) {
            return this.program.checker.createType({
                kind: "String",
                value,
            });
        },
        createNumeric(value) {
            const valueAsString = String(value);
            return this.program.checker.createType({
                kind: "Number",
                value,
                valueAsString,
                numericValue: Numeric(valueAsString),
            });
        },
        createBoolean(value) {
            return this.program.checker.createType({
                kind: "Boolean",
                value,
            });
        },
        isBoolean(type) {
            return type.kind === "Boolean";
        },
        isString(type) {
            return type.kind === "String";
        },
        isNumeric(type) {
            return type.kind === "Number";
        },
        is(type) {
            return (this.literal.isBoolean(type) || this.literal.isNumeric(type) || this.literal.isString(type));
        },
    },
});

/**
 * Filters the properties of a model by removing them from the model instance if
 * a given `filter` predicate is not satisfied.
 *
 * @param model - the model to filter properties on
 * @param filter - the predicate to filter properties with
 */
function filterModelPropertiesInPlace(model, filter) {
    for (const [key, prop] of model.properties) {
        if (!filter(prop)) {
            model.properties.delete(key);
        }
    }
}
/**
 * Creates a unique symbol for storing state on objects
 * @param name The name/description of the state
 */
function createStateSymbol(name) {
    return Symbol.for(`TypeSpec.${name}`);
}

// Copyright (c) Microsoft Corporation
// Licensed under the MIT license.
// TypeSpec Visibility System
// --------------------------
// This module defines the core visibility system of the TypeSpec language. The
// visibility system is used to decide when properties of a _conceptual resource_
// are present. The system is based on the concept of _visibility classes_,
// represented by TypeSpec enums. Each visibility class has a set of _visibility
// modifiers_ that can be applied to a model property, each modifier represented
// by a member of the visibility class enum.
//
// Each visibility class has a _default modifier set_ that is used when no
// modifiers are specified for a property, and each property has an _active
// modifier set_ that is used when analyzing the visibility of the property.
//
// Visibility can be _sealed_ for a program, property, or visibility class
// within a property. Once visibility is sealed, it cannot be unsealed, and any
// attempts to modify a sealed visibility will fail.
/**
 * The global visibility store.
 *
 * This store is used to track the visibility modifiers
 */
const [getVisibilityStore, setVisibilityStore] = useStateMap(createStateSymbol("visibilityStore"));
/**
 * Provides access to the visibility store.
 * @internal
 */
function getRawVisibilityStore(...params) {
    return getVisibilityStore(...params);
}
/**
 * Returns the visibility modifiers for a given `property` within a `program`.
 */
function getOrInitializeVisibilityModifiers(program, property) {
    let visibilityModifiers = getVisibilityStore(program, property);
    if (!visibilityModifiers) {
        visibilityModifiers = new Map();
        setVisibilityStore(program, property, visibilityModifiers);
    }
    return visibilityModifiers;
}
/**
 * Returns the active visibility modifier set for a given `property` and `visibilityClass`.
 *
 * If no visibility modifiers have been set for the given `property` and `visibilityClass`, the function will use the
 * provided `defaultSet` to initialize the visibility modifiers.
 *
 * @param program - the program in which the property occurs
 * @param property - the property to get visibility modifiers for
 * @param visibilityClass - the visibility class to get visibility modifiers for
 * @param defaultSet - the default set to use if no set has been initialized
 * @returns the active visibility modifier set for the given property and visibility class
 */
function getOrInitializeActiveModifierSetForClass(program, property, visibilityClass, defaultSet) {
    const visibilityModifiers = getOrInitializeVisibilityModifiers(program, property);
    let visibilityModifierSet = visibilityModifiers.get(visibilityClass);
    if (!visibilityModifierSet) {
        visibilityModifierSet = defaultSet;
        visibilityModifiers.set(visibilityClass, visibilityModifierSet);
    }
    return visibilityModifierSet;
}
/**
 * If a Program is in this set, visibility is sealed for all properties in that Program.
 */
const VISIBILITY_PROGRAM_SEALS = new WeakSet();
const [isVisibilitySealedForProperty, sealVisibilityForProperty] = useStateSet(createStateSymbol("propertyVisibilitySealed"));
const [getSealedVisibilityClasses, setSealedVisibilityClasses] = useStateMap(createStateSymbol("sealedVisibilityClasses"));
/**
 * Seals visibility modifiers for a property in a given visibility class.
 *
 * @param program - the program in which the property occurs
 * @param property - the property to seal visibility modifiers for
 * @param visibilityClass - the visibility class to seal visibility modifiers for
 */
function sealVisibilityModifiersForClass(program, property, visibilityClass) {
    let sealedClasses = getSealedVisibilityClasses(program, property);
    if (!sealedClasses) {
        sealedClasses = new Set();
        setSealedVisibilityClasses(program, property, sealedClasses);
    }
    sealedClasses.add(visibilityClass);
}
/**
 * Stores the default modifier set for a given visibility class.
 */
const [getDefaultModifiers, setDefaultModifiers] = useStateMap(createStateSymbol("defaultVisibilityModifiers"));
/**
 * Gets the default modifier set for a visibility class. If no default modifier set has been set, this function will
 * initialize the default modifier set to ALL the visibility class's members.
 *
 * @param program - the program in which the visibility class occurs
 * @param visibilityClass - the visibility class to get the default modifier set for
 * @returns the default modifier set for the visibility class
 */
function getDefaultModifierSetForClass(program, visibilityClass) {
    const cached = getDefaultModifiers(program, visibilityClass);
    if (cached)
        return new Set(cached);
    const defaultModifierSet = new Set(visibilityClass.members.values());
    setDefaultModifiers(program, visibilityClass, defaultModifierSet);
    // Explicitly clone the set again to prevent accidental modification of the default set.
    return new Set(defaultModifierSet);
}
/**
 * Set the default visibility modifier set for a visibility class.
 *
 * This function may only be called ONCE per visibility class and must be called
 * before the default modifier set is used by any operation.
 */
function setDefaultModifierSetForVisibilityClass(program, visibilityClass, defaultSet) {
    compilerAssert(!getDefaultModifiers(program, visibilityClass), "The default modifier set for a visibility class may only be set once.");
    setDefaultModifiers(program, visibilityClass, defaultSet);
}
/**
 * Convert a sequence of visibility modifiers into a map of visibility classes to their respective modifiers in the
 * sequence.
 *
 * @param modifiers - the visibility modifiers to group
 * @returns a map of visibility classes to their respective modifiers in the input list
 */
function groupModifiersByVisibilityClass(modifiers) {
    const enumMap = new Map();
    // Prepare new modifier sets for each visibility class
    for (const modifier of modifiers) {
        const visibilityClass = modifier.enum;
        let modifierSet = enumMap.get(visibilityClass);
        if (!modifierSet) {
            modifierSet = new Set();
            enumMap.set(visibilityClass, modifierSet);
        }
        modifierSet.add(modifier);
    }
    return enumMap;
}
// #region Visibility Management API
/**
 * Check if a property has had its visibility modifiers sealed.
 *
 * If the property has been sealed globally, this function will return true. If the property has been sealed for the
 * given visibility class, this function will return true.
 *
 * Otherwise, this function returns false.
 *
 * @param property - the property to check
 * @param visibilityClass - the optional visibility class to check
 * @returns true if the property is sealed for the given visibility class, false otherwise
 */
function isSealed(program, property, visibilityClass) {
    if (VISIBILITY_PROGRAM_SEALS.has(program))
        return true;
    const classSealed = visibilityClass
        ? getSealedVisibilityClasses(program, property)?.has(visibilityClass)
        : false;
    return classSealed || isVisibilitySealedForProperty(program, property);
}
/**
 * Seals a property's visibility modifiers.
 *
 * If the `visibilityClass` is provided, the property's visibility modifiers will be sealed for that visibility class
 * only. Otherwise, the property's visibility modifiers will be sealed for all visibility classes (globally).
 *
 * @param property - the property to seal
 * @param visibilityClass - the optional visibility class to seal the property for
 */
function sealVisibilityModifiers(program, property, visibilityClass) {
    if (visibilityClass) {
        sealVisibilityModifiersForClass(program, property, visibilityClass);
    }
    else {
        sealVisibilityForProperty(program, property);
    }
}
/**
 * Seals a program's visibility modifiers.
 *
 * This affects all properties in the program and prevents any further modifications to visibility modifiers within the
 * program.
 *
 * Once the modifiers for a program are sealed, they cannot be unsealed.
 *
 * @param program - the program to seal
 */
function sealVisibilityModifiersForProgram(program) {
    VISIBILITY_PROGRAM_SEALS.add(program);
}
/**
 * Add visibility modifiers to a property.
 *
 * This function will add all the `modifiers` to the active set of visibility modifiers for the given `property`.
 *
 * If no set of active modifiers exists for the given `property`, an empty set will be created for the property.
 *
 * If the visibility modifiers for `property` in the given modifier's visibility class have been sealed, this function
 * will issue a diagnostic and ignore that modifier, but it will still add the rest of the modifiers whose classes have
 * not been sealed.
 *
 * @param program - the program in which the ModelProperty occurs
 * @param property - the property to add visibility modifiers to
 * @param modifiers - the visibility modifiers to add
 * @param context - the optional decorator context to use for displaying diagnostics
 */
function addVisibilityModifiers(program, property, modifiers, context) {
    const modifiersByClass = groupModifiersByVisibilityClass(modifiers);
    for (const [visibilityClass, newModifiers] of modifiersByClass.entries()) {
        const target = context?.decoratorTarget ?? property;
        if (isSealed(program, property, visibilityClass)) {
            reportDiagnostic(program, {
                code: "visibility-sealed",
                format: {
                    propName: property.name,
                },
                target,
            });
            continue;
        }
        const modifierSet = getOrInitializeActiveModifierSetForClass(program, property, visibilityClass, 
        /* defaultSet: */ new Set());
        for (const modifier of newModifiers) {
            modifierSet.add(modifier);
        }
    }
}
/**
 * Remove visibility modifiers from a property.
 *
 * This function will remove all the `modifiers` from the active set of visibility modifiers for the given `property`.
 *
 * If no set of active modifiers exists for the given `property`, the default set for the modifier's visibility class
 * will be used.
 *
 * If the visibility modifiers for `property` in the given modifier's visibility class have been sealed, this function
 * will issue a diagnostic and ignore that modifier, but it will still remove the rest of the modifiers whose classes
 * have not been sealed.
 *
 * @param program - the program in which the ModelProperty occurs
 * @param property - the property to remove visibility modifiers from
 * @param modifiers - the visibility modifiers to remove
 * @param context - the optional decorator context to use for displaying diagnostics
 */
function removeVisibilityModifiers(program, property, modifiers, context) {
    const modifiersByClass = groupModifiersByVisibilityClass(modifiers);
    for (const [visibilityClass, newModifiers] of modifiersByClass.entries()) {
        const target = context?.decoratorTarget ?? property;
        if (isSealed(program, property, visibilityClass)) {
            reportDiagnostic(program, {
                code: "visibility-sealed",
                format: {
                    propName: property.name,
                },
                target,
            });
            continue;
        }
        const modifierSet = getOrInitializeActiveModifierSetForClass(program, property, visibilityClass, 
        /* defaultSet: */ getDefaultModifierSetForClass(program, visibilityClass));
        for (const modifier of newModifiers) {
            modifierSet.delete(modifier);
        }
    }
}
/**
 * Clears the visibility modifiers for a property in a given visibility class.
 *
 * If the visibility modifiers for the given class are sealed, this function will issue a diagnostic and leave the
 * visibility modifiers unchanged.
 *
 * @param program - the program in which the ModelProperty occurs
 * @param property - the property to clear visibility modifiers for
 * @param visibilityClass - the visibility class to clear visibility modifiers for
 * @param context - the optional decorator context to use for displaying diagnostics
 */
function clearVisibilityModifiersForClass(program, property, visibilityClass, context) {
    const target = context?.decoratorTarget ?? property;
    if (isSealed(program, property, visibilityClass)) {
        reportDiagnostic(program, {
            code: "visibility-sealed",
            format: {
                propName: property.name,
            },
            target,
        });
        return;
    }
    const modifierSet = getOrInitializeActiveModifierSetForClass(program, property, visibilityClass, 
    /* defaultSet: */ new Set());
    modifierSet.clear();
}
/**
 * Resets the visibility modifiers for a property in a given visibility class.
 *
 * This does not clear the modifiers. It resets them to the _uninitialized_ state.
 *
 * This is useful when cloning properties and you want to reset the visibility modifiers on the clone.
 *
 * If the visibility modifiers for this property and given visibility class are sealed, this function will issue a
 * diagnostic and leave the visibility modifiers unchanged.
 *
 * @param program - the program in which the property occurs
 * @param property - the property to reset visibility modifiers for
 * @param visibilityClass - the visibility class to reset visibility modifiers for
 * @param context - the optional decorator context to use for displaying diagnostics
 */
function resetVisibilityModifiersForClass(program, property, visibilityClass, context) {
    const target = context?.decoratorTarget ?? property;
    if (isSealed(program, property, visibilityClass)) {
        reportDiagnostic(program, {
            code: "visibility-sealed",
            format: {
                propName: property.name,
            },
            target,
        });
        return;
    }
    getOrInitializeVisibilityModifiers(program, property).delete(visibilityClass);
}
// #endregion
// #region Visibility Analysis API
/**
 * Returns the active visibility modifiers for a property in a given visibility class.
 *
 * This function is infallible. If the visibility modifiers for the given class have not been set explicitly, it will
 * return the default visibility modifiers for the class.
 *
 * @param program - the program in which the property occurs
 * @param property - the property to get visibility modifiers for
 * @param visibilityClass - the visibility class to get visibility modifiers for
 * @returns the set of active modifiers (enum members) for the property and visibility class
 */
function getVisibilityForClass(program, property, visibilityClass) {
    const store = getVisibilityStore(program, property);
    return store?.get(visibilityClass) ?? getDefaultModifierSetForClass(program, visibilityClass);
}
/**
 * Determines if a property has a specified visibility modifier.
 *
 * If no visibility modifiers have been set for the visibility class of the modifier, the visibility class's default
 * modifier set is used.
 *
 * @param program - the program in which the property occurs
 * @param property - the property to check
 * @param modifier - the visibility modifier to check for
 * @returns true if the property has the visibility modifier, false otherwise
 */
function hasVisibility(program, property, modifier) {
    const visibilityClass = modifier.enum;
    return getVisibilityForClass(program, property, visibilityClass).has(modifier) ?? false;
}
const VisibilityFilter = {
    /**
     * Convert a TypeSpec `GeneratedVisibilityFilter` value to a `VisibilityFilter`.
     *
     * @param filter - the decorator argument filter to convert
     * @returns a `VisibilityFilter` object that can be consumed by the visibility APIs
     */
    fromDecoratorArgument(filter) {
        return {
            all: filter.all && new Set(filter.all.map((v) => v.value)),
            any: filter.any && new Set(filter.any.map((v) => v.value)),
            none: filter.none && new Set(filter.none.map((v) => v.value)),
        };
    },
    /**
     * Extracts the unique visibility classes referred to by the modifiers in a
     * visibility filter.
     *
     * @param filter - the visibility filter to extract visibility classes from
     * @returns a set of visibility classes referred to by the filter
     */
    getVisibilityClasses(filter) {
        const classes = new Set();
        if (filter.all)
            filter.all.forEach((v) => classes.add(v.enum));
        if (filter.any)
            filter.any.forEach((v) => classes.add(v.enum));
        if (filter.none)
            filter.none.forEach((v) => classes.add(v.enum));
        return classes;
    },
};
/**
 * Determines if a property is visible according to the given visibility filter.
 *
 * @see VisibilityFilter
 *
 * @param program - the program in which the property occurs
 * @param property - the property to check
 * @param filter - the visibility filter to use
 * @returns true if the property is visible according to the filter, false otherwise
 */
function isVisible(program, property, filter) {
    // Validate that property has ALL of the required visibilities of filter.all
    if (filter.all) {
        for (const modifier of filter.all) {
            if (!hasVisibility(program, property, modifier))
                return false;
        }
    }
    // Validate that property has NONE of the excluded visibilities of filter.none
    if (filter.none) {
        for (const modifier of filter.none) {
            if (hasVisibility(program, property, modifier))
                return false;
        }
    }
    if (filter.any) {
        for (const modifier of filter.any) {
            if (hasVisibility(program, property, modifier))
                return true;
        }
        return false;
    }
    return true;
}
// #endregion

/**
 * Validate the decorator target is matching the expected value.
 * @param context
 * @param target
 * @param decoratorName
 * @param expectedType
 * @returns
 */
function validateDecoratorTarget(context, target, decoratorName, expectedType) {
    const isCorrectType = isTypeSpecValueTypeOf(target, expectedType);
    if (!isCorrectType) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            format: {
                decorator: decoratorName,
                to: target.kind,
            },
            target: context.decoratorTarget,
        });
        return false;
    }
    return true;
}
function isIntrinsicType(program, type, kind) {
    return ignoreDiagnostics(program.checker.isTypeAssignableTo(type, program.checker.getStdType(kind), type));
}
/**
 * Check if the given target is of any of the TypeSpec types.
 * @param target Target to validate.
 * @param expectedType One or multiple allowed TypeSpec types.
 * @returns boolean if the target is of one of the allowed types.
 */
function isTypeSpecValueTypeOf(target, expectedType) {
    const kind = getTypeKind(target);
    if (kind === undefined) {
        return false;
    }
    return typeof expectedType === "string"
        ? expectedType === "Any" || kind === expectedType
        : expectedType.includes("Any") || expectedType.includes(kind);
}
function getTypeKind(target) {
    switch (typeof target) {
        case "object":
            return target.kind;
        case "string":
            return "String";
        case "number":
            return "Number";
        case "boolean":
            return "Boolean";
        default:
            return undefined;
    }
}
/**
 * Convert a TypeSpec type to a serializable Json object.
 * Emits diagnostics if the given type is invalid
 * @param typespecType The type to convert to Json data
 * @param target The diagnostic target in case of errors.
 */
function typespecTypeToJson(typespecType, target) {
    if (typeof typespecType !== "object") {
        return [typespecType, []];
    }
    return typespecTypeToJsonInternal(typespecType, target, []);
}
function typespecTypeToJsonInternal(typespecType, target, path) {
    switch (typespecType.kind) {
        case "String":
        case "Boolean":
        case "Number":
            return [typespecType.value, []];
        case "EnumMember":
            return [typespecType.value ?? typespecType.name, []];
        case "Tuple": {
            const result = [];
            for (const [index, type] of typespecType.values.entries()) {
                const [item, diagnostics] = typespecTypeToJsonInternal(type, target, [
                    ...path,
                    index.toString(),
                ]);
                if (diagnostics.length > 0) {
                    return [undefined, diagnostics];
                }
                result.push(item);
            }
            return [result, []];
        }
        case "Model": {
            const result = {};
            for (const [name, type] of typespecType.properties.entries()) {
                const [item, diagnostics] = typespecTypeToJsonInternal(type.type, target, [
                    ...path,
                    name.toString(),
                ]);
                if (diagnostics.length > 0) {
                    return [undefined, diagnostics];
                }
                result[name] = item;
            }
            return [result, []];
        }
        case "StringTemplate":
            if (typespecType.stringValue) {
                return [typespecType.stringValue, []];
            }
        // By design
        // eslint-disable-next-line no-fallthrough
        default:
            const diagnostic = path.length === 0
                ? createDiagnostic({
                    code: "invalid-value",
                    format: {
                        kind: typespecType.kind,
                    },
                    target,
                })
                : createDiagnostic({
                    code: "invalid-value",
                    messageId: "atPath",
                    format: {
                        kind: typespecType.kind,
                        path: path.join("."),
                    },
                    target,
                });
            return [undefined, [diagnostic]];
    }
}
function validateDecoratorUniqueOnNode(context, type, decorator) {
    compilerAssert("decorators" in type, "Type should have decorators");
    const sameDecorators = type.decorators.filter((x) => x.decorator === decorator &&
        x.node?.kind === SyntaxKind.DecoratorExpression &&
        x.node?.parent === type.node);
    if (sameDecorators.length > 1) {
        reportDiagnostic(context.program, {
            code: "duplicate-decorator",
            format: { decoratorName: "@" + decorator.name.slice(1) },
            target: context.decoratorTarget,
        });
        return false;
    }
    return true;
}
/**
 * Validate that a given decorator is not on a type or any of its base types.
 * Useful to check for decorator usage that conflicts with another decorator.
 * @param context Decorator context
 * @param type The type to check
 * @param badDecorator The decorator we don't want present
 * @param givenDecorator The decorator that is the reason why we don't want the bad decorator present
 * @param includeHeritage Whether to check base types for the bad decorator too
 * @returns Whether the decorator application is valid
 */
function validateDecoratorNotOnType(context, type, badDecorator, givenDecorator) {
    compilerAssert("decorators" in type, "Type should have decorators");
    const decAppsToCheck = [];
    let base = type;
    while (base) {
        decAppsToCheck.push(...base.decorators);
        base = getHeritage(base);
    }
    for (const decapp of decAppsToCheck) {
        if (decapp.decorator === badDecorator) {
            reportDiagnostic(context.program, {
                code: "decorator-conflict",
                format: {
                    decoratorName: "@" + badDecorator.name.slice(1),
                    otherDecoratorName: "@" + givenDecorator.name.slice(1),
                },
                target: context.decoratorTarget,
            });
            return false;
        }
    }
    return true;
    function getHeritage(type) {
        if (type.kind === "Model") {
            return type.baseModel;
        }
        else if (type.kind === "Scalar") {
            return type.baseScalar;
        }
        else {
            return undefined;
        }
    }
}
/**
 * Return the type of the property or the model itself.
 */
function getPropertyType(target) {
    if (target.kind === "ModelProperty") {
        return target.type;
    }
    else {
        return target;
    }
}

const deprecatedKey = createStateSymbol("deprecated");
/**
 * Check if the given type is deprecated
 * @param program Program
 * @param type Type
 */
function isDeprecated(program, type) {
    return program.stateMap(deprecatedKey).has(type);
}
/**
 * Returns complete deprecation details for the given type or node
 * @param program Program
 * @param typeOrNode A Type or Node to check for deprecation
 */
function getDeprecationDetails(program, typeOrNode) {
    function isType(maybeType) {
        return typeof maybeType.kind === "string";
    }
    // If we're looking at a type, pull the deprecation details from the state map
    if (isType(typeOrNode)) {
        return program.stateMap(deprecatedKey).get(typeOrNode);
    }
    else {
        // Look at the node for a deprecation directive
        const deprecatedDirective = (typeOrNode.directives ?? []).find((directive) => directive.target.sv === "deprecated");
        if (deprecatedDirective?.arguments[0].kind === SyntaxKind.StringLiteral) {
            return {
                message: deprecatedDirective.arguments[0].value,
            };
        }
    }
    return undefined;
}
/**
 * Mark the given type as deprecated with the provided details.
 * @param program Program
 * @param type Type
 * @param details Details of the deprecation
 */
function markDeprecated(program, type, details) {
    program.stateMap(deprecatedKey).set(type, details);
}

// Contains all intrinsic data setter or getter
// Anything that the TypeSpec check might should be here.
const stateKeys = {
    minValues: createStateSymbol("minValues"),
    maxValues: createStateSymbol("maxValues"),
    minValueExclusive: createStateSymbol("minValueExclusive"),
    maxValueExclusive: createStateSymbol("maxValueExclusive"),
    minLength: createStateSymbol("minLengthValues"),
    maxLength: createStateSymbol("maxLengthValues"),
    minItems: createStateSymbol("minItems"),
    maxItems: createStateSymbol("maxItems"),
    docs: createStateSymbol("docs"),
    returnDocs: createStateSymbol("returnsDocs"),
    errorsDocs: createStateSymbol("errorDocs"),
    discriminator: createStateSymbol("discriminator"),
};
// #region @minValue
function setMinValue(program, target, value) {
    program.stateMap(stateKeys.minValues).set(target, value);
}
function getMinValueAsNumeric(program, target) {
    return program.stateMap(stateKeys.minValues).get(target);
}
function getMinValue(program, target) {
    return getMinValueAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @minValue
// #region @maxValue
function setMaxValue(program, target, value) {
    program.stateMap(stateKeys.maxValues).set(target, value);
}
function getMaxValueAsNumeric(program, target) {
    return program.stateMap(stateKeys.maxValues).get(target);
}
function getMaxValue(program, target) {
    return getMaxValueAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @maxValue
// #region @minValueExclusive
function setMinValueExclusive(program, target, value) {
    program.stateMap(stateKeys.minValueExclusive).set(target, value);
}
function getMinValueExclusiveAsNumeric(program, target) {
    return program.stateMap(stateKeys.minValueExclusive).get(target);
}
function getMinValueExclusive(program, target) {
    return getMinValueExclusiveAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @minValueExclusive
// #region @maxValueExclusive
function setMaxValueExclusive(program, target, value) {
    program.stateMap(stateKeys.maxValueExclusive).set(target, value);
}
function getMaxValueExclusiveAsNumeric(program, target) {
    return program.stateMap(stateKeys.maxValueExclusive).get(target);
}
function getMaxValueExclusive(program, target) {
    return getMaxValueExclusiveAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @maxValueExclusive
// #region @minLength
function setMinLength(program, target, value) {
    program.stateMap(stateKeys.minLength).set(target, value);
}
/**
 * Get the minimum length of a string type as a {@link Numeric} value.
 * @param program Current program
 * @param target Type with the `@minLength` decorator
 */
function getMinLengthAsNumeric(program, target) {
    return program.stateMap(stateKeys.minLength).get(target);
}
function getMinLength(program, target) {
    return getMinLengthAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @minLength
// #region @maxLength
function setMaxLength(program, target, value) {
    program.stateMap(stateKeys.maxLength).set(target, value);
}
/**
 * Get the minimum length of a string type as a {@link Numeric} value.
 * @param program Current program
 * @param target Type with the `@maxLength` decorator
 */
function getMaxLengthAsNumeric(program, target) {
    return program.stateMap(stateKeys.maxLength).get(target);
}
function getMaxLength(program, target) {
    return getMaxLengthAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @maxLength
// #region @minItems
function setMinItems(program, target, value) {
    program.stateMap(stateKeys.minItems).set(target, value);
}
function getMinItemsAsNumeric(program, target) {
    return program.stateMap(stateKeys.minItems).get(target);
}
function getMinItems(program, target) {
    return getMinItemsAsNumeric(program, target)?.asNumber() ?? undefined;
}
// #endregion @minItems
// #region @minItems
function setMaxItems(program, target, value) {
    program.stateMap(stateKeys.maxItems).set(target, value);
}
function getMaxItemsAsNumeric(program, target) {
    return program.stateMap(stateKeys.maxItems).get(target);
}
function getMaxItems(program, target) {
    return getMaxItemsAsNumeric(program, target)?.asNumber() ?? undefined;
}
/** @internal */
function setDocData(program, target, key, data) {
    program.stateMap(getDocKey(key)).set(target, data);
}
function getDocKey(target) {
    switch (target) {
        case "self":
            return stateKeys.docs;
        case "returns":
            return stateKeys.returnDocs;
        case "errors":
            return stateKeys.errorsDocs;
    }
}
/**
 * @internal
 * Get the documentation information for the given type. In most cases you probably just want to use {@link getDoc}
 * @param program Program
 * @param target Type
 * @returns Doc data with source information.
 */
function getDocDataInternal(program, target, key) {
    return program.stateMap(getDocKey(key)).get(target);
}
/**
 * Get the documentation information for the given type. In most cases you probably just want to use {@link getDoc}
 * @param program Program
 * @param target Type
 * @returns Doc data with source information.
 */
function getDocData(program, target) {
    return getDocDataInternal(program, target, "self");
}
function setDiscriminator(program, entity, discriminator) {
    program.stateMap(stateKeys.discriminator).set(entity, discriminator);
}
function getDiscriminator(program, entity) {
    return program.stateMap(stateKeys.discriminator).get(entity);
}
function getDiscriminatedTypes(program) {
    return [...program.stateMap(stateKeys.discriminator).entries()];
}
const [getDiscriminatedOptions, setDiscriminatedOptions] = useStateMap(createStateSymbol("discriminated"));
// #endregion

function getDiscriminatedUnion(typeOrProgram, typeOrDiscriminator) {
    return getDiscriminatedUnionForUnion(typeOrProgram, typeOrDiscriminator);
}
/**
 * Run the validation on all discriminated models to make sure the discriminator are valid.
 * This has to be done after the checker so we can have the full picture of all the dervied models.
 */
function validateInheritanceDiscriminatedUnions(program) {
    for (const [type, discriminator] of getDiscriminatedTypes(program)) {
        // Union would have already reported the issue.
        if (type.kind === "Model") {
            const [_, diagnostics] = getDiscriminatedUnionFromInheritance(type, discriminator);
            program.reportDiagnostics(diagnostics);
        }
    }
}
function getDiscriminatedUnionForUnion(program, type) {
    const options = getDiscriminatedOptions(program, type);
    const diagnostics = [];
    if (options === undefined) {
        return [undefined, []];
    }
    const variants = new Map();
    let defaultVariant;
    // If there is not envelope then every variant needs to be a model.
    for (const variant of type.variants.values()) {
        if (typeof variant.name !== "string") {
            if (defaultVariant) {
                diagnostics.push(createDiagnostic({
                    code: "invalid-discriminated-union-variant",
                    messageId: "duplicateDefaultVariant",
                    target: variant,
                }));
            }
            else {
                defaultVariant = variant.type;
            }
            continue;
        }
        variants.set(variant.name, variant.type);
        if (options.envelope === "none") {
            if (variant.type.kind !== "Model") {
                diagnostics.push(createDiagnostic({
                    code: "invalid-discriminated-union-variant",
                    messageId: "noEnvelopeModel",
                    format: { name: variant.name.toString() },
                    target: variant,
                }));
                continue;
            }
            const prop = variant.type.properties.get(options.discriminatorPropertyName);
            if (prop !== undefined) {
                const key = getStringValue(prop.type);
                if (key !== variant.name) {
                    diagnostics.push(createDiagnostic({
                        code: "invalid-discriminated-union-variant",
                        messageId: "discriminantMismatch",
                        format: {
                            name: variant.name.toString(),
                            discriminant: options.discriminatorPropertyName,
                            propertyValue: key,
                            variantName: String(variant.name),
                        },
                        target: variant.type,
                    }));
                }
            }
        }
    }
    return [
        {
            type,
            options,
            variants,
        },
        diagnostics,
    ];
}
function getDiscriminatedUnionFromInheritance(type, discriminator) {
    const variants = new Map();
    const diagnostics = [];
    const duplicates = new DuplicateTracker();
    function checkForVariantsIn(current) {
        for (const derivedModel of current.derivedModels) {
            if (isTemplateDeclarationOrInstance(derivedModel)) {
                continue; // Skip template instances as they should be used with `model is`
            }
            const keys = getDiscriminatorValues(derivedModel, discriminator, diagnostics);
            if (keys === undefined) {
                if (derivedModel.derivedModels.length === 0) {
                    diagnostics.push(createDiagnostic({
                        code: "missing-discriminator-property",
                        format: { discriminator: discriminator.propertyName },
                        target: derivedModel,
                    }));
                }
                else {
                    checkForVariantsIn(derivedModel);
                }
            }
            else {
                for (const key of keys) {
                    duplicates.track(key, derivedModel);
                    variants.set(key, derivedModel);
                }
            }
        }
    }
    checkForVariantsIn(type);
    reportDuplicateDiscriminatorValues(duplicates, diagnostics);
    const discriminatedUnion = {
        kind: "legacy",
        propertyName: discriminator.propertyName,
        variants,
    };
    return [discriminatedUnion, diagnostics];
}
function reportDuplicateDiscriminatorValues(duplicates, diagnostics) {
    for (const [duplicateKey, models] of duplicates.entries()) {
        for (const model of models) {
            diagnostics.push(createDiagnostic({
                code: "invalid-discriminator-value",
                messageId: "duplicate",
                format: { discriminator: duplicateKey },
                target: model,
            }));
        }
    }
}
function getDiscriminatorProperty(model, discriminator, diagnostics) {
    const prop = model.properties.get(discriminator.propertyName);
    if (prop && prop.optional) {
        diagnostics.push(createDiagnostic({
            code: "invalid-discriminator-value",
            messageId: "required",
            target: prop,
        }));
    }
    return prop;
}
function getDiscriminatorValues(model, discriminator, diagnostics) {
    const prop = getDiscriminatorProperty(model, discriminator, diagnostics);
    if (!prop)
        return undefined;
    const keys = getStringValues(prop.type);
    if (keys.length === 0) {
        diagnostics.push(createDiagnostic({
            code: "invalid-discriminator-value",
            format: { kind: prop.type.kind },
            target: prop,
        }));
    }
    return keys;
}
function getStringValues(type) {
    switch (type.kind) {
        case "String":
            return [type.value];
        case "Union":
            return [...type.variants.values()].flatMap((x) => getStringValues(x.type)).filter(isDefined);
        case "EnumMember":
            return typeof type.value !== "number" ? [type.value ?? type.name] : [];
        case "UnionVariant":
            return getStringValues(type.type);
        default:
            return [];
    }
}
function getStringValue(type) {
    switch (type.kind) {
        case "String":
            return type.value;
        case "EnumMember":
            return typeof type.value !== "number" ? (type.value ?? type.name) : undefined;
        default:
            return undefined;
    }
}

const regex = /^(application|audio|font|example|image|message|model|multipart|text|video|x-(?:[0-9A-Za-z!#$%&'*+.^_`|~-]+))\/([0-9A-Za-z!#$%&'*+.^_`|~-]+)$/;
function parseMimeType(mimeType) {
    const match = mimeType.match(regex);
    if (match == null) {
        return undefined;
    }
    const type = {
        type: match[1],
        ...parseSubType(match[2]),
    };
    return type;
}
function parseSubType(value) {
    if (!value.includes("+"))
        return { subtype: value };
    const [subtype, suffix] = value.split("+", 2);
    return { subtype, suffix };
}

const [getKey, setKey] = useStateMap(createStateSymbol("key"));
function isKey(program, property) {
    return getKey(program, property) !== undefined;
}

const [getEncodedNamesMap, setEncodedNamesMap, getEncodedNamesStateMap] = useStateMap(createStateSymbol("encodedName"));
function $encodedName(context, target, mimeType, name) {
    let existing = getEncodedNamesMap(context.program, target);
    if (existing === undefined) {
        existing = new Map();
        setEncodedNamesMap(context.program, target, existing);
    }
    const mimeTypeObj = parseMimeType(mimeType);
    if (mimeTypeObj === undefined) {
        reportDiagnostic(context.program, {
            code: "invalid-mime-type",
            format: { mimeType },
            target: context.getArgumentTarget(0),
        });
    }
    else if (mimeTypeObj.suffix) {
        reportDiagnostic(context.program, {
            code: "no-mime-type-suffix",
            format: { mimeType, suffix: mimeTypeObj.suffix },
            target: context.getArgumentTarget(0),
        });
    }
    existing.set(mimeType, name);
}
function getEncodedName(program, target, mimeType) {
    const mimeTypeObj = parseMimeType(mimeType);
    if (mimeTypeObj === undefined) {
        return undefined;
    }
    const resolvedMimeType = mimeTypeObj?.suffix
        ? `${mimeTypeObj.type}/${mimeTypeObj.suffix}`
        : mimeType;
    return getEncodedNamesMap(program, target)?.get(resolvedMimeType);
}
/**
 * Resolve the encoded name for the given type when serialized to the given mime type.
 * If a specific value was provided by `@encodedName` decorator for that mime type it will return that otherwise it will return the name of the type.
 *
 * @example
 *
 * For the given
 * ```tsp
 * model Certificate {
 *   @encodedName("application/json", "exp")
 *   @encodedName("application/xml", "expiry")
 *   expireAt: utcDateTime;
 *
 * }
 * ```
 *
 * ```ts
 * resolveEncodedName(program, type, "application/json") // exp
 * resolveEncodedName(program, type, "application/merge-patch+json") // exp
 * resolveEncodedName(program, type, "application/xml") // expireAt
 * resolveEncodedName(program, type, "application/yaml") // expiry
 * ```
 */
function resolveEncodedName(program, target, mimeType) {
    return getEncodedName(program, target, mimeType) ?? target.name;
}
/**
 * Validate encoded names for conflicts. Validate the encoded names doesn't encode to an existing property name and that 2 encoded names don't map to the same name for the same mime type.
 * @internal
 */
function validateEncodedNamesConflicts(program) {
    const duplicateTrackers = new Map();
    function getOrCreateDuplicateTracker(type, mimeType) {
        let perMimeTypes = duplicateTrackers.get(type);
        if (perMimeTypes === undefined) {
            perMimeTypes = new Map();
            duplicateTrackers.set(type, perMimeTypes);
        }
        let tracker = perMimeTypes.get(mimeType);
        if (tracker === undefined) {
            tracker = new DuplicateTracker();
            perMimeTypes.set(mimeType, tracker);
        }
        return tracker;
    }
    for (const [target, map] of getEncodedNamesStateMap(program).entries()) {
        const scope = getScope(target);
        if (scope === undefined) {
            return;
        }
        for (const [mimeType, name] of map.entries()) {
            const duplicateTracker = getOrCreateDuplicateTracker(scope.parent, mimeType);
            duplicateTracker.track(name, target);
            if (scope.members.has(name)) {
                reportDiagnostic(program, {
                    code: "encoded-name-conflict",
                    format: { name, mimeType },
                    target: target,
                });
            }
        }
    }
    for (const perMimeTypes of duplicateTrackers.values()) {
        for (const [mimeType, tracker] of perMimeTypes.entries()) {
            for (const [duplicateName, items] of tracker.entries()) {
                for (const item of items) {
                    reportDiagnostic(program, {
                        code: "encoded-name-conflict",
                        messageId: "duplicate",
                        format: { name: duplicateName, mimeType },
                        target: item,
                    });
                }
            }
        }
    }
}
function getScope(type) {
    switch (type.kind) {
        case "ModelProperty":
            return type.model && { parent: type.model, members: type.model.properties };
        case "EnumMember":
            return { parent: type.enum, members: type.enum.members };
        case "UnionVariant":
            return { parent: type.union, members: type.union.variants };
        default:
            return undefined;
    }
}

const defaultOptions = {
    includeTemplateDeclaration: false,
};
/**
 * Navigate all types in the program.
 * @param program Program to navigate.
 * @param listeners Listener called when visiting types.
 * @param options Navigation options.
 */
function navigateProgram(program, listeners, options = {}) {
    const context = createNavigationContext(listeners, options);
    context.emit("root", program);
    navigateNamespaceType(program.getGlobalNamespaceType(), context);
}
/**
 * Navigate the given type and all the types that are used in it.
 * @param type Type to navigate.
 * @param listeners Listener for the types found.
 * @param options Navigation options
 */
function navigateType(type, listeners, options) {
    const context = createNavigationContext(listeners, options);
    navigateTypeInternal(type, context);
}
/**
 * Scope the current navigation to the given namespace.
 * @param namespace Namespace the traversal shouldn't leave.
 * @param listeners Type listeners.
 * @param options Scope options
 * @returns wrapped listeners that that can be used with `navigateType`
 */
function scopeNavigationToNamespace(namespace, listeners, options = {}) {
    const wrappedListeners = {};
    for (const [name, callback] of Object.entries(listeners)) {
        wrappedListeners[name] = (x) => {
            if (x !== namespace && "namespace" in x) {
                if (options.skipSubNamespaces && x.namespace !== namespace) {
                    return ListenerFlow.NoRecursion;
                }
                if (x.namespace && !isSubNamespace(x.namespace, namespace)) {
                    return ListenerFlow.NoRecursion;
                }
            }
            return callback(x);
        };
    }
    return wrappedListeners;
}
function navigateTypesInNamespace(namespace, listeners, options = {}) {
    navigateType(namespace, scopeNavigationToNamespace(namespace, listeners, options), options);
}
/**
 * Create a Semantic node listener from an event emitter.
 * @param eventEmitter Event emitter.
 * @returns Semantic node listener.
 */
function mapEventEmitterToNodeListener(eventEmitter) {
    const listener = {};
    for (const eventName of eventNames) {
        listener[eventName] = (...args) => {
            eventEmitter.emit(eventName, ...args);
        };
    }
    return listener;
}
function isSubNamespace(subNamespace, namespace) {
    let current = subNamespace;
    while (current !== undefined) {
        if (current === namespace) {
            return true;
        }
        current = current.namespace;
    }
    return false;
}
function createNavigationContext(listeners, options = {}) {
    return {
        visited: new Set(),
        emit: (key, ...args) => listeners[key]?.(...args),
        options: computeOptions(options),
    };
}
function computeOptions(options) {
    return { ...defaultOptions, ...options };
}
function navigateNamespaceType(namespace, context) {
    if (context.emit("namespace", namespace) === ListenerFlow.NoRecursion)
        return;
    for (const model of namespace.models.values()) {
        navigateModelType(model, context);
    }
    for (const scalar of namespace.scalars.values()) {
        navigateScalarType(scalar, context);
    }
    for (const operation of namespace.operations.values()) {
        navigateOperationType(operation, context);
    }
    for (const subNamespace of namespace.namespaces.values()) {
        if (!(namespace.name === "TypeSpec" && subNamespace.name === "Prototypes")) {
            navigateNamespaceType(subNamespace, context);
        }
    }
    for (const union of namespace.unions.values()) {
        navigateUnionType(union, context);
    }
    for (const iface of namespace.interfaces.values()) {
        navigateInterfaceType(iface, context);
    }
    for (const enumType of namespace.enums.values()) {
        navigateEnumType(enumType, context);
    }
    for (const decorator of namespace.decoratorDeclarations.values()) {
        navigateDecoratorDeclaration(decorator, context);
    }
    context.emit("exitNamespace", namespace);
}
function checkVisited(visited, item) {
    if (visited.has(item)) {
        return true;
    }
    visited.add(item);
    return false;
}
function shouldNavigateTemplatableType(context, type) {
    if (context.options.includeTemplateDeclaration) {
        return type.isFinished || isTemplateDeclaration(type);
    }
    else {
        return type.isFinished;
    }
}
function navigateOperationType(operation, context) {
    if (checkVisited(context.visited, operation)) {
        return;
    }
    if (!shouldNavigateTemplatableType(context, operation)) {
        return;
    }
    if (context.emit("operation", operation) === ListenerFlow.NoRecursion)
        return;
    for (const parameter of operation.parameters.properties.values()) {
        navigateTypeInternal(parameter, context);
    }
    navigateTypeInternal(operation.returnType, context);
    if (operation.sourceOperation) {
        navigateTypeInternal(operation.sourceOperation, context);
    }
    context.emit("exitOperation", operation);
}
function navigateModelType(model, context) {
    if (checkVisited(context.visited, model)) {
        return;
    }
    if (!shouldNavigateTemplatableType(context, model)) {
        return;
    }
    if (context.emit("model", model) === ListenerFlow.NoRecursion)
        return;
    for (const property of model.properties.values()) {
        navigateModelTypeProperty(property, context);
    }
    if (model.baseModel) {
        navigateModelType(model.baseModel, context);
    }
    if (model.indexer && model.indexer.value) {
        navigateTypeInternal(model.indexer.value, context);
    }
    if (context.options.visitDerivedTypes) {
        for (const derived of model.derivedModels) {
            navigateModelType(derived, context);
        }
    }
    context.emit("exitModel", model);
}
function navigateModelTypeProperty(property, context) {
    if (checkVisited(context.visited, property)) {
        return;
    }
    if (context.emit("modelProperty", property) === ListenerFlow.NoRecursion)
        return;
    navigateTypeInternal(property.type, context);
    context.emit("exitModelProperty", property);
}
function navigateScalarType(scalar, context) {
    if (checkVisited(context.visited, scalar)) {
        return;
    }
    if (context.emit("scalar", scalar) === ListenerFlow.NoRecursion)
        return;
    if (scalar.baseScalar) {
        navigateScalarType(scalar.baseScalar, context);
    }
    for (const constructor of scalar.constructors.values()) {
        navigateScalarConstructor(constructor, context);
    }
    context.emit("exitScalar", scalar);
}
function navigateInterfaceType(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (!shouldNavigateTemplatableType(context, type)) {
        return;
    }
    context.emit("interface", type);
    for (const op of type.operations.values()) {
        navigateOperationType(op, context);
    }
    context.emit("exitInterface", type);
}
function navigateEnumType(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    context.emit("enum", type);
    for (const member of type.members.values()) {
        navigateTypeInternal(member, context);
    }
    context.emit("exitEnum", type);
}
function navigateUnionType(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (!shouldNavigateTemplatableType(context, type)) {
        return;
    }
    if (context.emit("union", type) === ListenerFlow.NoRecursion)
        return;
    for (const variant of type.variants.values()) {
        navigateUnionTypeVariant(variant, context);
    }
    context.emit("exitUnion", type);
}
function navigateUnionTypeVariant(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("unionVariant", type) === ListenerFlow.NoRecursion)
        return;
    navigateTypeInternal(type.type, context);
    context.emit("exitUnionVariant", type);
}
function navigateTupleType(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("tuple", type) === ListenerFlow.NoRecursion)
        return;
    for (const value of type.values) {
        navigateTypeInternal(value, context);
    }
    context.emit("exitTuple", type);
}
function navigateStringTemplate(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("stringTemplate", type) === ListenerFlow.NoRecursion)
        return;
    for (const value of type.spans) {
        navigateTypeInternal(value, context);
    }
}
function navigateStringTemplateSpan(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("stringTemplateSpan", type) === ListenerFlow.NoRecursion)
        return;
    navigateTypeInternal(type.type, context);
}
function navigateTemplateParameter(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("templateParameter", type) === ListenerFlow.NoRecursion)
        return;
}
function navigateDecoratorDeclaration(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("decorator", type) === ListenerFlow.NoRecursion)
        return;
}
function navigateScalarConstructor(type, context) {
    if (checkVisited(context.visited, type)) {
        return;
    }
    if (context.emit("scalarConstructor", type) === ListenerFlow.NoRecursion)
        return;
}
function navigateTypeInternal(type, context) {
    switch (type.kind) {
        case "Model":
            return navigateModelType(type, context);
        case "Scalar":
            return navigateScalarType(type, context);
        case "ModelProperty":
            return navigateModelTypeProperty(type, context);
        case "Namespace":
            return navigateNamespaceType(type, context);
        case "Interface":
            return navigateInterfaceType(type, context);
        case "Enum":
            return navigateEnumType(type, context);
        case "Operation":
            return navigateOperationType(type, context);
        case "Union":
            return navigateUnionType(type, context);
        case "UnionVariant":
            return navigateUnionTypeVariant(type, context);
        case "Tuple":
            return navigateTupleType(type, context);
        case "StringTemplate":
            return navigateStringTemplate(type, context);
        case "StringTemplateSpan":
            return navigateStringTemplateSpan(type, context);
        case "TemplateParameter":
            return navigateTemplateParameter(type, context);
        case "Decorator":
            return navigateDecoratorDeclaration(type, context);
        case "ScalarConstructor":
            return navigateScalarConstructor(type, context);
        case "FunctionParameter":
        case "Boolean":
        case "EnumMember":
        case "Intrinsic":
        case "Number":
        case "String":
            return;
        default:
            return;
    }
}
// Return property from type, nesting into baseTypes as needed.
function getProperty(type, propertyName) {
    while (type.baseModel) {
        if (type.properties.has(propertyName)) {
            return type.properties.get(propertyName);
        }
        type = type.baseModel;
    }
    return type.properties.get(propertyName);
}
class EventEmitter {
    listeners = new Map();
    on(name, listener) {
        const array = this.listeners.get(name);
        if (array) {
            array.push(listener);
        }
        else {
            this.listeners.set(name, [listener]);
        }
    }
    emit(name, ...args) {
        const listeners = this.listeners.get(name);
        if (listeners) {
            for (const listener of listeners) {
                listener(...args);
            }
        }
    }
}
const eventNames = [
    "root",
    "templateParameter",
    "exitTemplateParameter",
    "scalar",
    "exitScalar",
    "model",
    "exitModel",
    "modelProperty",
    "exitModelProperty",
    "interface",
    "exitInterface",
    "enum",
    "exitEnum",
    "enumMember",
    "exitEnumMember",
    "namespace",
    "exitNamespace",
    "operation",
    "exitOperation",
    "string",
    "exitString",
    "number",
    "exitNumber",
    "boolean",
    "exitBoolean",
    "tuple",
    "exitTuple",
    "union",
    "exitUnion",
    "unionVariant",
    "exitUnionVariant",
    "intrinsic",
    "exitIntrinsic",
];

const [
/**
 * Check if the given operation is used to page through a list.
 * @param program Program
 * @param target Operation
 */
isList, markList, 
/** {@inheritdoc ListDecorator} */
listDecorator,] = createMarkerDecorator("list");
const [
/**
 * Check if the given property is the `@offset` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */
isOffsetProperty, markOffset, 
/** {@inheritdoc OffsetDecorator} */
offsetDecorator,] = createMarkerDecorator("offset", createNumericValidation("offset"));
const [
/**
 * Check if the given property is the `@pageIndex` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */
isPageIndexProperty, markPageIndexProperty, 
/** {@inheritdoc PageIndexDecorator} */
pageIndexDecorator,] = createMarkerDecorator("pageIndex", createNumericValidation("pageIndex"));
const [
/**
 * Check if the given property is the `@pageIndex` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */
isPageSizeProperty, markPageSizeProperty, 
/** {@inheritdoc PageSizeDecorator} */
pageSizeDecorator,] = createMarkerDecorator("pageSize", createNumericValidation("pageSize"));
const [
/**
 * Check if the given property is the `@pageIndex` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */
isPageItemsProperty, markPageItemsProperty, 
/** {@inheritdoc PageItemsDecorator} */
pageItemsDecorator,] = createMarkerDecorator("pageItems", (context, target) => {
    if (target.type.kind !== "Model" || !isArrayModelType(context.program, target.type)) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            messageId: "withExpected",
            format: { decorator: "continuationToken", expected: "Array", to: getTypeName(target.type) },
            target: context.decoratorTarget,
        });
        return false;
    }
    return true;
});
const [
/**
 * Check if the given property is the `@pageIndex` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */
isContinuationTokenProperty, markContinuationTokenProperty, 
/** {@inheritdoc ContinuationTokenDecorator} */
continuationTokenDecorator,] = createMarkerDecorator("continuationToken", (context, target) => {
    if (!isStringType(context.program, target.type)) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            messageId: "withExpected",
            format: { decorator: "continuationToken", expected: "string", to: getTypeName(target.type) },
            target: context.decoratorTarget,
        });
        return false;
    }
    return true;
});
const [
/**
 * Check if the given property is the `@nextLink` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */
isNextLink, markNextLink, 
/** {@inheritdoc NextLinkDecorator} */
nextLinkDecorator,] = createMarkerDecorator("nextLink");
const [
/**
 * Check if the given property is the `@prevLink` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */ isPrevLink, markPrevLink, 
/** {@inheritdoc PrevLinkDecorator} */
prevLinkDecorator,] = createMarkerDecorator("prevLink");
const [
/**
 * Check if the given property is the `@firstLink` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */ isFirstLink, markFirstLink, 
/** {@inheritdoc FirstLinkDecorator} */
firstLinkDecorator,] = createMarkerDecorator("firstLink");
const [
/**
 * Check if the given property is the `@lastLink` property for a paging operation.
 * @param program Program
 * @param target Model Property
 */ isLastLink, markLastLink, 
/** {@inheritdoc LastLinkDecorator} */
lastLinkDecorator,] = createMarkerDecorator("lastLink");
function validatePagingOperations(program) {
    navigateProgram(program, {
        operation: (op) => {
            if (isList(program, op)) {
                validatePagingOperation(program, op);
            }
        },
    });
}
const inputProps = new Set(["offset", "pageIndex", "pageSize", "continuationToken"]);
const outputProps = new Set([
    "pageItems",
    "nextLink",
    "prevLink",
    "firstLink",
    "lastLink",
    "continuationToken",
]);
function findPagingProperties(program, op, base, source) {
    const diags = createDiagnosticCollector();
    const acceptableProps = source === "input" ? inputProps : outputProps;
    const duplicateTracker = new DuplicateTracker();
    const data = {};
    navigateProperties(base, (property) => {
        const kind = diags.pipe(getPagingProperty(program, property));
        if (kind === undefined) {
            return;
        }
        duplicateTracker.track(kind, property);
        if (acceptableProps.has(kind)) {
            data[kind] = { property };
        }
        else {
            diags.add(createDiagnostic({
                code: "invalid-paging-prop",
                messageId: source === "input" ? "input" : "output",
                format: { kind },
                target: property,
            }));
        }
    });
    for (const [key, duplicates] of duplicateTracker.entries()) {
        for (const prop of duplicates) {
            diags.add(createDiagnostic({
                code: "duplicate-paging-prop",
                format: { kind: key, operationName: op.name },
                target: prop,
            }));
        }
    }
    return diags.wrap(data);
}
function getPagingOperation(program, op) {
    const diags = createDiagnosticCollector();
    const result = {
        input: diags.pipe(findPagingProperties(program, op, op.parameters, "input")),
        output: diags.pipe(findPagingProperties(program, op, op.returnType, "output")),
    };
    if (result.output.pageItems === undefined) {
        diags.add(createDiagnostic({
            code: "missing-paging-items",
            format: { operationName: op.name },
            target: op,
        }));
        return diags.wrap(undefined);
    }
    return diags.wrap(result);
}
function navigateProperties(type, callback) {
    switch (type.kind) {
        case "Model":
            for (const prop of type.properties.values()) {
                callback(prop);
            }
            break;
        case "Union":
            for (const member of type.variants.values()) {
                navigateProperties(member, callback);
            }
            break;
        case "UnionVariant":
            navigateProperties(type.type, callback);
            break;
    }
}
function getPagingProperty(program, prop) {
    const diagnostics = [];
    const props = {
        offset: isOffsetProperty(program, prop),
        pageIndex: isPageIndexProperty(program, prop),
        pageItems: isPageItemsProperty(program, prop),
        pageSize: isPageSizeProperty(program, prop),
        continuationToken: isContinuationTokenProperty(program, prop),
        nextLink: isNextLink(program, prop),
        prevLink: isPrevLink(program, prop),
        lastLink: isLastLink(program, prop),
        firstLink: isFirstLink(program, prop),
    };
    const defined = Object.entries(props).filter((x) => !!x[1]);
    if (defined.length > 1) {
        diagnostics.push(createDiagnostic({
            code: "incompatible-paging-props",
            format: { kinds: defined.map((x) => x[0]).join(", ") },
            target: prop,
        }));
    }
    if (defined.length === 0) {
        return [undefined, diagnostics];
    }
    return [defined[0][0], diagnostics];
}
function validatePagingOperation(program, op) {
    const [_, diagnostics] = getPagingOperation(program, op);
    program.reportDiagnostics(diagnostics);
}
function createMarkerDecorator(key, validate) {
    const [isLink, markLink] = useStateSet(createStateSymbol(key));
    const decorator = (...args) => {
        if (validate && !validate(...args)) {
            return;
        }
        const [context, target] = args;
        markLink(context.program, target);
    };
    return [isLink, markLink, decorator];
}
function createNumericValidation(decoratorName) {
    return (context, target) => {
        if (!isNumericType(context.program, target.type)) {
            reportDiagnostic(context.program, {
                code: "decorator-wrong-target",
                messageId: "withExpected",
                format: {
                    decorator: decoratorName,
                    expected: "numeric",
                    to: getTypeName(target.type),
                },
                target: context.decoratorTarget,
            });
            return false;
        }
        return true;
    };
}

const [getService, setService, getServiceMap] = useStateMap(Symbol.for("@typespec/compiler.services"));
/**
 * List all the services defined in the TypeSpec program
 * @param program Program
 * @returns List of service.
 */
function listServices(program) {
    return [...getServiceMap(program).values()].filter((x) => !Realm.realmForType.has(x.type));
}
/**
 * Check if the namespace is defined as a service.
 * @param program Program
 * @param namespace Namespace
 * @returns Boolean
 */
function isService(program, namespace) {
    return getServiceMap(program).has(namespace);
}
/**
 * Mark the given namespace as a service.
 * @param program Program
 * @param namespace Namespace
 * @param details Service details
 */
function addService(program, namespace, details = {}) {
    const serviceMap = getServiceMap(program);
    const existing = serviceMap.get(namespace) ?? {};
    setService(program, namespace, { ...existing, ...details, type: namespace });
}
const $service = (context, target, options) => {
    validateDecoratorUniqueOnNode(context, target, $service);
    addService(context.program, target, options);
};

// Copyright (c) Microsoft Corporation
// Licensed under the MIT license.
/**
 * A cache for the `TypeSpec.Visibility.Lifecycle` enum per Program instance.
 */
const LIFECYCLE_ENUM_CACHE = new WeakMap();
/**
 * Returns the instance of `TypeSpec.Visibility.Lifecycle` for the given `program`.
 *
 * @param program - the program to get the lifecycle visibility enum for
 * @returns a reference to the lifecycle visibility enum
 */
function getLifecycleVisibilityEnum(program) {
    const cached = LIFECYCLE_ENUM_CACHE.get(program);
    if (cached)
        return cached;
    const [type, diagnostics] = program.resolveTypeReference("TypeSpec.Lifecycle");
    compilerAssert(diagnostics.length === 0, "Encountered diagnostics when resolving the `TypeSpec.Lifecycle` visibility class enum");
    compilerAssert(type.kind === "Enum", "Expected `TypeSpec.Visibility.Lifecycle` to be an enum");
    LIFECYCLE_ENUM_CACHE.set(program, type);
    return type;
}

function getLocationContext(program, type) {
    const sourceLocation = getSourceLocation(type);
    if (sourceLocation.isSynthetic) {
        return { type: "synthetic" };
    }
    return program.getSourceFileLocationContext(sourceLocation.file);
}

/**
 * Flow control for mutators.
 *
 * When filtering types in a mutator, the filter function may return MutatorFlow flags to control how mutation should
 * proceed.
 *
 * @see {@link MutatorFilterFn}
 *
 * @experimental
 */
var MutatorFlow;
(function (MutatorFlow) {
    /**
     * Mutate the type and recur, further mutating the type's children. This is the default behavior.
     */
    MutatorFlow[MutatorFlow["MutateAndRecur"] = 0] = "MutateAndRecur";
    /**
     * If this flag is set, the type will not be mutated.
     */
    MutatorFlow[MutatorFlow["DoNotMutate"] = 1] = "DoNotMutate";
    /**
     * If this flag is set, the mutator will not proceed recursively into the children of the type.
     */
    MutatorFlow[MutatorFlow["DoNotRecur"] = 2] = "DoNotRecur";
})(MutatorFlow || (MutatorFlow = {}));
/**
 * Determines if a type is mutable.
 *
 * @experimental
 */
function isMutableTypeWithNamespace(type) {
    switch (type.kind) {
        case "TemplateParameter":
        case "Intrinsic":
        case "Decorator":
        case "FunctionParameter":
            return false;
        default:
            return true;
    }
}
/**
 * Determines if a type is mutable.
 *
 * @experimental
 */
function isMutableType(type) {
    switch (type.kind) {
        case "TemplateParameter":
        case "Intrinsic":
        case "Decorator":
        case "FunctionParameter":
        case "Namespace":
            return false;
        default:
            return true;
    }
}
// #endregion
// #region Mutator Application
const typeId = CustomKeyMap.objectKeyer();
const mutatorId = CustomKeyMap.objectKeyer();
const seen = new CustomKeyMap(([type, mutators]) => {
    const key = `${typeId.getKey(type)}-${[...mutators.values()]
        .map((v) => mutatorId.getKey(v))
        .join("-")}`;
    return key;
});
/**
 * Mutate the type graph, allowing namespaces to be mutated.
 *
 * **Warning**: This function will likely mutate the entire type graph. Most TypeSpec types relate to namespaces
 * in some way (e.g. through namespace parent links, or the `namespace` property of a Model).
 *
 * @param program - The program in which the `type` occurs.
 * @param mutators - An array of mutators to apply to the type graph rooted at `type`.
 * @param type - The type to mutate.
 *
 * @returns an object containing the mutated `type` and a nullable `Realm` in which the mutated type resides.
 *
 * @see {@link mutateSubgraph}
 *
 * @experimental
 */
function mutateSubgraphWithNamespace(program, mutators, type) {
    const engine = createMutatorEngine(program, mutators, {
        mutateNamespaces: true,
    });
    const mutated = engine.mutate(type);
    if (mutated === type) {
        return { realm: null, type };
    }
    return { realm: engine.realm, type: mutated };
}
/**
 * Mutate the type graph.
 *
 * Mutators clone the input `type`, creating a new type instance that is mutated in place.
 *
 * The mutator returns the mutated type and optionally a `realm` in which the mutated clone resides.
 *
 * @see {@link Mutator}
 * @see {@link Realm}
 *
 * **Warning**: Mutators _SHOULD NOT_ modify the source type. Modifications to the source type
 * will be visible to other emitters or libraries that view the original source type, and will
 * be sensitive to the order in which the mutator was applied. Only edit the `clone` type.
 * Furthermore, mutators must take care not to modify elements of the source and clone types
 * that are shared between the two types, such as the properties of any parent references
 * or the `decorators` of the type without taking care to clone them first.
 *
 * @param program - The program in which the `type` occurs.
 * @param mutators - An array of mutators to apply to the type graph rooted at `type`.
 * @param type - The type to mutate.
 *
 * @returns an object containing the mutated `type` and a nullable `Realm` in which the mutated type resides.
 *
 * @experimental
 */
function mutateSubgraph(program, mutators, type) {
    const engine = createMutatorEngine(program, mutators, {
        mutateNamespaces: false,
    });
    const mutated = engine.mutate(type);
    if (mutated === type) {
        return { realm: null, type };
    }
    return { realm: engine.realm, type: mutated };
}
function createMutatorEngine(program, mutators, options) {
    const realm = new Realm(program, `Mutator realm ${mutators.map((m) => m.name).join(", ")}`);
    const interstitialFunctions = [];
    let preparingNamespace = false;
    const muts = new Set(mutators);
    const postVisits = [];
    const namespacesVisitedContent = new Set();
    // If we are mutating namespace we need to make sure we mutate everything first
    if (options.mutateNamespaces) {
        preparingNamespace = true;
        // Prepare namespaces first
        mutateSubgraphWorker(program.getGlobalNamespaceType(), muts);
        preparingNamespace = false;
        postVisits.forEach((visit) => visit());
    }
    return {
        realm,
        mutate: (type) => {
            return mutateSubgraphWorker(type, muts);
        },
    };
    /** Resolve the mutators to apply. */
    function resolveMutators(activeMutators, type) {
        const mutatorsWithOptions = [];
        const newMutators = new Set(activeMutators.values());
        // step 1: see what mutators to run
        for (const mutator of activeMutators) {
            const record = mutator[type.kind];
            if (!record) {
                continue;
            }
            let mutationFn = null;
            let replaceFn = null;
            let mutate = false;
            let recurse = false;
            if (typeof record === "function") {
                mutationFn = record;
                mutate = true;
                recurse = true;
            }
            else {
                mutationFn = "mutate" in record ? record.mutate : null;
                replaceFn = "replace" in record ? record.replace : null;
                if (record.filter) {
                    const filterResult = record.filter(type, program, realm);
                    if (filterResult === true) {
                        mutate = true;
                        recurse = true;
                    }
                    else if (filterResult === false) {
                        mutate = false;
                        recurse = true;
                    }
                    else {
                        mutate = (filterResult & MutatorFlow.DoNotMutate) === 0;
                        recurse = (filterResult & MutatorFlow.DoNotRecur) === 0;
                    }
                }
                else {
                    mutate = true;
                    recurse = true;
                }
            }
            if (!recurse) {
                newMutators.delete(mutator);
            }
            if (mutate) {
                mutatorsWithOptions.push({ mutator, mutationFn, replaceFn });
            }
        }
        return { mutatorsWithOptions, newMutators };
    }
    function applyMutations(type, clone, mutatorsWithOptions) {
        let resolved = clone;
        for (const { mutationFn, replaceFn } of mutatorsWithOptions) {
            // todo: handle replace earlier in the mutation chain
            const result = (mutationFn ?? replaceFn)(type, resolved, program, realm);
            if (replaceFn && result !== undefined) {
                resolved = result;
            }
        }
        return resolved;
    }
    function mutateSubgraphWorker(type, activeMutators, mutateSubNamespace = true) {
        let existing = seen.get([type, activeMutators]);
        if (existing) {
            if (mutateSubNamespace &&
                existing.kind === "Namespace" &&
                !namespacesVisitedContent.has(existing)) {
                namespacesVisitedContent.add(existing);
                mutateSubMap(existing, "namespaces", true, activeMutators);
            }
            clearInterstitialFunctions();
            return existing;
        }
        // Mutating compiler types breaks a lot of things in the type checker. It is better to keep any of those as they are.
        if (getLocationContext(program, type).type === "compiler" && !isTemplateInstance(type)) {
            return type;
        }
        let clone = null;
        const { mutatorsWithOptions, newMutators } = resolveMutators(activeMutators, type);
        const mutatorsToApply = mutatorsWithOptions.map((v) => v.mutator);
        let mutating = false;
        // if we have no mutators to apply, let's bail out.
        if (mutatorsWithOptions.length === 0) {
            if (newMutators.size > 0) {
                // we might need to clone this type later if something in our subgraph needs mutated.
                interstitialFunctions.push(initializeClone);
                seen.set([type, activeMutators], type);
                visitSubgraph();
                visitParents();
                interstitialFunctions.pop();
                return clone ?? type;
            }
            else {
                // we don't need to clone this type, so let's just return it.
                return type;
            }
        }
        clearInterstitialFunctions();
        // step 2: see if we need to mutate based on the set of mutators we're actually going to run
        existing = seen.get([type, mutatorsToApply]);
        if (existing) {
            return existing;
        }
        // step 3: run the mutators
        initializeClone();
        const result = applyMutations(type, clone, mutatorsWithOptions);
        if (result !== clone) {
            clone = result;
            seen.set([type, activeMutators], clone);
            seen.set([type, mutatorsToApply], clone);
        }
        if (newMutators.size > 0) {
            if (preparingNamespace && type.kind === "Namespace") {
                compilerAssert(mutating, "Cannot be preparing namespaces and not have cloned it.");
                prepareNamespace(clone);
                postVisits.push(() => visitNamespaceContents(clone));
            }
            else {
                visitSubgraph();
            }
        }
        if (type.isFinished) {
            $(realm).type.finishType(clone);
        }
        if (type.kind === "Namespace" && mutateSubNamespace) {
            compilerAssert(mutating, "Cannot be preparing namespaces and not have cloned it.");
            visitSubNamespaces(clone);
        }
        if (type.kind !== "Namespace") {
            visitParents();
        }
        return clone;
        function initializeClone() {
            clone = $(realm).type.clone(type);
            mutating = true;
            seen.set([type, activeMutators], clone);
            seen.set([type, mutatorsToApply], clone);
        }
        function clearInterstitialFunctions() {
            for (const interstitial of interstitialFunctions) {
                interstitial();
            }
            interstitialFunctions.length = 0;
        }
        function mutateNamespaceProperty(root) {
            if (!options.mutateNamespaces) {
                return;
            }
            if ("namespace" in root && root.namespace) {
                const newNs = mutateSubgraphWorker(root.namespace, newMutators, false);
                compilerAssert(newNs.kind === "Namespace", `Expected to be mutated to a namespace`);
                clone.namespace = newNs;
            }
        }
        function prepareNamespace(root) {
            visitDecorators(root, true, newMutators);
            mutateNamespaceProperty(root);
        }
        function visitSubNamespaces(type) {
            namespacesVisitedContent.add(type);
            mutateSubMap(type, "namespaces", true, newMutators);
        }
        function visitNamespaceContents(root) {
            mutateSubMap(root, "models", mutating, newMutators);
            mutateSubMap(root, "operations", mutating, newMutators);
            mutateSubMap(root, "interfaces", mutating, newMutators);
            mutateSubMap(root, "enums", mutating, newMutators);
            mutateSubMap(root, "unions", mutating, newMutators);
            mutateSubMap(root, "scalars", mutating, newMutators);
        }
        function visitModel(root) {
            mutateSubMap(root, "properties", mutating, newMutators);
            if (root.indexer) {
                const res = mutateSubgraphWorker(root.indexer.value, newMutators);
                if (mutating) {
                    root.indexer.value = res;
                }
            }
            for (const [index, prop] of root.sourceModels.entries()) {
                const newModel = mutateSubgraphWorker(prop.model, newMutators);
                if (mutating) {
                    mutate(root.sourceModels[index]).model = newModel;
                }
            }
            mutateProperty(root, "sourceModel", mutating, newMutators);
            mutateProperty(root, "baseModel", mutating, newMutators);
            mutateSubArray(root, "derivedModels", mutating, newMutators);
        }
        function visitSubgraph() {
            const root = clone ?? type;
            switch (root.kind) {
                case "Namespace":
                    visitNamespaceContents(root);
                    break;
                case "Model":
                    visitModel(root);
                    break;
                case "ModelProperty":
                    mutateProperty(root, "type", mutating, newMutators);
                    mutateProperty(root, "sourceProperty", mutating, newMutators);
                    break;
                case "Operation":
                    mutateProperty(root, "parameters", mutating, newMutators);
                    mutateProperty(root, "returnType", mutating, newMutators);
                    mutateProperty(root, "sourceOperation", mutating, newMutators);
                    break;
                case "Interface":
                    mutateSubMap(root, "operations", mutating, newMutators);
                    break;
                case "Enum":
                    mutateSubMap(root, "members", mutating, newMutators);
                    break;
                case "EnumMember":
                    break;
                case "Union":
                    mutateSubMap(root, "variants", mutating, newMutators);
                    break;
                case "UnionVariant":
                    mutateProperty(root, "type", mutating, newMutators);
                    break;
                case "Scalar":
                    mutateSubMap(root, "constructors", mutating, newMutators);
                    mutateProperty(root, "baseScalar", mutating, newMutators);
                    mutateSubArray(root, "derivedScalars", mutating, newMutators);
                    break;
                case "ScalarConstructor":
                    mutateProperty(root, "scalar", mutating, newMutators);
                    break;
            }
            if ("templateMapper" in root) {
                mutateTemplateMapper(root, mutating, newMutators);
            }
            if ("decorators" in root) {
                visitDecorators(root, mutating, newMutators);
            }
            mutateNamespaceProperty(root);
        }
        // Parents needs to be visited after the type is finished
        function visitParents() {
            const root = clone ?? type;
            switch (root.kind) {
                case "ModelProperty":
                    mutateProperty(root, "model", mutating, newMutators);
                    break;
                case "Operation":
                    mutateProperty(root, "interface", mutating, newMutators);
                    break;
                case "EnumMember":
                    mutateProperty(root, "enum", mutating, newMutators);
                    break;
                case "UnionVariant":
                    mutateProperty(root, "union", mutating, newMutators);
                    break;
                case "ScalarConstructor":
                    mutateProperty(root, "scalar", mutating, newMutators);
                    break;
            }
        }
    }
    function visitDecorators(type, mutating, newMutators) {
        for (const [index, dec] of type.decorators.entries()) {
            const args = [];
            for (const arg of dec.args) {
                const jsValue = typeof arg.jsValue === "object" &&
                    isType(arg.jsValue) &&
                    isMutableTypeWithNamespace(arg.jsValue)
                    ? mutateSubgraphWorker(arg.jsValue, newMutators)
                    : arg.jsValue;
                args.push({
                    ...arg,
                    value: isType(arg.value) && isMutableTypeWithNamespace(arg.value)
                        ? mutateSubgraphWorker(arg.value, newMutators)
                        : arg.value,
                    jsValue,
                });
            }
            if (mutating) {
                type.decorators[index] = { ...dec, args };
            }
        }
    }
    function mutateTemplateMapper(type, mutating, newMutators) {
        if (type.templateMapper === undefined) {
            return;
        }
        const mutatedMapper = {
            ...type.templateMapper,
            args: [],
            map: new Map(),
        };
        for (const arg of type.templateMapper.args) {
            mutate(mutatedMapper.args).push(mutateSubgraphWorker(arg, newMutators));
        }
        for (const [param, paramType] of type.templateMapper.map) {
            mutatedMapper.map.set(param, mutateSubgraphWorker(paramType, newMutators));
        }
        if (mutating) {
            type.templateMapper = mutatedMapper;
        }
    }
    function mutateSubMap(type, prop, mutate, newMutators) {
        for (const [key, value] of type[prop].entries()) {
            const newValue = mutateSubgraphWorker(value, newMutators);
            if (mutate) {
                type[prop].set(key, newValue);
                if (newValue.name !== value.name) {
                    type[prop].rekey(key, newValue.name);
                }
            }
        }
    }
    function mutateSubArray(type, prop, mutate, newMutators) {
        for (const [index, value] of type[prop].entries()) {
            const newValue = mutateSubgraphWorker(value, newMutators);
            if (mutate) {
                type[prop][index] = newValue;
            }
        }
    }
    function mutateProperty(type, prop, mutating, newMutators) {
        if (type[prop] === undefined) {
            return;
        }
        const newValue = mutateSubgraphWorker(type[prop], newMutators);
        if (mutating) {
            type[prop] = newValue;
        }
    }
}
// #endregion

// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
const $withDefaultKeyVisibility = (context, entity, visibility) => {
    const keyProperties = [...entity.properties].filter(([_, prop]) => {
        // Keep track of any key property without a visibility
        return isKey(context.program, prop) && !getRawVisibilityStore(context.program, prop);
    });
    // For each key property without a visibility, clone it and add the specified
    // default visibility value
    for (const [name, keyProp] of keyProperties) {
        entity.properties.set(name, context.program.checker.cloneType(keyProp, {
            decorators: [
                ...keyProp.decorators,
                {
                    decorator: $visibility,
                    args: [
                        {
                            value: visibility,
                            jsValue: visibility,
                        },
                    ],
                },
            ],
        }));
    }
};
const [getOperationVisibilityConfigRaw, setOperationVisibilityConfigRaw] = useStateMap(createStateSymbol("operationVisibilityConfig"));
function getOperationVisibilityConfig(program, operation) {
    let config = getOperationVisibilityConfigRaw(program, operation);
    if (!config) {
        config = {};
        setOperationVisibilityConfigRaw(program, operation, config);
    }
    return config;
}
const $parameterVisibility = (context, operation, ...modifiers) => {
    validateDecoratorUniqueOnNode(context, operation, $parameterVisibility);
    if (modifiers.length === 0) {
        reportDiagnostic(context.program, {
            code: "operation-visibility-constraint-empty",
            messageId: "parameter",
            target: context.decoratorTarget,
        });
        return;
    }
    getOperationVisibilityConfig(context.program, operation).parameters = modifiers.map((m) => m.value);
};
/**
 * An empty visibility provider. This provider returns an empty filter that considers all properties visible. This filter
 * is used when no context-specific visibility provider is available.
 *
 * When working with an HTTP specification, use the `HttpVisibilityProvider` from the `@typespec/http` library instead.
 */
const EmptyVisibilityProvider = {
    parameters: () => ({}),
    returnType: () => ({}),
};
/**
 * Get the visibility filter that should apply to the parameters of the given operation, or `undefined` if no parameter
 * visibility is set.
 *
 * If you are not working in a protocol that has specific implicit visibility semantics, you can use the
 * {@link EmptyVisibilityProvider} as a default provider. If you working in a protocol or context where parameters have
 * implicit visibility transformations (like HTTP), you should use the visibility provider from that library (for HTTP,
 * use the `HttpVisibilityProvider` from the `@typespec/http` library).
 *
 * @param program - the Program in which the operation is defined
 * @param operation - the Operation to get the parameter visibility filter for
 * @param defaultProvider - a provider for visibility filters that apply when no visibility constraints are explicitly
 *                         set. Defaults to an empty provider that returns an empty filter if not provided.
 * @returns a visibility filter for the parameters of the operation, or `undefined` if no parameter visibility is set
 */
function getParameterVisibilityFilter(program, operation, defaultProvider) {
    const operationVisibilityConfig = getOperationVisibilityConfig(program, operation);
    if (!operationVisibilityConfig.parameters)
        return defaultProvider.parameters(program, operation);
    compilerAssert(operationVisibilityConfig.parameters.length !== 0, "Empty parameter visibility constraint.");
    return {
        // WARNING: the HTTP library depends on `any` being the only key in the filter object returned by this method.
        //          if you change this logic, you will need to update the HTTP library to account for differences in the
        //          returned object. HTTP does not currently have a way to express `all` or `none` constraints in the same
        //          way that the core visibility system does.
        any: new Set(operationVisibilityConfig.parameters),
    };
}
const $returnTypeVisibility = (context, operation, ...modifiers) => {
    validateDecoratorUniqueOnNode(context, operation, $parameterVisibility);
    if (modifiers.length === 0) {
        reportDiagnostic(context.program, {
            code: "operation-visibility-constraint-empty",
            messageId: "returnType",
            target: context.decoratorTarget,
        });
        return;
    }
    getOperationVisibilityConfig(context.program, operation).returnType = modifiers.map((m) => m.value);
};
/**
 * Get the visibility filter that should apply to the return type of the given operation, or `undefined` if no return
 * type visibility is set.
 *
 * @param program - the Program in which the operation is defined
 * @param operation - the Operation to get the return type visibility filter for
 * @param defaultProvider - a provider for visibility filters that apply when no visibility constraints are explicitly
 *                          set. Defaults to an empty provider that returns an empty filter if not provided.
 * @returns a visibility filter for the return type of the operation, or `undefined` if no return type visibility is set
 */
function getReturnTypeVisibilityFilter(program, operation, defaultProvider) {
    const visibilityConfig = getOperationVisibilityConfig(program, operation);
    if (!visibilityConfig.returnType)
        return defaultProvider.returnType(program, operation);
    compilerAssert(visibilityConfig.returnType.length !== 0, "Empty return type visibility constraint.");
    return {
        // WARNING: the HTTP library depends on `any` being the only key in the filter object returned by this method.
        //          if you change this logic, you will need to update the HTTP library to account for differences in the
        //          returned object. HTTP does not currently have a way to express `all` or `none` constraints in the same
        //          way that the core visibility system does.
        any: new Set(visibilityConfig.returnType),
    };
}
// #endregion
// #region Core Visibility Decorators
// -- @visibility decorator ---------------------
const $visibility = (context, target, ...modifiers) => {
    addVisibilityModifiers(context.program, target, modifiers.map((m) => m.value), context);
};
// -- @removeVisibility decorator ---------------------
const $removeVisibility = (context, target, ...visibilities) => {
    removeVisibilityModifiers(context.program, target, visibilities.map((v) => v.value));
};
// -- @invisible decorator ---------------------
const $invisible = (context, target, visibilityClass) => {
    clearVisibilityModifiersForClass(context.program, target, visibilityClass);
};
// -- @defaultVisibility decorator ------------------
const $defaultVisibility = (context, target, ...visibilities) => {
    validateDecoratorUniqueOnNode(context, target, $defaultVisibility);
    const modifierSet = new Set();
    for (const visibility of visibilities) {
        if (visibility.value.enum !== target) {
            reportDiagnostic(context.program, {
                code: "default-visibility-not-member",
                target: context.decoratorTarget,
            });
        }
        else {
            modifierSet.add(visibility.value);
        }
    }
    setDefaultModifierSetForVisibilityClass(context.program, target, modifierSet);
};
// #endregion
// #region Legacy Visibility Transforms
// -- @withVisibility decorator ---------------------
const $withVisibility = (context, target, ...modifiers) => {
    const modifierMembers = modifiers.map((m) => m.value);
    const filter = {
        all: new Set(modifierMembers),
    };
    const visibilityClasses = new Set(modifierMembers.map((m) => m.enum));
    filterModelPropertiesInPlace(target, (p) => isVisible(context.program, p, filter));
    for (const p of target.properties.values()) {
        for (const c of visibilityClasses) {
            resetVisibilityModifiersForClass(context.program, p, c);
        }
    }
};
// -- @withUpdateableProperties decorator ----------------------
/**
 * Filters a model for properties that are updateable.
 *
 * @param context - the program context
 * @param target - Model to filter for updateable properties
 */
const $withUpdateableProperties = (context, target) => {
    if (!validateDecoratorTarget(context, target, "@withUpdateableProperties", "Model")) {
        return;
    }
    const filter = {
        all: new Set([getLifecycleVisibilityEnum(context.program).members.get("Update")]),
    };
    filterModelPropertiesInPlace(target, (p) => isVisible(context.program, p, filter));
};
// #endregion
// #region Mutator Driven Transforms
// -- @withVisibilityFilter decorator ----------------------
const $withVisibilityFilter = (context, target, _filter) => {
    const filter = VisibilityFilter.fromDecoratorArgument(_filter);
    const vfMutator = createVisibilityFilterMutator(filter, {
        decoratorFn: $withVisibilityFilter,
    });
    const { type } = mutateSubgraph(context.program, [vfMutator], target);
    target.properties = type.properties;
};
// -- @withLifecycleUpdate decorator ----------------------
const $withLifecycleUpdate = (context, target) => {
    const lifecycle = getLifecycleVisibilityEnum(context.program);
    const lifecycleUpdate = {
        all: new Set([lifecycle.members.get("Update")]),
    };
    const lifecycleCreateOrUpdate = {
        any: new Set([lifecycle.members.get("Create"), lifecycle.members.get("Update")]),
    };
    const createOrUpdateMutator = createVisibilityFilterMutator(lifecycleCreateOrUpdate);
    const updateMutator = createVisibilityFilterMutator(lifecycleUpdate, {
        recur: createOrUpdateMutator,
        decoratorFn: $withLifecycleUpdate,
    });
    const { type } = mutateSubgraph(context.program, [updateMutator], target);
    target.properties = type.properties;
};
/**
 * Create a mutator that applies a visibility filter to a type.
 *
 * @param filter - The visibility filter to apply
 * @param options - optional settings for the mutator
 * @returns
 */
function createVisibilityFilterMutator(filter, options = {}) {
    const visibilityClasses = VisibilityFilter.getVisibilityClasses(filter);
    const mpMutator = {
        name: "VisibilityFilterProperty",
        ModelProperty: {
            filter: () => MutatorFlow.DoNotRecur,
            mutate: (prop, clone, program) => {
                // We need to create a copy of the decorators array to avoid modifying the original.
                // Decorators are _NOT_ cloned by the type kit, so we have to be careful not to modify the decorator arguments
                // of the original type.
                const decorators = [];
                for (const decorator of prop.decorators) {
                    const decFn = decorator.decorator;
                    if (decFn === $visibility || decFn === $removeVisibility) {
                        const nextArgs = decorator.args.filter((arg) => {
                            if (arg.value.entityKind !== "Value")
                                return false;
                            const isString = arg.value.valueKind === "StringValue";
                            const isOperativeVisibility = arg.value.valueKind === "EnumValue" && visibilityClasses.has(arg.value.value.enum);
                            return !(isString || isOperativeVisibility);
                        });
                        if (nextArgs.length > 0) {
                            decorators.push({
                                ...decorator,
                                args: nextArgs,
                            });
                        }
                    }
                    else if (decFn !== $invisible) {
                        decorators.push(decorator);
                    }
                }
                clone.decorators = decorators;
                for (const visibilityClass of visibilityClasses) {
                    resetVisibilityModifiersForClass(program, clone, visibilityClass);
                }
                if (isMutableType(prop.type)) {
                    clone.type = mutateSubgraph(program, [options.recur ?? self], prop.type).type;
                }
            },
        },
    };
    const self = {
        name: "VisibilityFilter",
        Union: {
            filter: () => MutatorFlow.DoNotRecur,
            mutate: (union, clone, program) => {
                for (const [key, member] of union.variants) {
                    if (member.type.kind === "Model" || member.type.kind === "Union") {
                        const variant = {
                            ...member,
                            type: mutateSubgraph(program, [self], member.type).type,
                        };
                        clone.variants.set(key, variant);
                    }
                }
            },
        },
        Model: {
            filter: () => MutatorFlow.DoNotRecur,
            mutate: (model, clone, program, realm) => {
                for (const [key, prop] of model.properties) {
                    if (!isVisible(program, prop, filter)) {
                        // Property is not visible, remove it
                        clone.properties.delete(key);
                        realm.remove(clone);
                    }
                    else {
                        const mutated = mutateSubgraph(program, [mpMutator], prop);
                        clone.properties.set(key, mutated.type);
                    }
                }
                if (options.decoratorFn) {
                    clone.decorators = clone.decorators.filter((d) => d.decorator !== options.decoratorFn);
                }
            },
        },
        ModelProperty: {
            filter: () => MutatorFlow.DoNotRecur,
            mutate: (prop, clone, program) => {
                if (isMutableType(prop.type)) {
                    clone.type = mutateSubgraph(program, [self], prop.type).type;
                }
            },
        },
        Tuple: {
            filter: () => MutatorFlow.DoNotRecur,
            mutate: (tuple, clone, program) => {
                for (const [index, element] of tuple.values.entries()) {
                    if (isMutableType(element)) {
                        clone.values[index] = mutateSubgraph(program, [self], element).type;
                    }
                }
            },
        },
    };
    return self;
}
// #endregion

function replaceTemplatedStringFromProperties(formatString, sourceObject) {
    // Template parameters are not valid source objects, just skip them
    if (sourceObject.kind === "TemplateParameter") {
        return formatString;
    }
    return formatString.replace(/{(\w+)}/g, (_, propName) => {
        return sourceObject[propName];
    });
}
const [getSummary, setSummary] = useStateMap(createStateSymbol("summary"));
/**
 * @summary attaches a documentation string. It is typically used to give a short, single-line
 * description, and can be used in combination with or instead of @doc.
 *
 * The first argument to @summary is a string, which may contain template parameters, enclosed in braces,
 * which are replaced with an attribute for the type (commonly "name") passed as the second (optional) argument.
 *
 * @summary can be specified on any language element -- a model, an operation, a namespace, etc.
 */
const $summary = (context, target, text, sourceObject) => {
    if (sourceObject) {
        text = replaceTemplatedStringFromProperties(text, sourceObject);
    }
    setSummary(context.program, target, text);
};
/**
 * @doc attaches a documentation string. Works great with multi-line string literals.
 *
 * The first argument to @doc is a string, which may contain template parameters, enclosed in braces,
 * which are replaced with an attribute for the type (commonly "name") passed as the second (optional) argument.
 *
 * @doc can be specified on any language element -- a model, an operation, a namespace, etc.
 */
const $doc = (context, target, text, sourceObject) => {
    validateDecoratorUniqueOnNode(context, target, $doc);
    if (sourceObject) {
        text = replaceTemplatedStringFromProperties(text, sourceObject);
    }
    setDocData(context.program, target, "self", { value: text, source: "decorator" });
};
/**
 * Get the documentation string for the given type.
 * @param program Program
 * @param target Type
 * @returns Documentation value
 */
function getDoc(program, target) {
    return getDocDataInternal(program, target, "self")?.value;
}
const $returnsDoc = (context, target, text) => {
    validateDecoratorUniqueOnNode(context, target, $doc);
    setDocData(context.program, target, "returns", { value: text, source: "decorator" });
};
/**
 * Get the documentation information for the return success types of an operation. In most cases you probably just want to use {@link getReturnsDoc}
 * @param program Program
 * @param target Type
 * @returns Doc data with source information.
 */
function getReturnsDocData(program, target) {
    return getDocDataInternal(program, target, "returns");
}
/**
 * Get the documentation string for the return success types of an operation.
 * @param program Program
 * @param target Type
 * @returns Documentation value
 */
function getReturnsDoc(program, target) {
    return getDocDataInternal(program, target, "returns")?.value;
}
const $errorsDoc = (context, target, text) => {
    validateDecoratorUniqueOnNode(context, target, $doc);
    setDocData(context.program, target, "errors", { value: text, source: "decorator" });
};
/**
 * Get the documentation information for the return errors types of an operation. In most cases you probably just want to use {@link getErrorsDoc}
 * @param program Program
 * @param target Type
 * @returns Doc data with source information.
 */
function getErrorsDocData(program, target) {
    return getDocDataInternal(program, target, "errors");
}
/**
 * Get the documentation string for the return errors types of an operation.
 * @param program Program
 * @param target Type
 * @returns Documentation value
 */
function getErrorsDoc(program, target) {
    return getDocDataInternal(program, target, "errors")?.value;
}
const $inspectType = (context, target, text) => {
    // eslint-disable-next-line no-console
    if (text)
        console.log(text);
    // eslint-disable-next-line no-console
    console.dir(target, { depth: 3 });
};
const $inspectTypeName = (context, target, text) => {
    // eslint-disable-next-line no-console
    if (text)
        console.log(text);
    // eslint-disable-next-line no-console
    console.log(getTypeName(target));
};
function isStringType(program, target) {
    const stringType = program.checker.getStdType("string");
    return (target.kind === "Scalar" && program.checker.isTypeAssignableTo(target, stringType, target)[0]);
}
function isNumericType(program, target) {
    const numericType = program.checker.getStdType("numeric");
    return (target.kind === "Scalar" && program.checker.isTypeAssignableTo(target, numericType, target)[0]);
}
/**
 * Check the given type is matching the given condition or is a union of null and types matching the condition.
 * @param type Type to test
 * @param condition Condition
 * @returns Boolean
 */
function isTypeIn(type, condition) {
    if (type.kind === "Union") {
        return [...type.variants.values()].some((v) => condition(v.type));
    }
    return condition(type);
}
function validateTargetingANumeric(context, target, decoratorName) {
    const valid = isTypeIn(getPropertyType(target), (x) => isNumericType(context.program, x));
    if (!valid) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            format: {
                decorator: decoratorName,
                to: `type it is not a numeric`,
            },
            target: context.decoratorTarget,
        });
    }
    return valid;
}
/**
 * Validate the given target is a string type or a union containing at least a string type.
 */
function validateTargetingAString(context, target, decoratorName) {
    const valid = isTypeIn(getPropertyType(target), (x) => isStringType(context.program, x));
    if (!valid) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            format: {
                decorator: decoratorName,
                to: `type it is not a string`,
            },
            target: context.decoratorTarget,
        });
    }
    return valid;
}
// -- @error decorator ----------------------
const [getErrorState, setErrorState] = useStateSet(createStateSymbol("error"));
/**
 * `@error` decorator marks a model as an error type.
 *  Any derived models (using extends) will also be seen as error types.
 */
const $error = (context, entity) => {
    validateDecoratorUniqueOnNode(context, entity, $error);
    setErrorState(context.program, entity);
};
/**
 * Check if the type is an error model or a descendant of an error model.
 */
function isErrorModel(program, target) {
    if (target.kind !== "Model") {
        return false;
    }
    let current = target;
    while (current) {
        if (getErrorState(program, current)) {
            return true;
        }
        current = current.baseModel;
    }
    return false;
}
// -- @mediaTypeHint decorator --------------
const [_getMediaTypeHint, setMediaTypeHint] = useStateMap(createStateSymbol("mediaTypeHint"));
const $mediaTypeHint = (context, target, mediaType) => {
    validateDecoratorUniqueOnNode(context, target, $mediaTypeHint);
    const mimeTypeObj = parseMimeType(mediaType);
    if (mimeTypeObj === undefined) {
        reportDiagnostic(context.program, {
            code: "invalid-mime-type",
            format: { mimeType: mediaType },
            target: context.getArgumentTarget(0),
        });
    }
    else if (mimeTypeObj.suffix) {
        reportDiagnostic(context.program, {
            code: "no-mime-type-suffix",
            format: { mimeType: mediaType, suffix: mimeTypeObj.suffix },
            target: context.getArgumentTarget(0),
        });
    }
    setMediaTypeHint(context.program, target, mediaType);
};
/**
 * Get the default media type hint for the given target type.
 *
 * This value is a hint _ONLY_. Emitters are not required to use it, but may use it to get the default media type
 * associated with a TypeSpec type.
 *
 * @param program - the Program containing the target
 * @param target - the target to get the MIME type for
 * @returns the default media type hint for the target, if any
 */
function getMediaTypeHint(program, target) {
    switch (target.kind) {
        case "Scalar":
            // This special-casing is necessary because we cannot apply a decorator with a `valueof string` argument to
            // `string` itself. It creates a circular dependency in the checker where initializing the decorator value depends
            // on `string` already being initialized, which it isn't if we're still checking its decorators. This simple
            // special-casing allows us to avoid that circularity or having to special-case the initialization of string
            // itself.
            const isTypeSpecString = target.name === "string" &&
                target.namespace?.name === "TypeSpec" &&
                target.namespace.namespace === program.getGlobalNamespaceType();
            if (isTypeSpecString)
                return "text/plain";
        // Intentional fallthrough
        // eslint-disable-next-line no-fallthrough
        case "Union":
        case "Enum":
        case "Model": {
            const hint = _getMediaTypeHint(program, target);
            if (!hint) {
                // Look up the hierarchy for a MIME type hint if we don't have one on this type.
                const ancestor = target.kind === "Model"
                    ? target.baseModel
                    : target.kind === "Scalar"
                        ? target.baseScalar
                        : undefined;
                if (ancestor) {
                    return getMediaTypeHint(program, ancestor);
                }
            }
            return hint;
        }
        default:
            return undefined;
    }
}
// -- @format decorator ---------------------
const [getFormat, setFormat] = useStateMap(createStateSymbol("format"));
/**
 * `@format` - specify the data format hint for a string type
 *
 * The first argument is a string that identifies the format that the string type expects.  Any string
 * can be entered here, but a TypeSpec emitter must know how to interpret
 *
 * For TypeSpec specs that will be used with an OpenAPI emitter, the OpenAPI specification describes possible
 * valid values for a string type's format:
 *
 * https://github.com/OAI/OpenAPI-Specification/blob/3.0.3/versions/3.0.3.md#dataTypes
 *
 * `@format` can be specified on a type that extends from `string` or a `string`-typed model property.
 */
const $format = (context, target, format) => {
    validateDecoratorUniqueOnNode(context, target, $format);
    if (!validateTargetingAString(context, target, "@format")) {
        return;
    }
    setFormat(context.program, target, format);
};
// -- @pattern decorator ---------------------
const [getPatternData, setPatternData] = useStateMap(createStateSymbol("patternValues"));
const $pattern = (context, target, pattern, validationMessage) => {
    validateDecoratorUniqueOnNode(context, target, $pattern);
    if (!validateTargetingAString(context, target, "@pattern")) {
        return;
    }
    try {
        new RegExp(pattern);
    }
    catch (e) {
        reportDiagnostic(context.program, {
            code: "invalid-pattern-regex",
            target: target,
        });
    }
    const patternData = {
        pattern,
        validationMessage,
    };
    setPatternData(context.program, target, patternData);
};
/**
 * Gets the pattern regular expression associated with a given type, if one has been set.
 *
 * @see getPatternData
 *
 * @param program - the Program containing the target Type
 * @param target - the type to get the pattern for
 * @returns the pattern string, if one was set
 */
function getPattern(program, target) {
    return getPatternData(program, target)?.pattern;
}
// -- @minLength decorator ---------------------
const $minLength = (context, target, minLength) => {
    validateDecoratorUniqueOnNode(context, target, $minLength);
    if (!validateTargetingAString(context, target, "@minLength") ||
        !validateRange(context, minLength, getMaxLengthAsNumeric(context.program, target))) {
        return;
    }
    setMinLength(context.program, target, minLength);
};
// -- @maxLength decorator ---------------------
const $maxLength = (context, target, maxLength) => {
    validateDecoratorUniqueOnNode(context, target, $maxLength);
    if (!validateTargetingAString(context, target, "@maxLength") ||
        !validateRange(context, getMinLengthAsNumeric(context.program, target), maxLength)) {
        return;
    }
    setMaxLength(context.program, target, maxLength);
};
// -- @minItems decorator ---------------------
const $minItems = (context, target, minItems) => {
    validateDecoratorUniqueOnNode(context, target, $minItems);
    if (!isArrayModelType(context.program, target.kind === "Model" ? target : target.type)) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            format: {
                decorator: "@minItems",
                to: `non Array type`,
            },
            target: context.decoratorTarget,
        });
    }
    if (!validateRange(context, minItems, getMaxItemsAsNumeric(context.program, target))) {
        return;
    }
    setMinItems(context.program, target, minItems);
};
// -- @maxLength decorator ---------------------
const $maxItems = (context, target, maxItems) => {
    validateDecoratorUniqueOnNode(context, target, $maxItems);
    if (!isArrayModelType(context.program, target.kind === "Model" ? target : target.type)) {
        reportDiagnostic(context.program, {
            code: "decorator-wrong-target",
            format: {
                decorator: "@maxItems",
                to: `non Array type`,
            },
            target: context.decoratorTarget,
        });
    }
    if (!validateRange(context, getMinItemsAsNumeric(context.program, target), maxItems)) {
        return;
    }
    setMaxItems(context.program, target, maxItems);
};
// -- @minValue decorator ---------------------
const $minValue = (context, target, minValue) => {
    validateDecoratorUniqueOnNode(context, target, $minValue);
    validateDecoratorNotOnType(context, target, $minValueExclusive, $minValue);
    const { program } = context;
    if (!validateTargetingANumeric(context, target, "@minValue")) {
        return;
    }
    if (!validateRange(context, minValue, getMaxValueAsNumeric(context.program, target) ??
        getMaxValueExclusiveAsNumeric(context.program, target))) {
        return;
    }
    setMinValue(program, target, minValue);
};
// -- @maxValue decorator ---------------------
const $maxValue = (context, target, maxValue) => {
    validateDecoratorUniqueOnNode(context, target, $maxValue);
    validateDecoratorNotOnType(context, target, $maxValueExclusive, $maxValue);
    const { program } = context;
    if (!validateTargetingANumeric(context, target, "@maxValue")) {
        return;
    }
    if (!validateRange(context, getMinValueAsNumeric(context.program, target) ??
        getMinValueExclusiveAsNumeric(context.program, target), maxValue)) {
        return;
    }
    setMaxValue(program, target, maxValue);
};
// -- @minValueExclusive decorator ---------------------
const $minValueExclusive = (context, target, minValueExclusive) => {
    validateDecoratorUniqueOnNode(context, target, $minValueExclusive);
    validateDecoratorNotOnType(context, target, $minValue, $minValueExclusive);
    const { program } = context;
    if (!validateTargetingANumeric(context, target, "@minValueExclusive")) {
        return;
    }
    if (!validateRange(context, minValueExclusive, getMaxValueAsNumeric(context.program, target) ??
        getMaxValueExclusiveAsNumeric(context.program, target))) {
        return;
    }
    setMinValueExclusive(program, target, minValueExclusive);
};
// -- @maxValueExclusive decorator ---------------------
const $maxValueExclusive = (context, target, maxValueExclusive) => {
    validateDecoratorUniqueOnNode(context, target, $maxValueExclusive);
    validateDecoratorNotOnType(context, target, $maxValue, $maxValueExclusive);
    const { program } = context;
    if (!validateTargetingANumeric(context, target, "@maxValueExclusive")) {
        return;
    }
    if (!validateRange(context, getMinValueAsNumeric(context.program, target) ??
        getMinValueExclusiveAsNumeric(context.program, target), maxValueExclusive)) {
        return;
    }
    setMaxValueExclusive(program, target, maxValueExclusive);
};
// -- @secret decorator ---------------------
const [isSecret, markSecret] = useStateSet(createStateSymbol("secretTypes"));
/**
 * Mark a string as a secret value that should be treated carefully to avoid exposure
 * @param context Decorator context
 * @param target Decorator target, either a string model or a property with type string.
 */
const $secret = (context, target) => {
    validateDecoratorUniqueOnNode(context, target, $secret);
    if (!validateTargetingAString(context, target, "@secret")) {
        return;
    }
    markSecret(context.program, target);
};
const [getEncode, setEncodeData] = useStateMap(createStateSymbol("encode"));
const $encode = (context, target, encoding, encodeAs) => {
    validateDecoratorUniqueOnNode(context, target, $encode);
    const encodeData = computeEncoding(context.program, encoding, encodeAs);
    if (encodeData === undefined) {
        return;
    }
    const targetType = getPropertyType(target);
    validateEncodeData(context, targetType, encodeData);
    setEncodeData(context.program, target, encodeData);
};
function computeEncoding(program, encodingOrEncodeAs, encodeAs) {
    const strType = program.checker.getStdType("string");
    const resolvedEncodeAs = encodeAs ?? strType;
    if (typeof encodingOrEncodeAs === "string") {
        return { encoding: encodingOrEncodeAs, type: resolvedEncodeAs };
    }
    else if (isValue(encodingOrEncodeAs)) {
        const member = encodingOrEncodeAs.value;
        if (member.value && typeof member.value === "string") {
            return { encoding: member.value, type: resolvedEncodeAs };
        }
        else {
            return { encoding: getTypeName(member), type: resolvedEncodeAs };
        }
    }
    else {
        const originalType = encodingOrEncodeAs;
        if (originalType !== strType) {
            reportDiagnostic(program, {
                code: "invalid-encode",
                messageId: "firstArg",
                target: encodingOrEncodeAs,
            });
            return undefined;
        }
        return { type: encodingOrEncodeAs };
    }
}
function validateEncodeData(context, target, encodeData) {
    function check(validTargets, validEncodeTypes) {
        const checker = context.program.checker;
        const isTargetValid = isTypeIn(target, (type) => validTargets.some((validTarget) => {
            return ignoreDiagnostics(checker.isTypeAssignableTo(type, checker.getStdType(validTarget), target));
        }));
        if (!isTargetValid) {
            reportDiagnostic(context.program, {
                code: "invalid-encode",
                messageId: "wrongType",
                format: {
                    encoding: encodeData.encoding ?? "string",
                    type: getTypeName(target),
                    expected: validTargets.join(", "),
                },
                target: context.decoratorTarget,
            });
        }
        const isEncodingTypeValid = validEncodeTypes.some((validEncoding) => {
            return ignoreDiagnostics(checker.isTypeAssignableTo(encodeData.type, checker.getStdType(validEncoding), target));
        });
        if (!isEncodingTypeValid) {
            const typeName = getTypeName(encodeData.type);
            reportDiagnostic(context.program, {
                code: "invalid-encode",
                messageId: ["unixTimestamp", "seconds"].includes(encodeData.encoding ?? "string")
                    ? "wrongNumericEncodingType"
                    : "wrongEncodingType",
                format: {
                    encoding: encodeData.encoding,
                    type: getTypeName(target),
                    expected: validEncodeTypes.join(", "),
                    actual: typeName,
                },
                target: context.decoratorTarget,
            });
        }
    }
    switch (encodeData.encoding) {
        case "rfc3339":
            return check(["utcDateTime", "offsetDateTime"], ["string"]);
        case "rfc7231":
            return check(["utcDateTime", "offsetDateTime"], ["string"]);
        case "unixTimestamp":
            return check(["utcDateTime"], ["integer"]);
        case "seconds":
            return check(["duration"], ["numeric"]);
        case "base64":
            return check(["bytes"], ["string"]);
        case "base64url":
            return check(["bytes"], ["string"]);
        case undefined:
            return check(["numeric"], ["string"]);
    }
}
// -- @withOptionalProperties decorator ---------------------
const $withOptionalProperties = (context, target) => {
    // Make all properties of the target type optional
    target.properties.forEach((p) => (p.optional = true));
};
// -- @withoutOmittedProperties decorator ----------------------
const $withoutOmittedProperties = (context, target, omitProperties) => {
    // Get the property or properties to omit
    const omitNames = new Set();
    if (omitProperties.kind === "String") {
        omitNames.add(omitProperties.value);
    }
    else if (omitProperties.kind === "Union") {
        for (const variant of omitProperties.variants.values()) {
            if (variant.type.kind === "String") {
                omitNames.add(variant.type.value);
            }
        }
    }
    // Remove all properties to be omitted
    filterModelPropertiesInPlace(target, (prop) => !omitNames.has(prop.name));
};
// -- @withPickedProperties decorator ----------------------
const $withPickedProperties = (context, target, pickedProperties) => {
    // Get the property or properties to pick
    const pickedNames = new Set();
    if (pickedProperties.kind === "String") {
        pickedNames.add(pickedProperties.value);
    }
    else if (pickedProperties.kind === "Union") {
        for (const variant of pickedProperties.variants.values()) {
            if (variant.type.kind === "String") {
                pickedNames.add(variant.type.value);
            }
        }
    }
    // Remove all properties not picked
    filterModelPropertiesInPlace(target, (prop) => pickedNames.has(prop.name));
};
// -- @withoutDefaultValues decorator ----------------------
const $withoutDefaultValues = (context, target) => {
    // remove all read-only properties from the target type
    target.properties.forEach((p) => {
        delete p.defaultValue;
    });
};
// -- @tag decorator ---------------------
const [getTagsState, setTags] = useStateMap(createStateSymbol("tagProperties"));
// Set a tag on an operation, interface, or namespace.  There can be multiple tags on an
// operation, interface, or namespace.
const $tag = (context, target, tag) => {
    const tags = getTagsState(context.program, target);
    if (tags) {
        tags.push(tag);
    }
    else {
        setTags(context.program, target, [tag]);
    }
};
// Return the tags set on an operation or namespace
function getTags(program, target) {
    return getTagsState(program, target) || [];
}
// Merge the tags for a operation with the tags that are on the namespace or
// interface it resides within.
function getAllTags(program, target) {
    const tags = new Set();
    let current = target;
    while (current !== undefined) {
        for (const t of getTags(program, current)) {
            tags.add(t);
        }
        // Move up to the parent
        if (current.kind === "Operation") {
            current = current.interface ?? current.namespace;
        }
        else {
            // Type is a namespace or interface
            current = current.namespace;
        }
    }
    return tags.size > 0 ? Array.from(tags).reverse() : undefined;
}
// -- @friendlyName decorator ---------------------
const [getFriendlyName, setFriendlyName] = useStateMap(createStateSymbol("friendlyNames"));
const $friendlyName = (context, target, friendlyName, sourceObject) => {
    // workaround for current lack of functionality in compiler
    // https://github.com/microsoft/typespec/issues/2717
    if (target.kind === "Model" || target.kind === "Operation") {
        if (context.decoratorTarget.kind === SyntaxKind.AugmentDecoratorStatement) {
            if (ignoreDiagnostics(context.program.checker.resolveTypeReference(context.decoratorTarget.targetType))?.node !== target.node) {
                return;
            }
        }
        if (context.decoratorTarget.kind === SyntaxKind.DecoratorExpression) {
            if (context.decoratorTarget.parent !== target.node) {
                return;
            }
        }
    }
    // If an object was passed in, use it to format the friendly name
    if (sourceObject) {
        friendlyName = replaceTemplatedStringFromProperties(friendlyName, sourceObject);
    }
    setFriendlyName(context.program, target, friendlyName);
};
/**
 * `@key` - mark a model property as the key to identify instances of that type
 *
 * The optional first argument accepts an alternate key name which may be used by emitters.
 * Otherwise, the name of the target property will be used.
 *
 * `@key` can only be applied to model properties.
 */
const $key = (context, entity, altName) => {
    // Ensure that the key property is not marked as optional
    if (entity.optional) {
        reportDiagnostic(context.program, {
            code: "no-optional-key",
            format: { propertyName: entity.name },
            target: entity,
        });
        return;
    }
    // Register the key property
    setKey(context.program, entity, altName || entity.name);
};
/**
 * Return the deprecated message or undefined if not deprecated
 * @param program Program
 * @param type Type
 */
function getDeprecated(program, type) {
    return getDeprecationDetails(program, type)?.message;
}
const [getOverloads, setOverloads] = useStateMap(createStateSymbol("overloadedByKey"));
const [getOverloadedOperation, setOverloadBase] = useStateMap(createStateSymbol("overloadsOperation"));
/**
 * `@overload` - Indicate that the target overloads (specializes) the overloads type.
 * @param context DecoratorContext
 * @param target The specializing operation declaration
 * @param overloadBase The operation to be overloaded.
 */
const $overload = (context, target, overloadBase) => {
    // Ensure that the overloaded method arguments are a subtype of the original operation.
    const [paramValid, paramDiagnostics] = context.program.checker.isTypeAssignableTo(target.parameters, overloadBase.parameters, target);
    if (!paramValid)
        context.program.reportDiagnostics(paramDiagnostics);
    const [returnTypeValid, returnTypeDiagnostics] = context.program.checker.isTypeAssignableTo(target.returnType, overloadBase.returnType, target);
    if (!returnTypeValid)
        context.program.reportDiagnostics(returnTypeDiagnostics);
    if (!areOperationsInSameContainer(target, overloadBase)) {
        reportDiagnostic(context.program, {
            code: "overload-same-parent",
            target: context.decoratorTarget,
        });
    }
    // Save the information about the overloaded operation
    setOverloadBase(context.program, target, overloadBase);
    const existingOverloads = getOverloads(context.program, overloadBase) || new Array();
    setOverloads(context.program, overloadBase, existingOverloads.concat(target));
};
function areOperationsInSameContainer(op1, op2) {
    return op1.interface || op2.interface
        ? op1.interface === op2.interface
        : op1.namespace === op2.namespace;
}
function validateRange(context, min, max) {
    if (min === undefined || max === undefined) {
        return true;
    }
    if (min.gt(max)) {
        reportDiagnostic(context.program, {
            code: "invalid-range",
            format: { start: min.toString(), end: max.toString() },
            target: context.decoratorTarget,
        });
        return false;
    }
    return true;
}
const discriminatedDecorator = (context, entity, options = {}) => {
    setDiscriminatedOptions(context.program, entity, {
        envelope: "object",
        discriminatorPropertyName: "kind",
        envelopePropertyName: "value",
        ...options,
    });
    const [_, diagnostics] = getDiscriminatedUnion(context.program, entity);
    context.program.reportDiagnostics(diagnostics);
};
const $discriminator = (context, entity, propertyName) => {
    setDiscriminator(context.program, entity, { propertyName });
};
const [getExamplesState, setExamples] = useStateMap(createStateSymbol("examples"));
const $example = (context, target, _example, options) => {
    const decorator = target.decorators.find((d) => d.decorator === $example && d.node === context.decoratorTarget);
    compilerAssert(decorator, `Couldn't find @example decorator`, context.decoratorTarget);
    const rawExample = decorator.args[0].value;
    // skip validation in cloned types
    if (Realm.realmForType.get(target) === undefined) {
        if (!checkExampleValid(context.program, rawExample, target.kind === "ModelProperty" ? target.type : target, context.getArgumentTarget(0))) {
            return;
        }
    }
    let list = getExamplesState(context.program, target);
    if (list === undefined) {
        list = [];
        setExamples(context.program, target, list);
    }
    list.push({ value: rawExample, ...options });
};
function getExamples(program, target) {
    return getExamplesState(program, target) ?? [];
}
const [getOpExamplesState, setOpExamples] = useStateMap(createStateSymbol("opExamples"));
const $opExample = (context, target, _example, options) => {
    const decorator = target.decorators.find((d) => d.decorator === $opExample && d.node === context.decoratorTarget);
    compilerAssert(decorator, `Couldn't find @opExample decorator`, context.decoratorTarget);
    const rawExampleConfig = decorator.args[0].value;
    const parameters = rawExampleConfig.properties.get("parameters")?.value;
    const returnType = rawExampleConfig.properties.get("returnType")?.value;
    // skip validation in cloned types
    if (Realm.realmForType.get(target) === undefined) {
        if (parameters &&
            !checkExampleValid(context.program, parameters, target.parameters, context.getArgumentTarget(0))) {
            return;
        }
        if (returnType &&
            !checkExampleValid(context.program, returnType, target.returnType, context.getArgumentTarget(0))) {
            return;
        }
    }
    let list = getOpExamplesState(context.program, target);
    if (list === undefined) {
        list = [];
        setOpExamples(context.program, target, list);
    }
    list.push({ parameters, returnType, ...options });
};
function checkExampleValid(program, value, target, diagnosticTarget) {
    const exactType = program.checker.getValueExactType(value);
    const [assignable, diagnostics] = program.checker.isTypeAssignableTo(exactType ?? value.type, target, diagnosticTarget);
    if (!assignable) {
        program.reportDiagnostics(diagnostics);
    }
    return assignable;
}
function getOpExamples(program, target) {
    return getOpExamplesState(program, target) ?? [];
}

defineKit({
    modelProperty: {
        is(type) {
            return type.kind === "ModelProperty";
        },
        getEncoding(type) {
            return getEncode(this.program, type) ?? getEncode(this.program, type.type);
        },
        getFormat(type) {
            return getFormat(this.program, type) ?? getFormat(this.program, type.type);
        },
        getVisibilityForClass(property, visibilityClass) {
            return getVisibilityForClass(this.program, property, visibilityClass);
        },
        create(desc) {
            return this.program.checker.createType({
                kind: "ModelProperty",
                name: desc.name,
                node: undefined,
                type: desc.type,
                optional: desc.optional ?? false,
                decorators: [],
                defaultValue: desc.defaultValue,
            });
        },
    },
});

const spreadCache = new Map();
defineKit({
    model: {
        create(desc) {
            const properties = createRekeyableMap(Array.from(Object.entries(desc.properties)));
            const model = this.program.checker.createType({
                kind: "Model",
                name: desc.name ?? "",
                decorators: decoratorApplication(this, desc.decorators),
                properties: properties,
                node: undefined,
                derivedModels: desc.derivedModels ?? [],
                sourceModels: desc.sourceModels ?? [],
                indexer: desc.indexer,
            });
            this.program.checker.finishType(model);
            return model;
        },
        is(type) {
            return type.kind === "Model";
        },
        isExpresion(type) {
            return type.name === "";
        },
        getEffectiveModel(model, filter) {
            return getEffectiveModelType(this.program, model, filter);
        },
        getSpreadType(model) {
            if (spreadCache.has(model)) {
                return spreadCache.get(model);
            }
            if (!model.indexer) {
                return undefined;
            }
            if (model.indexer.key.name === "string") {
                const record = this.record.create(model.indexer.value);
                spreadCache.set(model, record);
                return record;
            }
            if (model.indexer.key.name === "integer") {
                const array = this.array.create(model.indexer.value);
                spreadCache.set(model, array);
                return array;
            }
            return model.indexer.value;
        },
        getProperties(model, options = {}) {
            // Add explicitly defined properties
            const properties = copyMap(model.properties);
            // Add discriminator property if it exists
            const discriminator = this.type.getDiscriminator(model);
            if (discriminator) {
                const discriminatorName = discriminator.propertyName;
                properties.set(discriminatorName, this.modelProperty.create({ name: discriminatorName, type: this.builtin.string }));
            }
            if (options.includeExtended) {
                let base = model.baseModel;
                while (base) {
                    for (const [key, value] of base.properties) {
                        if (!properties.has(key)) {
                            properties.set(key, value);
                        }
                    }
                    base = base.baseModel;
                }
            }
            // TODO: Add Spread?
            return properties;
        },
        getAdditionalPropertiesRecord(model) {
            // model MyModel is Record<> {} should be model with additional properties
            if (this.model.is(model) && model.sourceModel && this.record.is(model.sourceModel)) {
                return model.sourceModel;
            }
            // model MyModel extends Record<> {} should be model with additional properties
            if (model.baseModel && this.record.is(model.baseModel)) {
                return model.baseModel;
            }
            // model MyModel { ...Record<>} should be model with additional properties
            const spread = this.model.getSpreadType(model);
            if (spread && this.model.is(spread) && this.record.is(spread)) {
                return spread;
            }
            if (model.baseModel) {
                return this.model.getAdditionalPropertiesRecord(model.baseModel);
            }
            return undefined;
        },
        getDiscriminatedUnion(model) {
            const discriminator = getDiscriminator(this.program, model);
            if (!discriminator) {
                return undefined;
            }
            return ignoreDiagnostics(getDiscriminatedUnionFromInheritance(model, discriminator));
        },
    },
});

defineKit({
    operation: {
        is(type) {
            return type.kind === "Operation";
        },
        create(desc) {
            const parametersModel = this.model.create({
                name: `${desc.name}Parameters`,
                properties: desc.parameters.reduce((acc, property) => {
                    acc[property.name] = property;
                    return acc;
                }, {}),
            });
            const operation = this.program.checker.createType({
                kind: "Operation",
                name: desc.name,
                decorators: [],
                parameters: parametersModel,
                returnType: desc.returnType,
                node: undefined,
            });
            this.program.checker.finishType(operation);
            return operation;
        },
    },
});

defineKit({
    record: {
        is(type) {
            return (type.kind === "Model" && isRecordModelType(this.program, type) && type.properties.size === 0);
        },
        getElementType(type) {
            if (!this.record.is(type)) {
                throw new Error("Type is not a record.");
            }
            return type.indexer.value;
        },
        create(elementType) {
            return this.model.create({
                name: "Record",
                properties: {},
                indexer: {
                    key: this.builtin.string,
                    value: elementType,
                },
            });
        },
    },
});

defineKit({
    scalar: {
        is(type) {
            return type.kind === "Scalar";
        },
        extendsBoolean: extendsStdType("boolean"),
        extendsBytes: extendsStdType("bytes"),
        extendsDecimal: extendsStdType("decimal"),
        extendsDecimal128: extendsStdType("decimal128"),
        extendsDuration: extendsStdType("duration"),
        extendsFloat: extendsStdType("float"),
        extendsFloat32: extendsStdType("float32"),
        extendsFloat64: extendsStdType("float64"),
        extendsInt8: extendsStdType("int8"),
        extendsInt16: extendsStdType("int16"),
        extendsInt32: extendsStdType("int32"),
        extendsInt64: extendsStdType("int64"),
        extendsInteger: extendsStdType("integer"),
        extendsNumeric: extendsStdType("numeric"),
        extendsOffsetDateTime: extendsStdType("offsetDateTime"),
        extendsPlainDate: extendsStdType("plainDate"),
        extendsPlainTime: extendsStdType("plainTime"),
        extendsSafeint: extendsStdType("safeint"),
        extendsString: extendsStdType("string"),
        extendsUint8: extendsStdType("uint8"),
        extendsUint16: extendsStdType("uint16"),
        extendsUint32: extendsStdType("uint32"),
        extendsUint64: extendsStdType("uint64"),
        extendsUrl: extendsStdType("url"),
        extendsUtcDateTime: extendsStdType("utcDateTime"),
        isBoolean: isStdType("boolean"),
        isBytes: isStdType("bytes"),
        isDecimal: isStdType("decimal"),
        isDecimal128: isStdType("decimal128"),
        isDuration: isStdType("duration"),
        isFloat: isStdType("float"),
        isFloat32: isStdType("float32"),
        isFloat64: isStdType("float64"),
        isInt8: isStdType("int8"),
        isInt16: isStdType("int16"),
        isInt32: isStdType("int32"),
        isInt64: isStdType("int64"),
        isInteger: isStdType("integer"),
        isNumeric: isStdType("numeric"),
        isOffsetDateTime: isStdType("offsetDateTime"),
        isPlainDate: isStdType("plainDate"),
        isPlainTime: isStdType("plainTime"),
        isSafeint: isStdType("safeint"),
        isString: isStdType("string"),
        isUint8: isStdType("uint8"),
        isUint16: isStdType("uint16"),
        isUint32: isStdType("uint32"),
        isUint64: isStdType("uint64"),
        isUrl: isStdType("url"),
        isUtcDateTime: isStdType("utcDateTime"),
        getStdBase(type) {
            const tspNamespace = this.program.resolveTypeReference("TypeSpec")[0];
            let current = type;
            while (current) {
                if (current.namespace === tspNamespace) {
                    return current;
                }
                current = current.baseScalar;
            }
            return null;
        },
        getEncoding(type) {
            return getEncode(this.program, type);
        },
        getFormat(type) {
            return getFormat(this.program, type);
        },
    },
});
function isStdType(typeName) {
    return function (type) {
        return type === this.program.checker.getStdType(typeName);
    };
}
function extendsStdType(typeName) {
    return function (type) {
        if (!this.scalar.is(type)) {
            return false;
        }
        return isIntrinsicType(this.program, type, typeName);
    };
}

/**
 * Get a plausible name for the given type.
 * @experimental
 */
function getPlausibleName(type) {
    let name = type.name;
    if (!name) {
        name = "TypeExpression"; // TODO: Implement automatic name generation based on the type context
    }
    if (isTemplateInstance(type)) {
        const namePrefix = type.templateMapper.args.map((a) => ("name" in a && a.name) || "").join("_");
        name = `${namePrefix}${name}`;
    }
    return name;
}

defineKit({
    type: {
        finishType(type) {
            this.program.checker.finishType(type);
        },
        clone(type) {
            let clone;
            switch (type.kind) {
                case "Model":
                    clone = this.program.checker.createType({
                        ...type,
                        decorators: [...type.decorators],
                        derivedModels: [...type.derivedModels],
                        sourceModels: type.sourceModels.map((x) => ({ ...x })),
                        properties: copyMap(type.properties),
                        indexer: type.indexer ? { ...type.indexer } : undefined,
                    });
                    break;
                case "Union":
                    clone = this.program.checker.createType({
                        ...type,
                        decorators: [...type.decorators],
                        variants: copyMap(type.variants),
                        get options() {
                            return Array.from(this.variants.values()).map((v) => v.type);
                        },
                    });
                    break;
                case "Interface":
                    clone = this.program.checker.createType({
                        ...type,
                        decorators: [...type.decorators],
                        operations: copyMap(type.operations),
                    });
                    break;
                case "Enum":
                    clone = this.program.checker.createType({
                        ...type,
                        decorators: [...type.decorators],
                        members: copyMap(type.members),
                    });
                    break;
                case "Namespace":
                    clone = this.program.checker.createType({
                        ...type,
                        decorators: [...type.decorators],
                        instantiationParameters: type.instantiationParameters
                            ? [...type.instantiationParameters]
                            : undefined,
                        models: copyMap(type.models),
                        decoratorDeclarations: copyMap(type.decoratorDeclarations),
                        enums: copyMap(type.enums),
                        unions: copyMap(type.unions),
                        operations: copyMap(type.operations),
                        interfaces: copyMap(type.interfaces),
                        namespaces: copyMap(type.namespaces),
                        scalars: copyMap(type.scalars),
                    });
                    break;
                case "Scalar":
                    clone = this.program.checker.createType({
                        ...type,
                        decorators: [...type.decorators],
                        derivedScalars: [...type.derivedScalars],
                        constructors: copyMap(type.constructors),
                    });
                    break;
                default:
                    clone = this.program.checker.createType({
                        ...type,
                        ...("decorators" in type ? { decorators: [...type.decorators] } : {}),
                    });
                    break;
            }
            this.realm.addType(clone);
            return clone;
        },
        isError(type) {
            return isErrorType(type);
        },
        getEncodedName(type, encoding) {
            return resolveEncodedName(this.program, type, encoding);
        },
        getSummary(type) {
            return getSummary(this.program, type);
        },
        getDoc(type) {
            return getDoc(this.program, type);
        },
        getPlausibleName(type) {
            return getPlausibleName(type);
        },
        getDiscriminator(type) {
            let discriminator;
            if ($.model.is(type)) {
                discriminator = getDiscriminator(this.program, type);
            }
            else {
                const unionDiscriminator = ignoreDiagnostics(getDiscriminatedUnion(this.program, type));
                const propertyName = unionDiscriminator?.options.discriminatorPropertyName;
                if (propertyName) {
                    discriminator = { propertyName };
                }
            }
            return discriminator;
        },
        maxValue(type) {
            return getMaxValue(this.program, type);
        },
        minValue(type) {
            return getMinValue(this.program, type);
        },
        maxLength(type) {
            return getMaxLength(this.program, type);
        },
        minLength(type) {
            return getMinLength(this.program, type);
        },
        maxItems(type) {
            return getMaxItems(this.program, type);
        },
        maxValueExclusive(type) {
            return getMaxValueExclusive(this.program, type);
        },
        minValueExclusive(type) {
            return getMinValueExclusive(this.program, type);
        },
        minItems(type) {
            return getMinItems(this.program, type);
        },
        isNever(type) {
            return isNeverType(type);
        },
        isUserDefined(type) {
            return getLocationContext(this.program, type).type === "project";
        },
    },
});

defineKit({
    unionVariant: {
        create(desc) {
            const variant = this.program.checker.createType({
                kind: "UnionVariant",
                name: desc.name ?? Symbol("name"),
                decorators: decoratorApplication(this, desc.decorators),
                type: desc.type,
                node: undefined,
                union: desc.union,
            });
            this.program.checker.finishType(variant);
            return variant;
        },
        is(type) {
            return type.kind === "UnionVariant";
        },
    },
});

const UnionKit = defineKit({
    union: {
        filter(union, filterFn) {
            const variants = Array.from(union.variants.values()).filter(filterFn);
            return this.union.create({ variants });
        },
        create(desc) {
            const union = this.program.checker.createType({
                kind: "Union",
                name: desc.name,
                decorators: decoratorApplication(this, desc.decorators),
                variants: createRekeyableMap(),
                get options() {
                    return Array.from(this.variants.values()).map((v) => v.type);
                },
                expression: desc.name === undefined,
                node: undefined,
            });
            if (Array.isArray(desc.variants)) {
                for (const variant of desc.variants) {
                    union.variants.set(variant.name, variant);
                    variant.union = union;
                }
            }
            else if (desc.variants) {
                for (const [name, value] of Object.entries(desc.variants)) {
                    union.variants.set(name, this.unionVariant.create({ name, type: this.literal.create(value) }));
                }
            }
            this.program.checker.finishType(union);
            return union;
        },
        is(type) {
            return type.kind === "Union";
        },
        isValidEnum(type) {
            for (const variant of type.variants.values()) {
                if (!this.literal.isString(variant.type) && !this.literal.isNumeric(variant.type)) {
                    return false;
                }
            }
            return true;
        },
        isExtensible(type) {
            const variants = Array.from(type.variants.values());
            if (variants.length === 0) {
                return false;
            }
            for (let i = 0; i < variants.length; i++) {
                let isCommon = true;
                for (let j = 0; j < variants.length; j++) {
                    if (i === j) {
                        continue;
                    }
                    const assignable = ignoreDiagnostics(this.program.checker.isTypeAssignableTo(variants[j].type, variants[i].type, type));
                    if (!assignable) {
                        isCommon = false;
                        break;
                    }
                }
                if (isCommon) {
                    return true;
                }
            }
            return false;
        },
        isExpression(type) {
            return type.name === undefined || type.name === "";
        },
        getDiscriminatedUnion(type) {
            return ignoreDiagnostics(getDiscriminatedUnion(this.program, type));
        },
    },
});

defineKit({
    value: {
        is(value) {
            const type = value;
            return (this.value.isString(type) ||
                this.value.isNumeric(type) ||
                this.value.isBoolean(type) ||
                this.value.isArray(type) ||
                this.value.isObject(type) ||
                this.value.isEnum(type) ||
                this.value.isNull(type) ||
                this.value.isScalar(type));
        },
        create(value) {
            if (typeof value === "string") {
                return this.value.createString(value);
            }
            else if (typeof value === "number") {
                return this.value.createNumeric(value);
            }
            else {
                return this.value.createBoolean(value);
            }
        },
        createString(value) {
            return {
                entityKind: "Value",
                value: value,
                valueKind: "StringValue",
                type: this.literal.createString(value),
                scalar: undefined,
            };
        },
        createNumeric(value) {
            const valueAsString = String(value);
            return {
                entityKind: "Value",
                value: Numeric(valueAsString),
                valueKind: "NumericValue",
                type: this.literal.createNumeric(value),
                scalar: undefined,
            };
        },
        createBoolean(value) {
            return {
                entityKind: "Value",
                value: value,
                valueKind: "BooleanValue",
                type: this.literal.createBoolean(value),
                scalar: undefined,
            };
        },
        isBoolean(type) {
            return type.valueKind === "BooleanValue";
        },
        isString(type) {
            return type.valueKind === "StringValue";
        },
        isNumeric(type) {
            return type.valueKind === "NumericValue";
        },
        isArray(type) {
            return type.valueKind === "ArrayValue";
        },
        isObject(type) {
            return type.valueKind === "ObjectValue";
        },
        isEnum(type) {
            return type.valueKind === "EnumValue";
        },
        isNull(type) {
            return type.valueKind === "NullValue";
        },
        isScalar(type) {
            return type.valueKind === "ScalarValue";
        },
    },
});

/**
 * Create a new Typekit that operates in the given realm.
 *
 * Ordinarily, you should use the default typekit `$` to manipulate types in the current program, or call `$` with a
 * Realm or Program as the first argument if you want to work in a specific realm or in the default typekit realm of
 * a specific program.
 *
 * @param realm - The realm to create the typekit in.
 *
 * @experimental
 */
function createTypekit(realm) {
    const tk = Object.create(TypekitPrototype);
    const handler = {
        get(target, prop, receiver) {
            if (prop === "program") {
                // don't wrap program (probably need to ensure this isn't a nested program somewhere)
                return realm.program;
            }
            if (prop === "realm") {
                return realm;
            }
            const value = Reflect.get(target, prop, receiver);
            // Wrap functions to set `this` correctly
            if (typeof value === "function") {
                return function (...args) {
                    return value.apply(proxy, args);
                };
            }
            // Only wrap objects marked as Typekit namespaces
            if (typeof value === "object" && value !== null && isTypekitNamespace(value)) {
                return new Proxy(value, handler); // Wrap namespace objects
            }
            return value;
        },
    };
    const proxy = new Proxy(tk, handler);
    return proxy;
}
// Helper function to check if an object is a Typekit namespace
function isTypekitNamespace(obj) {
    return obj && !!obj[TypekitNamespaceSymbol];
}
// #region Default Typekit
const CURRENT_PROGRAM = Symbol.for("TypeSpec.Typekit.CURRENT_PROGRAM");
const DEFAULT_REALM = Symbol.for("TypeSpec.Typekit.DEFAULT_TYPEKIT_REALM");
function getCurrentProgram() {
    return globalThis[CURRENT_PROGRAM];
}
/**
 * Sets a given program as the current program for Typekit operations.
 *
 * This is necessary to enable use of the default Typekit `$` and should be
 * called whenever the compiler is working with a specific program
 *
 * @param program - The program to set as the current program.
 *
 * @internal
 */
function setCurrentProgram(program) {
    globalThis[CURRENT_PROGRAM] = program;
}
function _$(arg) {
    let realm;
    if (Object.hasOwn(arg, "projectRoot")) {
        // arg is a Program
        realm = arg[DEFAULT_REALM] ??= new Realm(arg, "default typekit realm");
    }
    else {
        // arg is a Realm
        realm = arg;
    }
    return realm.typekit;
}
/**
 * Typekit - Utilities for working with TypeSpec types.
 *
 * The default typekit `$` can be used to manipulate types in the current program.
 *
 * Each typekit is associated with a Realm in which it operates. The default typekit
 * will use the default typekit realm for the current program.
 *
 * Alternatively, to work in a specific realm, you can get the typekit associated
 * with that realm by calling `$` with the realm as an argument, or by calling
 * `$` with a program as an argument (in this case, it will use that program's
 * default typekit realm or create one if it does not already exist).
 *
 * @example
 * ```ts
 * import { $ } from "@typespec/compiler/experimental";
 *
 * const clone = $.type.clone(inputType);
 * ```
 *
 * @example
 * ```ts
 * import { $, Realm } from "@typespec/compiler/experimental";
 *
 * const realm = new Realm(program, "my custom realm");
 *
 * const clone = $(realm).type.clone(inputType);
 * ```
 *
 * @example
 * ```ts
 * import { $ } from "@typespec/compiler/experimental";
 *
 * const projectedProgram = projectProgram(program, ...);
 *
 * const clone = $(projectedProgram).type.clone(inputType);
 * ```
 *
 * @see {@link Realm}
 *
 * @experimental
 */
const $ = new Proxy(_$, {
    get(_target, prop, _receiver) {
        const currentProgram = getCurrentProgram();
        compilerAssert(currentProgram !== undefined, "Default typekits may not be used until a program is set in the compiler.");
        if (prop === "program")
            return currentProgram;
        const realm = (currentProgram[DEFAULT_REALM] ??= new Realm(currentProgram, "default typekit realm"));
        if (prop === "realm")
            return realm;
        const tk = _$(realm);
        return Reflect.get(tk, prop, tk);
    },
});
// #endregion

var _a;
/**
 * A Realm's view of a Program's state map for a given state key.
 *
 * For all operations, if a type was created within the realm, the realm's own state map is used. Otherwise, the owning'
 * Program's state map is used.
 *
 * @experimental
 */
class StateMapRealmView {
    #realm;
    #parentState;
    #realmState;
    constructor(realm, realmState, parentState) {
        this.#realm = realm;
        this.#parentState = parentState;
        this.#realmState = realmState;
    }
    has(t) {
        return this.#select(t).has(t) ?? false;
    }
    set(t, v) {
        this.#select(t).set(t, v);
        return this;
    }
    get(t) {
        return this.#select(t).get(t);
    }
    delete(t) {
        return this.#select(t).delete(t);
    }
    forEach(cb, thisArg) {
        for (const item of this.entries()) {
            cb.call(thisArg, item[1], item[0], this);
        }
        return this;
    }
    get size() {
        return this.#realmState.size + this.#parentState.size;
    }
    clear() {
        this.#realmState.clear();
    }
    *entries() {
        for (const item of this.#realmState) {
            yield item;
        }
        for (const item of this.#parentState) {
            yield item;
        }
        return undefined;
    }
    *values() {
        for (const item of this.entries()) {
            yield item[1];
        }
        return undefined;
    }
    *keys() {
        for (const item of this.entries()) {
            yield item[0];
        }
        return undefined;
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    [Symbol.toStringTag] = "StateMap";
    #select(keyType) {
        if (this.#realm.hasType(keyType)) {
            return this.#realmState;
        }
        return this.#parentState;
    }
}
/**
 * A Realm is an alternate view of a Program where types can be cloned, deleted, and modified without affecting the
 * original types in the Program.
 *
 * The realm stores the types that exist within the realm, views of state maps that only apply within the realm,
 * and a view of types that have been removed from the realm's view.
 *
 * @experimental
 */
class Realm {
    #program;
    /**
     * Stores all types owned by this realm.
     */
    #types = new Set();
    /**
     * Stores types that are deleted in this realm. When a realm is active and doing a traversal, you will
     * not find this type in e.g. collections. Deleted types are mapped to `null` if you ask for it.
     */
    #deletedTypes = new WeakSet();
    #stateMaps = new Map();
    key;
    /**
     * Create a new realm in the given program.
     *
     * @param program - The program to create the realm in.
     * @param description - A short description of the realm's purpose.
     */
    constructor(program, description) {
        this.key = Symbol(description);
        this.#program = program;
    }
    #_typekit;
    /**
     * The typekit instance bound to this realm.
     *
     * If the realm does not already have a typekit associated with it, one will be created and bound to this realm.
     */
    get typekit() {
        return (this.#_typekit ??= createTypekit(this));
    }
    /**
     * The program that this realm is associated with.
     */
    get program() {
        return this.#program;
    }
    /**
     * Gets a state map for the given state key symbol.
     *
     * This state map is a view of the program's state map for the given state key, with modifications made to the realm's
     * own state.
     *
     * @param stateKey - The symbol to use as the state key.
     * @returns The realm's state map for the given state key.
     */
    stateMap(stateKey) {
        let m = this.#stateMaps.get(stateKey);
        if (!m) {
            m = new Map();
            this.#stateMaps.set(stateKey, m);
        }
        return new StateMapRealmView(this, m, this.#program.stateMap(stateKey));
    }
    /**
     * Clones a type and adds it to the realm. This operation will use the realm's typekit to clone the type.
     *
     * @param type - The type to clone.
     * @returns A clone of the input type that exists within this realm.
     */
    clone(type) {
        compilerAssert(type, "Undefined type passed to clone");
        const clone = this.#cloneIntoRealm(type);
        this.typekit.type.finishType(clone);
        return clone;
    }
    /**
     * Removes a type from this realm. This operation will not affect the type in the program, only this realm's view
     * of the type.
     *
     * @param type - The TypeSpec type to remove from this realm.
     */
    remove(type) {
        this.#deletedTypes.add(type);
    }
    /**
     * Determines whether or not this realm contains a given type.
     *
     * @param type - The type to check.
     * @returns true if the type was created within this realm or added to this realm, false otherwise.
     */
    hasType(type) {
        return this.#types.has(type);
    }
    /**
     * Adds a type to this realm. Once a type is added to the realm, the realm considers it part of itself.
     *
     * A type can be present in multiple realms, but `Realm.realmForType` will only return the last realm that the type
     * was added to.
     *
     * @param type - The type to add to this realm.
     */
    addType(type) {
        this.#types.add(type);
        _a.realmForType.set(type, this);
    }
    #cloneIntoRealm(type) {
        const clone = this.typekit.type.clone(type);
        this.#types.add(clone);
        _a.realmForType.set(clone, this);
        return clone;
    }
    // TODO better way?
    /** @internal */
    get types() {
        return this.#types;
    }
    static realmForType = new WeakMap();
}
_a = Realm;

const indexTypeKey = Symbol.for(`TypeSpec.index`);
const indexerDecorator = (context, target, key, value) => {
    const indexer = { key, value };
    context.program.stateMap(indexTypeKey).set(target, indexer);
};
function getIndexer(program, target) {
    return program.stateMap(indexTypeKey).get(target);
}
/**
 * @internal to be used to set the `@doc` from doc comment.
 */
const docFromCommentDecorator = (context, target, key, text) => {
    setDocData(context.program, target, key, { value: text, source: "comment" });
};
const prototypeGetterKey = Symbol.for(`TypeSpec.Prototypes.getter`);
/** @internal */
function getterDecorator(context, target) {
    context.program.stateMap(prototypeGetterKey).set(target, true);
}

// Use a regular expression to define the prefix for TypeSpec-exposed functions
// defined in JavaScript modules
const DecoratorFunctionPattern = /^\$/;
const SymbolTable = class extends Map {
    duplicates = new Map();
    constructor(source) {
        super();
        if (source) {
            this.include(source);
        }
    }
    /** {@inheritdoc MutableSymboleTable} */
    include(source, parentSym) {
        for (const [key, value] of source) {
            super.set(key, { ...value, parent: parentSym ?? value.parent });
        }
        for (const [key, value] of source.duplicates) {
            this.duplicates.set(key, new Set(value));
        }
    }
    // First set for a given key wins, but record all duplicates for diagnostics.
    set(key, value) {
        const existing = super.get(key);
        if (existing === undefined) {
            super.set(key, value);
        }
        else {
            if (existing.flags & 8192 /* SymbolFlags.Using */) {
                mutate(existing).flags |= 16384 /* SymbolFlags.DuplicateUsing */;
            }
            const duplicateArray = this.duplicates.get(existing);
            if (duplicateArray) {
                duplicateArray.add(value);
            }
            else {
                this.duplicates.set(existing, new Set([existing, value]));
            }
        }
        return this;
    }
};
function createSymbolTable(source) {
    return new SymbolTable(source);
}
function createBinder(program) {
    let currentFile;
    let parentNode;
    let fileNamespace;
    let scope;
    return {
        bindSourceFile,
        bindJsSourceFile,
        bindNode,
    };
    function isFunctionName(name) {
        return DecoratorFunctionPattern.test(name);
    }
    function getFunctionName(name) {
        return name.replace(DecoratorFunctionPattern, "");
    }
    function bindJsSourceFile(sourceFile) {
        // cast because it causes TS to make the type of .symbol never other
        if (sourceFile.symbol !== undefined) {
            return;
        }
        fileNamespace = undefined;
        mutate(sourceFile).symbol = createSymbol(sourceFile, sourceFile.file.path, 32768 /* SymbolFlags.SourceFile */ | 1048576 /* SymbolFlags.Declaration */);
        const rootNs = sourceFile.esmExports["namespace"];
        for (const [key, member] of Object.entries(sourceFile.esmExports)) {
            let name;
            let kind;
            if (key === "$flags") {
                const context = getLocationContext(program, sourceFile);
                if (context.type === "library" || context.type === "project") {
                    mutate(context).flags = member;
                }
            }
            else if (key === "$decorators") {
                const value = member;
                for (const [namespaceName, decorators] of Object.entries(value)) {
                    for (const [decoratorName, decorator] of Object.entries(decorators)) {
                        bindFunctionImplementation(namespaceName === "" ? [] : namespaceName.split("."), "decorator", decoratorName, decorator, sourceFile);
                    }
                }
            }
            else if (typeof member === "function") {
                // lots of 'any' casts here because control flow narrowing `member` to Function
                // isn't particularly useful it turns out.
                if (isFunctionName(key)) {
                    name = getFunctionName(key);
                    kind = "decorator";
                    if (name === "onValidate") {
                        const context = getLocationContext(program, sourceFile);
                        const metadata = context.type === "library"
                            ? context.metadata
                            : { type: "file" };
                        program.onValidate(member, metadata);
                        continue;
                    }
                    else if (name === "onEmit") {
                        // nothing to do here this is loaded as emitter.
                        continue;
                    }
                }
                else {
                    name = key;
                    kind = "function";
                }
                const nsParts = resolveJSMemberNamespaceParts(rootNs, member);
                bindFunctionImplementation(nsParts, kind, name, member, sourceFile);
            }
        }
    }
    function bindFunctionImplementation(nsParts, kind, name, fn, sourceFile) {
        let containerSymbol = sourceFile.symbol;
        const tracer = program.tracer.sub("bind.js");
        for (const part of nsParts) {
            const existingBinding = containerSymbol.exports.get(part);
            const jsNamespaceNode = {
                kind: SyntaxKind.JsNamespaceDeclaration,
                id: {
                    kind: SyntaxKind.Identifier,
                    sv: part,
                    pos: 0,
                    end: 0,
                    flags: 0 /* NodeFlags.None */,
                    symbol: undefined,
                },
                pos: sourceFile.pos,
                end: sourceFile.end,
                parent: sourceFile,
                flags: 0 /* NodeFlags.None */,
                symbol: undefined,
            };
            const sym = createSymbol(jsNamespaceNode, part, 256 /* SymbolFlags.Namespace */ | 1048576 /* SymbolFlags.Declaration */, containerSymbol);
            mutate(jsNamespaceNode).symbol = sym;
            if (existingBinding) {
                if (existingBinding.flags & 256 /* SymbolFlags.Namespace */) {
                    // since the namespace was "declared" as part of this source file,
                    // we can simply re-use it.
                    containerSymbol = existingBinding;
                }
                else {
                    // we have some conflict, lets report a duplicate binding error.
                    mutate(containerSymbol.exports).set(part, sym);
                }
            }
            else {
                mutate(sym).exports = createSymbolTable();
                mutate(containerSymbol.exports).set(part, sym);
                containerSymbol = sym;
            }
        }
        let sym;
        if (kind === "decorator") {
            tracer.trace("decorator", `Bound decorator "@${name}" in namespace "${nsParts.join(".")}".`);
            sym = createSymbol(sourceFile, "@" + name, 512 /* SymbolFlags.Decorator */ | 1048576 /* SymbolFlags.Declaration */ | 2097152 /* SymbolFlags.Implementation */, containerSymbol);
        }
        else {
            tracer.trace("function", `Bound function "${name}" in namespace "${nsParts.join(".")}".`);
            sym = createSymbol(sourceFile, name, 2048 /* SymbolFlags.Function */ | 1048576 /* SymbolFlags.Declaration */ | 2097152 /* SymbolFlags.Implementation */, containerSymbol);
        }
        mutate(sym).value = fn;
        mutate(containerSymbol.exports).set(sym.name, sym);
    }
    function resolveJSMemberNamespaceParts(rootNs, member) {
        const memberNs = member.namespace;
        const nsParts = [];
        if (rootNs) {
            nsParts.push(...rootNs.split("."));
        }
        if (memberNs) {
            nsParts.push(...memberNs.split("."));
        }
        return nsParts;
    }
    function bindSourceFile(script) {
        if (script.locals !== undefined) {
            return;
        }
        mutate(script).locals = createSymbolTable();
        mutate(script).symbol = createSymbol(script, script.file.path, 32768 /* SymbolFlags.SourceFile */);
        mutate(script.symbol).exports = createSymbolTable();
        fileNamespace = undefined;
        currentFile = script;
        scope = script;
        bindNode(script);
    }
    function bindNode(node) {
        if (!node)
            return;
        // set the node's parent since we're going for a walk anyway
        mutate(node).parent = parentNode;
        switch (node.kind) {
            case SyntaxKind.ModelStatement:
                bindModelStatement(node);
                break;
            case SyntaxKind.ModelExpression:
                bindModelExpression(node);
                break;
            case SyntaxKind.ModelProperty:
                bindModelProperty(node);
                break;
            case SyntaxKind.IntersectionExpression:
                bindIntersectionExpression(node);
                break;
            case SyntaxKind.ScalarStatement:
                bindScalarStatement(node);
                break;
            case SyntaxKind.ScalarConstructor:
                bindScalarConstructor(node);
                break;
            case SyntaxKind.InterfaceStatement:
                bindInterfaceStatement(node);
                break;
            case SyntaxKind.UnionStatement:
                bindUnionStatement(node);
                break;
            case SyntaxKind.AliasStatement:
                bindAliasStatement(node);
                break;
            case SyntaxKind.ConstStatement:
                bindConstStatement(node);
                break;
            case SyntaxKind.EnumStatement:
                bindEnumStatement(node);
                break;
            case SyntaxKind.EnumMember:
                bindEnumMember(node);
                break;
            case SyntaxKind.UnionVariant:
                bindUnionVariant(node);
                break;
            case SyntaxKind.NamespaceStatement:
                bindNamespaceStatement(node);
                break;
            case SyntaxKind.OperationStatement:
                bindOperationStatement(node);
                break;
            case SyntaxKind.TemplateParameterDeclaration:
                bindTemplateParameterDeclaration(node);
                break;
            case SyntaxKind.UsingStatement:
                bindUsingStatement(node);
                break;
            case SyntaxKind.DecoratorDeclarationStatement:
                bindDecoratorDeclarationStatement(node);
                break;
            case SyntaxKind.FunctionDeclarationStatement:
                bindFunctionDeclarationStatement(node);
                break;
            case SyntaxKind.FunctionParameter:
                bindFunctionParameter(node);
                break;
        }
        const prevParent = parentNode;
        // set parent node when we walk into children
        parentNode = node;
        if (hasScope(node)) {
            const prevScope = scope;
            scope = node;
            visitChildren(node, bindNode);
            if ("locals" in node) {
                program.reportDuplicateSymbols(node.locals);
            }
            scope = prevScope;
        }
        else {
            visitChildren(node, bindNode);
        }
        // restore parent node
        parentNode = prevParent;
    }
    function bindTemplateParameterDeclaration(node) {
        declareSymbol(node, 1024 /* SymbolFlags.TemplateParameter */ | 1048576 /* SymbolFlags.Declaration */);
    }
    function bindModelStatement(node) {
        declareSymbol(node, 2 /* SymbolFlags.Model */ | 1048576 /* SymbolFlags.Declaration */);
        // Initialize locals for type parameters
        mutate(node).locals = new SymbolTable();
    }
    function bindModelExpression(node) {
        bindSymbol(node, 2 /* SymbolFlags.Model */);
    }
    function bindModelProperty(node) {
        declareMember(node, 65536 /* SymbolFlags.Member */, node.id.sv);
    }
    function bindIntersectionExpression(node) {
        bindSymbol(node, 2 /* SymbolFlags.Model */);
    }
    function bindScalarStatement(node) {
        declareSymbol(node, 4 /* SymbolFlags.Scalar */ | 1048576 /* SymbolFlags.Declaration */);
        // Initialize locals for type parameters
        mutate(node).locals = new SymbolTable();
    }
    function bindScalarConstructor(node) {
        declareMember(node, 65536 /* SymbolFlags.Member */, node.id.sv);
    }
    function bindInterfaceStatement(node) {
        declareSymbol(node, 32 /* SymbolFlags.Interface */ | 1048576 /* SymbolFlags.Declaration */);
        mutate(node).locals = new SymbolTable();
    }
    function bindUnionStatement(node) {
        declareSymbol(node, 64 /* SymbolFlags.Union */ | 1048576 /* SymbolFlags.Declaration */);
        mutate(node).locals = new SymbolTable();
    }
    function bindAliasStatement(node) {
        declareSymbol(node, 128 /* SymbolFlags.Alias */ | 1048576 /* SymbolFlags.Declaration */);
        // Initialize locals for type parameters
        mutate(node).locals = new SymbolTable();
    }
    function bindConstStatement(node) {
        declareSymbol(node, 131072 /* SymbolFlags.Const */ | 1048576 /* SymbolFlags.Declaration */);
    }
    function bindEnumStatement(node) {
        declareSymbol(node, 16 /* SymbolFlags.Enum */ | 1048576 /* SymbolFlags.Declaration */);
    }
    function bindEnumMember(node) {
        declareMember(node, 65536 /* SymbolFlags.Member */, node.id.sv);
    }
    function bindUnionVariant(node) {
        // cannot bind non named variant `union A { "a", "b"}`
        if (node.id) {
            declareMember(node, 65536 /* SymbolFlags.Member */, node.id.sv);
        }
    }
    function bindNamespaceStatement(statement) {
        const effectiveScope = fileNamespace ?? scope;
        // check if there's an existing symbol for this namespace
        const existingBinding = effectiveScope.symbol.exports.get(statement.id.sv);
        if (existingBinding && existingBinding.flags & 256 /* SymbolFlags.Namespace */) {
            mutate(statement).symbol = existingBinding;
            // locals are never shared.
            mutate(statement).locals = createSymbolTable();
            mutate(existingBinding.declarations).push(statement);
        }
        else {
            // Initialize locals for non-exported symbols
            mutate(statement).locals = createSymbolTable();
            declareSymbol(statement, 256 /* SymbolFlags.Namespace */ | 1048576 /* SymbolFlags.Declaration */);
        }
        currentFile.namespaces.push(statement);
        if (statement.statements === undefined) {
            fileNamespace = statement;
            let current = statement;
            while (current.kind !== SyntaxKind.TypeSpecScript) {
                currentFile.inScopeNamespaces.push(current);
                current = current.parent;
            }
        }
    }
    function bindUsingStatement(statement) {
        mutate(currentFile.usings).push(statement);
    }
    function bindOperationStatement(statement) {
        if (scope.kind === SyntaxKind.InterfaceStatement) {
            declareMember(statement, 8 /* SymbolFlags.Operation */ | 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */, statement.id.sv);
        }
        else {
            declareSymbol(statement, 8 /* SymbolFlags.Operation */ | 1048576 /* SymbolFlags.Declaration */);
        }
        mutate(statement).locals = createSymbolTable();
    }
    function bindDecoratorDeclarationStatement(node) {
        declareSymbol(node, 512 /* SymbolFlags.Decorator */ | 1048576 /* SymbolFlags.Declaration */, `@${node.id.sv}`);
    }
    function bindFunctionDeclarationStatement(node) {
        declareSymbol(node, 2048 /* SymbolFlags.Function */ | 1048576 /* SymbolFlags.Declaration */);
    }
    function bindFunctionParameter(node) {
        const symbol = createSymbol(node, node.id.sv, 4096 /* SymbolFlags.FunctionParameter */ | 1048576 /* SymbolFlags.Declaration */, scope.symbol);
        mutate(node).symbol = symbol;
    }
    /**
     * Declare a symbol for the given node in the current scope.
     * @param node Node
     * @param flags Symbol flags
     * @param name Optional symbol name, default to the node id.
     * @returns Created Symbol
     */
    function declareSymbol(node, flags, name) {
        compilerAssert(flags & 1048576 /* SymbolFlags.Declaration */, `Expected declaration symbol: ${name}`, node);
        switch (scope.kind) {
            case SyntaxKind.NamespaceStatement:
                return declareNamespaceMember(node, flags, name);
            case SyntaxKind.TypeSpecScript:
            case SyntaxKind.JsSourceFile:
                return declareScriptMember(node, flags, name);
            default:
                const key = name ?? node.id.sv;
                const symbol = createSymbol(node, key, flags, scope?.symbol);
                mutate(node).symbol = symbol;
                mutate(scope.locals).set(key, symbol);
                return symbol;
        }
    }
    function bindSymbol(node, flags) {
        const symbol = createSymbol(node, "-", flags, scope?.symbol);
        mutate(node).symbol = symbol;
        return symbol;
    }
    function declareNamespaceMember(node, flags, name) {
        if (flags & 256 /* SymbolFlags.Namespace */ &&
            mergeNamespaceDeclarations(node, scope)) {
            return;
        }
        const key = name ?? node.id.sv;
        const symbol = createSymbol(node, key, flags, scope.symbol);
        mutate(node).symbol = symbol;
        mutate(scope.symbol.exports).set(key, symbol);
        return symbol;
    }
    function declareScriptMember(node, flags, name) {
        const effectiveScope = fileNamespace ?? scope;
        if (flags & 256 /* SymbolFlags.Namespace */ &&
            mergeNamespaceDeclarations(node, effectiveScope)) {
            return;
        }
        const key = name ?? node.id.sv;
        const symbol = createSymbol(node, key, flags, fileNamespace?.symbol);
        mutate(node).symbol = symbol;
        mutate(effectiveScope.symbol.exports).set(key, symbol);
        return symbol;
    }
    /**
     * Declare a member of a model, enum, union, or interface.
     * @param node node of the member
     * @param flags symbol flags
     * @param name name of the symbol
     */
    function declareMember(node, flags, name) {
        const symbol = createSymbol(node, name, flags, scope.symbol);
        mutate(node).symbol = symbol;
        mutate(scope.symbol.members).set(name, symbol);
        return symbol;
    }
    function mergeNamespaceDeclarations(node, scope) {
        // we are declaring a namespace in either global scope, or a blockless namespace.
        const existingBinding = scope.symbol.exports.get(node.id.sv);
        if (existingBinding && existingBinding.flags & 256 /* SymbolFlags.Namespace */) {
            // we have an existing binding, so just push this node to its declarations
            mutate(existingBinding.declarations).push(node);
            mutate(node).symbol = existingBinding;
            return true;
        }
        return false;
    }
}
function hasScope(node) {
    switch (node.kind) {
        case SyntaxKind.ModelStatement:
        case SyntaxKind.ModelExpression:
        case SyntaxKind.ScalarStatement:
        case SyntaxKind.ConstStatement:
        case SyntaxKind.AliasStatement:
        case SyntaxKind.TypeSpecScript:
        case SyntaxKind.InterfaceStatement:
        case SyntaxKind.OperationStatement:
        case SyntaxKind.UnionStatement:
        case SyntaxKind.EnumStatement:
            return true;
        case SyntaxKind.NamespaceStatement:
            return node.statements !== undefined;
        default:
            return false;
    }
}
function createSymbol(node, name, flags, parent, value) {
    let exports;
    if (flags & 33024 /* SymbolFlags.ExportContainer */) {
        exports = createSymbolTable();
    }
    let members;
    if (flags & 118 /* SymbolFlags.MemberContainer */) {
        members = createSymbolTable();
    }
    compilerAssert(!(flags & 1048576 /* SymbolFlags.Declaration */) || node !== undefined, "Declaration without node");
    return {
        declarations: flags & 1048576 /* SymbolFlags.Declaration */ ? [node] : [],
        node: !(flags & 1048576 /* SymbolFlags.Declaration */) ? node : undefined,
        name,
        exports,
        members,
        flags,
        value,
        parent,
        metatypeMembers: createSymbolTable(),
    };
}
/**
 * Get the node attached to this symbol.
 * If a declaration symbol get the first one `.declarations[0]`
 * Otherwise get `.node`
 */
function getSymNode(sym) {
    return sym.flags & 1048576 /* SymbolFlags.Declaration */ ? sym.declarations[0] : sym.node;
}

function createChangeIdentifierCodeFix(node, newIdentifier) {
    return defineCodeFix({
        id: "change-identifier",
        label: `Change ${node.sv} to ${newIdentifier}`,
        fix: (context) => {
            const location = getSourceLocation(node);
            return context.replaceText(location, newIdentifier);
        },
    });
}

/**
 * Quick fix that convert a tuple to an array value.
 */
function createTupleToArrayValueCodeFix(node) {
    return defineCodeFix({
        id: "tuple-to-array-value",
        label: `Convert to an array value \`#[]\``,
        fix: (context) => {
            const result = [];
            addCreatedCodeFixResult(node);
            createChildTupleToArrValCodeFix(node, addCreatedCodeFixResult);
            return result;
            function addCreatedCodeFixResult(node) {
                const location = getSourceLocation(node);
                result.push(context.prependText(location, "#"));
            }
        },
    });
}
/**
 * Quick fix that convert a model expression to an object value.
 */
function createModelToObjectValueCodeFix(node) {
    return defineCodeFix({
        id: "model-to-object-value",
        label: `Convert to an object value \`#{}\``,
        fix: (context) => {
            const result = [];
            addCreatedCodeFixResult(node);
            createChildModelToObjValCodeFix(node, addCreatedCodeFixResult);
            return result;
            function addCreatedCodeFixResult(node) {
                const location = getSourceLocation(node);
                result.push(context.prependText(location, "#"));
            }
        },
    });
}
function createChildTupleToArrValCodeFix(node, addCreatedCodeFixResult) {
    for (const childNode of node.values) {
        if (childNode.kind === SyntaxKind.ModelExpression) {
            addCreatedCodeFixResult(childNode);
            createChildModelToObjValCodeFix(childNode, addCreatedCodeFixResult);
        }
        else if (childNode.kind === SyntaxKind.TupleExpression) {
            addCreatedCodeFixResult(childNode);
            createChildTupleToArrValCodeFix(childNode, addCreatedCodeFixResult);
        }
    }
}
function createChildModelToObjValCodeFix(node, addCreatedCodeFixResult) {
    for (const prop of node.properties.values()) {
        if (prop.kind === SyntaxKind.ModelProperty) {
            const childNode = prop.value;
            if (childNode.kind === SyntaxKind.ModelExpression) {
                addCreatedCodeFixResult(childNode);
                createChildModelToObjValCodeFix(childNode, addCreatedCodeFixResult);
            }
            else if (childNode.kind === SyntaxKind.TupleExpression) {
                addCreatedCodeFixResult(childNode);
                createChildTupleToArrValCodeFix(childNode, addCreatedCodeFixResult);
            }
        }
    }
}

function isStringTemplateSerializable(stringTemplate) {
    if (stringTemplate.stringValue !== undefined) {
        return [true, []];
    }
    else {
        return [false, explainStringTemplateNotSerializable(stringTemplate)];
    }
}
/**
 * get a list of diagnostic explaining why this string template cannot be converted to a string.
 */
function explainStringTemplateNotSerializable(stringTemplate) {
    const diagnostics = createDiagnosticCollector();
    for (const span of stringTemplate.spans) {
        if (span.isInterpolated) {
            switch (span.type.kind) {
                case "String":
                case "Number":
                case "Boolean":
                    break;
                case "StringTemplate":
                    diagnostics.pipe(isStringTemplateSerializable(span.type));
                    break;
                case "TemplateParameter":
                    if (span.type.constraint && span.type.constraint.valueType !== undefined) {
                        break; // Value types will be serializable in the template instance.
                    }
                // eslint-disable-next-line no-fallthrough
                default:
                    diagnostics.add(createDiagnostic({
                        code: "non-literal-string-template",
                        target: span.node,
                    }));
            }
        }
    }
    return diagnostics.diagnostics;
}

/**
 * Set of known numeric ranges
 */
const numericRanges = {
    int64: [
        Numeric("-9223372036854775808"),
        Numeric("9223372036854775807"),
        { int: true, isJsNumber: false },
    ],
    int32: [Numeric("-2147483648"), Numeric("2147483647"), { int: true, isJsNumber: true }],
    int16: [Numeric("-32768"), Numeric("32767"), { int: true, isJsNumber: true }],
    int8: [Numeric("-128"), Numeric("127"), { int: true, isJsNumber: true }],
    uint64: [Numeric("0"), Numeric("18446744073709551615"), { int: true, isJsNumber: false }],
    uint32: [Numeric("0"), Numeric("4294967295"), { int: true, isJsNumber: true }],
    uint16: [Numeric("0"), Numeric("65535"), { int: true, isJsNumber: true }],
    uint8: [Numeric("0"), Numeric("255"), { int: true, isJsNumber: true }],
    safeint: [
        Numeric(Number.MIN_SAFE_INTEGER.toString()),
        Numeric(Number.MAX_SAFE_INTEGER.toString()),
        { int: true, isJsNumber: true },
    ],
    float32: [Numeric("-3.4e38"), Numeric("3.4e38"), { int: false, isJsNumber: true }],
    float64: [
        Numeric(`${-Number.MAX_VALUE}`),
        Numeric(Number.MAX_VALUE.toString()),
        { int: false, isJsNumber: true },
    ],
};

function marshallTypeForJS(value, valueConstraint) {
    switch (value.valueKind) {
        case "BooleanValue":
        case "StringValue":
            return value.value;
        case "NumericValue":
            return numericValueToJs(value, valueConstraint);
        case "ObjectValue":
            return objectValueToJs(value);
        case "ArrayValue":
            return arrayValueToJs(value);
        case "EnumValue":
            return value;
        case "NullValue":
            return null;
        case "ScalarValue":
            return value;
    }
}
function isNumericScalar(scalar) {
    let current = scalar;
    while (current) {
        if (current.name === "numeric" && current.namespace?.name === "TypeSpec") {
            return true;
        }
        current = current.baseScalar;
    }
    return false;
}
function canNumericConstraintBeJsNumber(type) {
    if (type === undefined)
        return true;
    switch (type.kind) {
        case "Scalar":
            if (isNumericScalar(type)) {
                return numericRanges[type.name]?.[2].isJsNumber;
            }
            else {
                return true;
            }
        case "Union":
            return [...type.variants.values()].every((x) => canNumericConstraintBeJsNumber(x.type));
        default:
            return true;
    }
}
function numericValueToJs(type, valueConstraint) {
    const canBeANumber = canNumericConstraintBeJsNumber(valueConstraint);
    if (canBeANumber) {
        const asNumber = type.value.asNumber();
        compilerAssert(asNumber !== null, `Numeric value '${type.value.toString()}' is not a able to convert to a number without loosing precision.`);
        return asNumber;
    }
    return type.value;
}
function objectValueToJs(type) {
    const result = {};
    for (const [key, value] of type.properties) {
        result[key] = marshallTypeForJS(value.value, undefined);
    }
    return result;
}
function arrayValueToJs(type) {
    return type.values.map((x) => marshallTypeForJS(x, undefined));
}

var Related;
(function (Related) {
    Related[Related["false"] = 0] = "false";
    Related[Related["true"] = 1] = "true";
    Related[Related["maybe"] = 2] = "maybe";
})(Related || (Related = {}));
/**
 * Mapping from the reflection models to Type["kind"] value
 */
const ReflectionNameToKind = {
    Enum: "Enum",
    EnumMember: "EnumMember",
    Interface: "Interface",
    Model: "Model",
    ModelProperty: "ModelProperty",
    Namespace: "Namespace",
    Operation: "Operation",
    Scalar: "Scalar",
    TemplateParameter: "TemplateParameter",
    Tuple: "Tuple",
    Union: "Union",
    UnionVariant: "UnionVariant",
};
function createTypeRelationChecker(program, checker) {
    return {
        isTypeAssignableTo,
        isValueOfType,
        isReflectionType,
        areScalarsRelated,
    };
    /**
     * Check if the source type can be assigned to the target type.
     * @param source Source type
     * @param target Target type
     * @param diagnosticTarget Target for the diagnostic, unless something better can be inferred.
     */
    function isTypeAssignableTo(source, target, diagnosticTarget) {
        const [related, errors] = isTypeAssignableToInternal(source, target, diagnosticTarget, new MultiKeyMap());
        return [related === Related.true, convertErrorsToDiagnostics(errors, diagnosticTarget)];
    }
    function isTargetChildOf(target, base) {
        const errorNode = "kind" in target && typeof target.kind === "number" ? target : target.node;
        const baseNode = "kind" in base && typeof base.kind === "number" ? base : base.node;
        let currentNode = errorNode;
        while (currentNode) {
            if (currentNode === baseNode) {
                return true;
            }
            currentNode = currentNode.parent;
        }
        return false;
    }
    function convertErrorsToDiagnostics(errors, diagnosticBase) {
        return errors.flatMap((x) => convertErrorToDiagnostic(x, diagnosticBase));
    }
    function combineErrorMessage(error) {
        let message = error.message;
        let current = error.children[0];
        let indent = "  ";
        while (current !== undefined) {
            message += current.message
                .split("\n")
                .map((line) => `\n${indent}${line}`)
                .join("");
            indent += "  ";
            current = current.children[0];
        }
        return message;
    }
    function flattenErrors(error, diagnosticBase) {
        if (!isTargetChildOf(error.target, diagnosticBase)) {
            return [{ ...error, target: diagnosticBase }];
        }
        if (error.children.length === 0) {
            return [error];
        }
        return error.children.flatMap((x) => flattenErrors(x, error.target));
    }
    function convertErrorToDiagnostic(error, diagnosticBase) {
        const flattened = flattenErrors(error, diagnosticBase);
        return flattened.map((error) => {
            const messageBase = error.skipIfFirst && error.children.length > 0 ? error.children[0] : error;
            const message = combineErrorMessage(messageBase);
            return {
                severity: "error",
                code: error.code,
                message: message,
                target: error.target,
            };
        });
    }
    /**
     * Check if the given Value type is of the given type.
     * @param source Value
     * @param target Target type
     * @param diagnosticTarget Target for the diagnostic, unless something better can be inferred.
     */
    function isValueOfType(source, target, diagnosticTarget) {
        const [related, errors] = isValueOfTypeInternal(source, target, diagnosticTarget, new MultiKeyMap());
        return [related === Related.true, convertErrorsToDiagnostics(errors, diagnosticTarget)];
    }
    function isTypeAssignableToInternal(source, target, diagnosticTarget, relationCache) {
        const cached = relationCache.get([source, target]);
        if (cached !== undefined) {
            return [cached, []];
        }
        const [result, diagnostics] = isTypeAssignableToWorker(source, target, diagnosticTarget, new MultiKeyMap());
        relationCache.set([source, target], result);
        return [result, diagnostics];
    }
    function isTypeAssignableToWorker(source, target, diagnosticTarget, relationCache) {
        if ("kind" in source && source.kind === "TemplateParameter") {
            source = source.constraint ?? checker.anyType;
        }
        if (target.entityKind === "Indeterminate") {
            target = target.type;
        }
        if (source === target)
            return [Related.true, []];
        if (isValue(target)) {
            return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
        }
        if (source.entityKind === "Indeterminate") {
            return isIndeterminateEntityAssignableTo(source, target, diagnosticTarget, relationCache);
        }
        if (target.entityKind === "MixedParameterConstraint") {
            return isAssignableToMixedParameterConstraint(source, target, diagnosticTarget, relationCache);
        }
        if (isValue(source) || (source.entityKind === "MixedParameterConstraint" && source.valueType)) {
            return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
        }
        if (source.entityKind === "MixedParameterConstraint") {
            return isTypeAssignableToInternal(source.type, target, diagnosticTarget, relationCache);
        }
        const isSimpleTypeRelated = isSimpleTypeAssignableTo(source, target);
        if (isSimpleTypeRelated === true) {
            return [Related.true, []];
        }
        else if (isSimpleTypeRelated === false) {
            return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
        }
        if (source.kind === "Union") {
            for (const variant of source.variants.values()) {
                const [variantAssignable] = isTypeAssignableToInternal(variant.type, target, diagnosticTarget, relationCache);
                if (!variantAssignable) {
                    return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
                }
            }
            return [Related.true, []];
        }
        if (target.kind === "Model" &&
            source.kind === "Model" &&
            target.name !== "object" &&
            target.indexer === undefined &&
            source.indexer &&
            source.indexer.key.name === "integer") {
            return [
                Related.false,
                [
                    createTypeRelationError({
                        code: "missing-index",
                        format: {
                            indexType: getTypeName(source.indexer.key),
                            sourceType: getTypeName(target),
                        },
                        diagnosticTarget,
                    }),
                ],
            ];
        }
        else if (target.kind === "Model" &&
            isArrayModelType(program, target) &&
            source.kind === "Model") {
            if (isArrayModelType(program, source)) {
                return hasIndexAndIsAssignableTo(source, target, diagnosticTarget, relationCache);
            }
            else {
                // For other models just fallback to unassignable
                return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
            }
        }
        else if (target.kind === "Model" && source.kind === "Model") {
            return areModelsRelated(source, target, diagnosticTarget, relationCache);
        }
        else if (target.kind === "Model" &&
            isArrayModelType(program, target) &&
            source.kind === "Tuple") {
            return isTupleAssignableToArray(source, target, diagnosticTarget, relationCache);
        }
        else if (target.kind === "Tuple" && source.kind === "Tuple") {
            return isTupleAssignableToTuple(source, target, diagnosticTarget, relationCache);
        }
        else if (target.kind === "Union") {
            return isAssignableToUnion(source, target, diagnosticTarget, relationCache);
        }
        else if (target.kind === "Enum") {
            return isAssignableToEnum(source, target, diagnosticTarget);
        }
        return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
    }
    function isIndeterminateEntityAssignableTo(indeterminate, target, diagnosticTarget, relationCache) {
        const [typeRelated, typeDiagnostics] = isTypeAssignableToInternal(indeterminate.type, target, diagnosticTarget, relationCache);
        if (typeRelated) {
            return [Related.true, []];
        }
        if (target.entityKind === "MixedParameterConstraint" && target.valueType) {
            const [valueRelated] = isTypeAssignableToInternal(indeterminate.type, target.valueType, diagnosticTarget, relationCache);
            if (valueRelated) {
                return [Related.true, []];
            }
        }
        return [Related.false, typeDiagnostics];
    }
    function isAssignableToValueType(source, target, diagnosticTarget, relationCache) {
        if (!isValue(source)) {
            return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
        }
        return isValueOfTypeInternal(source, target, diagnosticTarget, relationCache);
    }
    function isAssignableToMixedParameterConstraint(source, target, diagnosticTarget, relationCache) {
        if ("entityKind" in source && source.entityKind === "MixedParameterConstraint") {
            if (source.type && target.type) {
                const [variantAssignable, diagnostics] = isTypeAssignableToInternal(source.type, target.type, diagnosticTarget, relationCache);
                if (variantAssignable === Related.false) {
                    return [Related.false, diagnostics];
                }
                return [Related.true, []];
            }
            if (source.valueType && target.valueType) {
                const [variantAssignable, diagnostics] = isTypeAssignableToInternal(source.valueType, target.valueType, diagnosticTarget, relationCache);
                if (variantAssignable === Related.false) {
                    return [Related.false, diagnostics];
                }
                return [Related.true, []];
            }
            return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
        }
        if (target.type) {
            const [related] = isTypeAssignableToInternal(source, target.type, diagnosticTarget, relationCache);
            if (related) {
                return [Related.true, []];
            }
        }
        if (target.valueType) {
            const [related] = isAssignableToValueType(source, target.valueType, diagnosticTarget, relationCache);
            if (related) {
                return [Related.true, []];
            }
        }
        return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
    }
    /** Check if the value is assignable to the given type. */
    function isValueOfTypeInternal(source, target, diagnosticTarget, relationCache) {
        return isTypeAssignableToInternal(source.type, target, diagnosticTarget, relationCache);
    }
    function isReflectionType(type) {
        return (type.kind === "Model" &&
            type.namespace?.name === "Reflection" &&
            type.namespace?.namespace?.name === "TypeSpec");
    }
    function isRelatedToScalar(source, target) {
        switch (source.kind) {
            case "Number":
                return isNumericLiteralRelatedTo(source, target);
            case "String":
            case "StringTemplate":
                return isStringLiteralRelatedTo(source, target);
            case "Boolean":
                return areScalarsRelated(target, checker.getStdType("boolean"));
            case "Scalar":
                return areScalarsRelated(source, target);
            case "Union":
                return undefined;
            default:
                return false;
        }
    }
    function areScalarsRelated(source, target) {
        let current = source;
        while (current) {
            if (current === target) {
                return true;
            }
            current = current.baseScalar;
        }
        return false;
    }
    function isSimpleTypeAssignableTo(source, target) {
        if (isNeverType(source))
            return true;
        if (isVoidType(target))
            return false;
        if (isUnknownType(target))
            return true;
        if (isReflectionType(target)) {
            return source.kind === ReflectionNameToKind[target.name];
        }
        if (target.kind === "Scalar") {
            return isRelatedToScalar(source, target);
        }
        if (source.kind === "Scalar" && target.kind === "Model") {
            return false;
        }
        if (target.kind === "String") {
            return ((source.kind === "String" && source.value === target.value) ||
                (source.kind === "StringTemplate" && source.stringValue === target.value));
        }
        if (target.kind === "StringTemplate" && target.stringValue) {
            return ((source.kind === "String" && source.value === target.stringValue) ||
                (source.kind === "StringTemplate" && source.stringValue === target.stringValue));
        }
        if (target.kind === "Number") {
            return source.kind === "Number" && target.value === source.value;
        }
        return undefined;
    }
    function isNumericLiteralRelatedTo(source, target) {
        // First check that the source numeric literal is assignable to the target scalar
        if (!isNumericAssignableToNumericScalar(source.numericValue, target)) {
            return false;
        }
        const min = getMinValueAsNumeric(program, target);
        const max = getMaxValueAsNumeric(program, target);
        const minExclusive = getMinValueExclusiveAsNumeric(program, target);
        const maxExclusive = getMaxValueExclusiveAsNumeric(program, target);
        if (min && source.numericValue.lt(min)) {
            return false;
        }
        if (minExclusive && source.numericValue.lte(minExclusive)) {
            return false;
        }
        if (max && source.numericValue.gt(max)) {
            return false;
        }
        if (maxExclusive && source.numericValue.gte(maxExclusive)) {
            return false;
        }
        return true;
    }
    function isNumericAssignableToNumericScalar(source, target) {
        // if the target does not derive from numeric, then it can't be assigned a numeric literal
        if (!areScalarsRelated(target, checker.getStdType("numeric"))) {
            return false;
        }
        // With respect to literal assignability a custom numeric scalar is
        // equivalent to its nearest TypeSpec.* base. Adjust target accordingly.
        while (!target.namespace || !isTypeSpecNamespace(target.namespace)) {
            compilerAssert(target.baseScalar, "Should not be possible to be derived from TypeSpec.numeric and not have a base when not in TypeSpec namespace.");
            target = target.baseScalar;
        }
        if (target.name === "numeric")
            return true;
        if (target.name === "decimal")
            return true;
        if (target.name === "decimal128")
            return true;
        const isInt = source.isInteger;
        if (target.name === "integer")
            return isInt;
        if (target.name === "float")
            return true;
        if (!(target.name in numericRanges))
            return false;
        const [low, high, options] = numericRanges[target.name];
        return source.gte(low) && source.lte(high) && (!options.int || isInt);
    }
    function isStringLiteralRelatedTo(source, target) {
        if (!areScalarsRelated(target, checker.getStdType("string"))) {
            return false;
        }
        if (source.kind === "StringTemplate") {
            return true;
        }
        const len = source.value.length;
        const min = getMinLength(program, target);
        const max = getMaxLength(program, target);
        if (min && len < min) {
            return false;
        }
        if (max && len > max) {
            return false;
        }
        return true;
    }
    function areModelsRelated(source, target, diagnosticTarget, relationCache) {
        relationCache.set([source, target], Related.maybe);
        const errors = [];
        const remainingProperties = new Map(source.properties);
        for (const prop of walkPropertiesInherited(target)) {
            const sourceProperty = getProperty(source, prop.name);
            if (sourceProperty === undefined) {
                if (!prop.optional) {
                    errors.push(createTypeRelationError({
                        code: "missing-property",
                        format: {
                            propertyName: prop.name,
                            sourceType: getTypeName(source),
                            targetType: getTypeName(target),
                        },
                        diagnosticTarget: source,
                    }));
                }
            }
            else {
                remainingProperties.delete(prop.name);
                if (sourceProperty.optional && !prop.optional) {
                    errors.push(createTypeRelationError({
                        code: "property-required",
                        format: {
                            propName: prop.name,
                            targetType: getTypeName(target),
                        },
                        diagnosticTarget,
                    }));
                }
                const [related, propErrors] = isTypeAssignableToInternal(sourceProperty.type, prop.type, diagnosticTarget, relationCache);
                if (!related) {
                    errors.push(...wrapUnassignablePropertyErrors(sourceProperty, propErrors));
                }
            }
        }
        if (target.indexer) {
            const [_, indexerDiagnostics] = arePropertiesAssignableToIndexer(remainingProperties, target.indexer.value, diagnosticTarget, relationCache);
            errors.push(...indexerDiagnostics);
            // For anonymous models we don't need an indexer
            if (source.name !== "" && target.indexer.key.name !== "integer") {
                const [related, indexDiagnostics] = hasIndexAndIsAssignableTo(source, target, diagnosticTarget, relationCache);
                if (!related) {
                    errors.push(...indexDiagnostics);
                }
            }
        }
        else if (shouldCheckExcessProperties(source)) {
            for (const [propName, prop] of remainingProperties) {
                if (shouldCheckExcessProperty(prop)) {
                    errors.push(createTypeRelationError({
                        code: "unexpected-property",
                        format: {
                            propertyName: propName,
                            type: getEntityName(target),
                        },
                        diagnosticTarget: prop,
                    }));
                }
            }
        }
        return [
            errors.length === 0 ? Related.true : Related.false,
            wrapUnassignableErrors(source, target, errors),
        ];
    }
    /** If we should check for excess properties on the given model. */
    function shouldCheckExcessProperties(model) {
        return model.node?.kind === SyntaxKind.ObjectLiteral;
    }
    /** If we should check for this specific property */
    function shouldCheckExcessProperty(prop) {
        return (prop.node?.kind === SyntaxKind.ObjectLiteralProperty && prop.node.parent === prop.model?.node);
    }
    function getProperty(model, name) {
        return (model.properties.get(name) ??
            (model.baseModel !== undefined ? getProperty(model.baseModel, name) : undefined));
    }
    function arePropertiesAssignableToIndexer(properties, indexerConstaint, diagnosticTarget, relationCache) {
        for (const prop of properties.values()) {
            const [related, diagnostics] = isTypeAssignableToInternal(prop.type, indexerConstaint, diagnosticTarget, relationCache);
            if (!related) {
                return [Related.false, diagnostics];
            }
        }
        return [Related.true, []];
    }
    /** Check that the source model has an index, the index key match and the value of the source index is assignable to the target index. */
    function hasIndexAndIsAssignableTo(source, target, diagnosticTarget, relationCache) {
        if (source.indexer === undefined || source.indexer.key !== target.indexer.key) {
            return [
                Related.false,
                [
                    createTypeRelationError({
                        code: "missing-index",
                        format: {
                            indexType: getTypeName(target.indexer.key),
                            sourceType: getTypeName(source),
                        },
                        diagnosticTarget,
                    }),
                ],
            ];
        }
        return isTypeAssignableToInternal(source.indexer.value, target.indexer.value, diagnosticTarget, relationCache);
    }
    function isTupleAssignableToArray(source, target, diagnosticTarget, relationCache) {
        const minItems = getMinItems(program, target);
        const maxItems = getMaxItems(program, target);
        if (minItems !== undefined && source.values.length < minItems) {
            return [
                Related.false,
                [
                    createUnassignableDiagnostic(source, target, diagnosticTarget, `Source has ${source.values.length} element(s) but target requires ${minItems}.`),
                ],
            ];
        }
        if (maxItems !== undefined && source.values.length > maxItems) {
            return [
                Related.false,
                [
                    createUnassignableDiagnostic(source, target, diagnosticTarget, `Source has ${source.values.length} element(s) but target only allows ${maxItems}.`),
                ],
            ];
        }
        for (const item of source.values) {
            const [related, diagnostics] = isTypeAssignableToInternal(item, target.indexer.value, diagnosticTarget, relationCache);
            if (!related) {
                return [Related.false, diagnostics];
            }
        }
        return [Related.true, []];
    }
    function isTupleAssignableToTuple(source, target, diagnosticTarget, relationCache) {
        if (source.values.length !== target.values.length) {
            return [
                Related.false,
                [
                    createUnassignableDiagnostic(source, target, diagnosticTarget, `Source has ${source.values.length} element(s) but target requires ${target.values.length}.`),
                ],
            ];
        }
        for (const [index, sourceItem] of source.values.entries()) {
            const targetItem = target.values[index];
            const [related, diagnostics] = isTypeAssignableToInternal(sourceItem, targetItem, diagnosticTarget, relationCache);
            if (!related) {
                return [Related.false, diagnostics];
            }
        }
        return [Related.true, []];
    }
    function isAssignableToUnion(source, target, diagnosticTarget, relationCache) {
        if (source.kind === "UnionVariant" && source.union === target) {
            return [Related.true, []];
        }
        for (const option of target.variants.values()) {
            const [related] = isTypeAssignableToInternal(source, option.type, diagnosticTarget, relationCache);
            if (related) {
                return [Related.true, []];
            }
        }
        return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
    }
    function isAssignableToEnum(source, target, diagnosticTarget) {
        switch (source.kind) {
            case "Enum":
                if (source === target) {
                    return [Related.true, []];
                }
                else {
                    return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
                }
            case "EnumMember":
                if (source.enum === target) {
                    return [Related.true, []];
                }
                else {
                    return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
                }
            default:
                return [Related.false, [createUnassignableDiagnostic(source, target, diagnosticTarget)]];
        }
    }
    function isTypeSpecNamespace(namespace) {
        return (namespace.name === "TypeSpec" && namespace.namespace === checker.getGlobalNamespaceType());
    }
}
function wrapUnassignableErrors(source, target, errors) {
    const error = createUnassignableDiagnostic(source, target, source);
    error.children = errors;
    return [error];
}
function wrapUnassignablePropertyErrors(source, errors) {
    const error = createTypeRelationError({
        code: "property-unassignable",
        diagnosticTarget: source,
        format: {
            propName: source.name,
        },
        skipIfFirst: true,
    });
    error.children = errors;
    return [error];
}
function createTypeRelationError({ code, format, details, diagnosticTarget, skipIfFirst, }) {
    const diag = createDiagnostic({
        code: code,
        format: format,
        target: NoTarget,
    });
    return {
        code: code,
        message: details ? `${diag.message}\n  ${details}` : diag.message,
        target: diagnosticTarget,
        skipIfFirst,
        children: [],
    };
}
function createUnassignableDiagnostic(source, target, diagnosticTarget, details) {
    return createTypeRelationError({
        code: "unassignable",
        format: {
            sourceType: getEntityName(source),
            targetType: getEntityName(target),
        },
        diagnosticTarget,
        details,
    });
}
// #endregion

/**
 * Maps type arguments to type instantiation.
 */
const TypeInstantiationMap = class extends MultiKeyMap {
};
function createChecker(program, resolver) {
    const stdTypes = {};
    const indeterminateEntities = new WeakMap();
    const docFromCommentForSym = new Map();
    const referenceSymCache = new WeakMap();
    const valueExactTypes = new WeakMap();
    let onCheckerDiagnostic = (x) => {
        program.reportDiagnostic(x);
    };
    const typePrototype = {};
    const globalNamespaceType = createGlobalNamespaceType();
    // Caches the deprecation test of nodes in the program
    const nodeDeprecationMap = new Map();
    const errorType = createType({ kind: "Intrinsic", name: "ErrorType" });
    const voidType = createType({ kind: "Intrinsic", name: "void" });
    const neverType = createType({ kind: "Intrinsic", name: "never" });
    const unknownType = createType({ kind: "Intrinsic", name: "unknown" });
    const nullType = createType({ kind: "Intrinsic", name: "null" });
    /**
     * Set keeping track of node pending type resolution.
     * Key is the SymId of a node. It can be retrieved with getNodeSymId(node)
     */
    const pendingResolutions = new PendingResolutions();
    const typespecNamespaceBinding = resolver.symbols.global.exports.get("TypeSpec");
    if (typespecNamespaceBinding) {
        initializeTypeSpecIntrinsics();
    }
    /**
     * Tracking the template parameters used or not.
     */
    const templateParameterUsageMap = new Map();
    const checker = {
        getTypeForNode,
        checkProgram,
        checkSourceFile,
        getLiteralType,
        getGlobalNamespaceType,
        cloneType,
        resolveRelatedSymbols,
        resolveCompletions,
        neverType,
        errorType,
        nullType,
        anyType: unknownType,
        voidType,
        typePrototype,
        createType,
        createAndFinishType,
        createLiteralType,
        finishType,
        isStdType,
        getStdType,
        resolveTypeReference,
        getValueForNode,
        getTypeOrValueForNode,
        getValueExactType,
        getTemplateParameterUsageMap,
        isTypeAssignableTo: undefined,
    };
    const relation = createTypeRelationChecker(program, checker);
    checker.isTypeAssignableTo = relation.isTypeAssignableTo;
    return checker;
    function getTemplateParameterUsageMap() {
        return templateParameterUsageMap;
    }
    function wrapInstantiationDiagnostic(diagnostic, templateMapper) {
        if (templateMapper === undefined || typeof diagnostic.target !== "object")
            return diagnostic;
        return {
            ...diagnostic,
            target: {
                node: diagnostic.target,
                templateMapper,
            },
        };
    }
    function reportCheckerDiagnostic(diagnostic, mapper) {
        onCheckerDiagnostic(wrapInstantiationDiagnostic(diagnostic, mapper));
    }
    function reportCheckerDiagnostics(diagnostics) {
        diagnostics.forEach((x) => reportCheckerDiagnostic(x));
    }
    function initializeTypeSpecIntrinsics() {
        // a utility function to log strings or numbers
        mutate(typespecNamespaceBinding.exports).set("log", {
            flags: 2048 /* SymbolFlags.Function */,
            name: "log",
            value(p, ...strs) {
                program.trace("log", strs.join(" "));
                return voidType;
            },
            declarations: [],
            node: undefined, // TODO: is this correct?
        });
        // Until we have an `unit` type for `null`
        mutate(resolver.symbols.null).type = nullType;
        getSymbolLinks(resolver.symbols.null).type = nullType;
    }
    function getStdType(name) {
        const type = stdTypes[name];
        if (type !== undefined) {
            return type;
        }
        const sym = typespecNamespaceBinding?.exports?.get(name);
        compilerAssert(sym, `Unexpected missing symbol to std type "${name}"`);
        if (sym.flags & 2 /* SymbolFlags.Model */) {
            checkModelStatement(sym.declarations[0], undefined);
        }
        else {
            checkScalar(sym.declarations[0], undefined);
        }
        const loadedType = stdTypes[name];
        compilerAssert(loadedType, `TypeSpec std type "${name}" should have been initalized before using array syntax.`);
        return loadedType;
    }
    /**
     * Create the link for the given type to the symbol links.
     * If currently instantiating a template it will link to the instantiations.
     * Else will link to the declaredType.
     * @param links Symbol link
     * @param type Type
     * @param mapper Type mapper if in an template instantiation
     */
    function linkType(links, type, mapper) {
        if (mapper === undefined) {
            links.declaredType = type;
            links.instantiations = new TypeInstantiationMap();
        }
        else if (links.instantiations) {
            links.instantiations.set(mapper.args, type);
        }
    }
    function linkMemberType(links, type, mapper) {
        if (mapper === undefined) {
            links.declaredType = type;
        }
    }
    /**
     * Check a member symbol.
     * @param sym Symbol binding a member node.
     * @param mapper Type mapper.
     * @returns Checked type for the given member symbol.
     */
    function checkMemberSym(sym, mapper) {
        const symbolLinks = getSymbolLinks(sym);
        const memberContainer = getTypeForNode(getSymNode(sym.parent), mapper);
        const type = symbolLinks.declaredType ?? symbolLinks.type;
        if (type) {
            return type;
        }
        else {
            return checkMember(getSymNode(sym), mapper, memberContainer);
        }
    }
    /**
     * Check a member node
     * @param node Member node to check
     * @param mapper Type mapper
     * @param containerType Member node container type(Interface, Model, Union, etc.)
     * @returns Checked member
     */
    function checkMember(node, mapper, containerType) {
        switch (node.kind) {
            case SyntaxKind.ModelProperty:
                return checkModelProperty(node, mapper);
            case SyntaxKind.EnumMember:
                return checkEnumMember(node, mapper, containerType);
            case SyntaxKind.OperationStatement:
                return checkOperation(node, mapper, containerType);
            case SyntaxKind.UnionVariant:
                return checkUnionVariant(node, mapper);
            case SyntaxKind.ScalarConstructor:
                return checkScalarConstructor(node, mapper, containerType);
        }
    }
    function getTypeForTypeOrIndeterminate(entity) {
        if (entity.entityKind === "Indeterminate") {
            return entity.type;
        }
        return entity;
    }
    function getTypeForNode(node, mapper) {
        const entity = checkNode(node, mapper);
        if (entity === null) {
            return errorType;
        }
        if (entity.entityKind === "Indeterminate") {
            return entity.type;
        }
        if (isValue(entity)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "value-in-type",
                target: node,
            }));
            return errorType;
        }
        if (entity.kind === "TemplateParameter") {
            if (entity.constraint?.valueType) {
                // means this template constraint will accept values
                reportCheckerDiagnostic(createDiagnostic({
                    code: "value-in-type",
                    messageId: "referenceTemplate",
                    target: node,
                }));
            }
        }
        return entity;
    }
    function getValueForNode(node, mapper, constraint, options = {}) {
        const initial = checkNode(node, mapper, constraint);
        if (initial === null) {
            return null;
        }
        let entity;
        if (initial.entityKind === "Indeterminate") {
            entity = getValueFromIndeterminate(initial.type, constraint, node);
        }
        else {
            entity = initial;
        }
        // if (options.legacyTupleAndModelCast && entity !== null && isType(entity)) {
        //   entity = legacy_tryTypeToValueCast(entity, constraint, node);
        // }
        if (entity === null) {
            return null;
        }
        if (isValue(entity)) {
            return constraint ? inferScalarsFromConstraints(entity, constraint.type) : entity;
        }
        reportExpectedValue(node, entity);
        return null;
    }
    function reportExpectedValue(target, type) {
        if (type.kind === "Model" && type.name === "" && target.kind === SyntaxKind.ModelExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                messageId: "model",
                format: { name: getTypeName(type) },
                codefixes: [createModelToObjectValueCodeFix(target)],
                target,
            }));
        }
        else if (type.kind === "Tuple" && target.kind === SyntaxKind.TupleExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                messageId: "tuple",
                format: { name: getTypeName(type) },
                codefixes: [createTupleToArrayValueCodeFix(target)],
                target,
            }));
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                format: { name: getTypeName(type) },
                target,
            }));
        }
    }
    /** In certain context for types that can also be value if the constraint allows it we try to use it as a value instead of a type. */
    function getValueFromIndeterminate(type, constraint, node) {
        switch (type.kind) {
            case "String":
            case "StringTemplate":
                return checkStringValue(type, constraint, node);
            case "Number":
                return checkNumericValue(type, constraint, node);
            case "Boolean":
                return checkBooleanValue(type, constraint, node);
            case "EnumMember":
                return checkEnumValue(type, constraint, node);
            case "UnionVariant":
                return getValueFromIndeterminate(type.type, constraint, node);
            case "Intrinsic":
                switch (type.name) {
                    case "null":
                        return checkNullValue(type, constraint, node);
                }
                return type;
            default:
                return type;
        }
    }
    /**
     * Try to intercept types used as values, either return null and emit a diagnostic with a codefix or return the type as is to be dealt with the normal way.
     */
    function interceptTypesUsedAsValue(type) {
        switch (type.kind) {
            case "Tuple":
                return interceptTupleUsedAsValue(type);
            case "Model":
                return interceptModelExpressionUsedAsValue(type);
            default:
                return type;
        }
    }
    // Legacy behavior to smooth transition to object values.
    function interceptModelExpressionUsedAsValue(model) {
        if (model.node?.kind !== SyntaxKind.ModelExpression) {
            return model; // we only want to convert model expressions
        }
        reportCheckerDiagnostic(createDiagnostic({
            code: "expect-value",
            codefixes: [createModelToObjectValueCodeFix(model.node)],
            messageId: "modelExpression",
            target: model.node,
        }));
        return null;
    }
    // Intercept tuple used as value and report diagnostic with a codefix
    function interceptTupleUsedAsValue(tuple) {
        if (tuple.node.kind !== SyntaxKind.TupleExpression) {
            return tuple; // we won't convert dynamic tuples to array values
        }
        reportCheckerDiagnostic(createDiagnostic({
            code: "expect-value",
            codefixes: [createTupleToArrayValueCodeFix(tuple.node)],
            messageId: "tuple",
            target: tuple.node,
        }));
        return null;
    }
    /** If the constraint only expect a value we can try to intercept types passed to it and call `interceptTypeUsedAsValue` to get a better error */
    function shouldTryInterceptTypeUsedAsValue(constraint) {
        return Boolean(constraint?.valueType && !constraint.type);
    }
    /**
     * Gets a type or value depending on the node and current constraint.
     * For nodes that can be both type or values(e.g. string), the value will be returned if the constraint expect a value of that type even if the constrain also allows the type.
     * This means that if the constraint is `string | valueof string` passing `"abc"` will send the value `"abc"` and not the type `"abc"`.
     */
    function getTypeOrValueForNode(node, mapper, constraint) {
        const valueConstraint = extractValueOfConstraints(constraint);
        const entity = checkNode(node, mapper, valueConstraint);
        if (entity === null) {
            return entity;
        }
        else if (isType(entity)) {
            if (shouldTryInterceptTypeUsedAsValue(constraint?.constraint)) {
                return interceptTypesUsedAsValue(entity);
            }
            else {
                return entity;
            }
        }
        else if (isValue(entity)) {
            return entity;
        }
        compilerAssert(entity.entityKind === "Indeterminate", "Expected indeterminate entity");
        if (valueConstraint) {
            const valueDiagnostics = [];
            const oldDiagnosticHook = onCheckerDiagnostic;
            onCheckerDiagnostic = (x) => valueDiagnostics.push(x);
            const result = getValueFromIndeterminate(entity.type, valueConstraint, node);
            onCheckerDiagnostic = oldDiagnosticHook;
            if (result) {
                // If there were diagnostic reported but we still got a value this means that the value might be invalid.
                reportCheckerDiagnostics(valueDiagnostics);
                return result;
            }
        }
        return entity.type;
    }
    /** Extact the type constraint a value should match. */
    function extractValueOfConstraints(constraint) {
        if (constraint?.constraint.valueType) {
            return { kind: constraint.kind, type: constraint.constraint.valueType };
        }
        else {
            return undefined;
        }
    }
    /**
     * Gets a type, value or indeterminate depending on the node and current constraint.
     * For nodes that can be both type or values(e.g. string literals), an indeterminate entity will be returned.
     * It is the job of of the consumer to decide if it should be a type or a value depending on the context.
     */
    function checkNode(node, mapper, valueConstraint) {
        switch (node.kind) {
            case SyntaxKind.ModelExpression:
                return checkModel(node, mapper);
            case SyntaxKind.ModelStatement:
                return checkModel(node, mapper);
            case SyntaxKind.ModelProperty:
                return checkModelProperty(node, mapper);
            case SyntaxKind.ScalarStatement:
                return checkScalar(node, mapper);
            case SyntaxKind.AliasStatement:
                return checkAlias(node, mapper);
            case SyntaxKind.EnumStatement:
                return checkEnum(node, mapper);
            case SyntaxKind.EnumMember:
                return checkEnumMember(node, mapper);
            case SyntaxKind.InterfaceStatement:
                return checkInterface(node, mapper);
            case SyntaxKind.UnionStatement:
                return checkUnion(node, mapper);
            case SyntaxKind.UnionVariant:
                return checkUnionVariant(node, mapper);
            case SyntaxKind.NamespaceStatement:
            case SyntaxKind.JsNamespaceDeclaration:
                return checkNamespace(node);
            case SyntaxKind.OperationStatement:
                return checkOperation(node, mapper);
            case SyntaxKind.NumericLiteral:
                return checkNumericLiteral(node);
            case SyntaxKind.BooleanLiteral:
                return checkBooleanLiteral(node);
            case SyntaxKind.StringLiteral:
                return checkStringLiteral(node);
            case SyntaxKind.TupleExpression:
                return checkTupleExpression(node, mapper);
            case SyntaxKind.StringTemplateExpression:
                return checkStringTemplateExpresion(node, mapper);
            case SyntaxKind.ArrayExpression:
                return checkArrayExpression(node, mapper);
            case SyntaxKind.UnionExpression:
                return checkUnionExpression(node, mapper);
            case SyntaxKind.IntersectionExpression:
                return checkIntersectionExpression(node, mapper);
            case SyntaxKind.DecoratorDeclarationStatement:
                return checkDecoratorDeclaration(node, mapper);
            case SyntaxKind.FunctionDeclarationStatement:
                return checkFunctionDeclaration(node);
            case SyntaxKind.TypeReference:
                return checkTypeOrValueReference(node, mapper);
            case SyntaxKind.TemplateArgument:
                return checkTemplateArgument(node, mapper);
            case SyntaxKind.TemplateParameterDeclaration:
                return checkTemplateParameterDeclaration(node, mapper);
            case SyntaxKind.VoidKeyword:
                return voidType;
            case SyntaxKind.NeverKeyword:
                return neverType;
            case SyntaxKind.UnknownKeyword:
                return unknownType;
            case SyntaxKind.ObjectLiteral:
                return checkObjectValue(node, mapper, valueConstraint);
            case SyntaxKind.ArrayLiteral:
                return checkArrayValue(node, mapper, valueConstraint);
            case SyntaxKind.ConstStatement:
                return checkConst(node);
            case SyntaxKind.CallExpression:
                return checkCallExpression(node, mapper);
            case SyntaxKind.TypeOfExpression:
                return checkTypeOfExpression(node, mapper);
            case SyntaxKind.AugmentDecoratorStatement:
                return checkAugmentDecorator(node);
            case SyntaxKind.UsingStatement:
                return checkUsings(node);
            default:
                return errorType;
        }
    }
    /**
     * Return a fully qualified id of node
     */
    function getNodeSym(node) {
        const symbol = node.kind === SyntaxKind.OperationStatement &&
            node.parent?.kind === SyntaxKind.InterfaceStatement
            ? getSymbolForMember(node)
            : node.symbol;
        return symbol;
    }
    /**
     * Check if the given namespace is the standard library `TypeSpec` namespace.
     */
    function isTypeSpecNamespace(namespace) {
        return namespace.name === "TypeSpec" && namespace.namespace === globalNamespaceType;
    }
    /**
     * Check if the given type is defined right in the TypeSpec namespace.
     */
    function isInTypeSpecNamespace(type) {
        return Boolean(type.namespace && isTypeSpecNamespace(type.namespace));
    }
    function checkTemplateParameterDeclaration(node, mapper) {
        const parentNode = node.parent;
        const grandParentNode = parentNode.parent;
        const links = getSymbolLinks(node.symbol);
        if (!templateParameterUsageMap.has(node)) {
            templateParameterUsageMap.set(node, false);
        }
        if (pendingResolutions.has(getNodeSym(node), ResolutionKind.Constraint)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-constraint",
                    format: { typeName: node.id.sv },
                    target: node.constraint,
                }));
            }
            return errorType;
        }
        let type = links.declaredType;
        if (type === undefined) {
            if (grandParentNode) {
                if (grandParentNode.locals?.has(node.id.sv)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "shadow",
                        format: { name: node.id.sv },
                        target: node,
                    }));
                }
            }
            const index = parentNode.templateParameters.findIndex((v) => v === node);
            type = links.declaredType = createAndFinishType({
                kind: "TemplateParameter",
                node: node,
            });
            if (node.constraint) {
                pendingResolutions.start(getNodeSym(node), ResolutionKind.Constraint);
                type.constraint = getParamConstraintEntityForNode(node.constraint);
                pendingResolutions.finish(getNodeSym(node), ResolutionKind.Constraint);
            }
            if (node.default) {
                type.default = checkTemplateParameterDefault(node.default, parentNode.templateParameters, index, type.constraint);
            }
        }
        return mapper ? mapper.getMappedType(type) : type;
    }
    function getResolvedTypeParameterDefault(declaredType, node, mapper) {
        if (declaredType.default === undefined) {
            return undefined;
        }
        if ((isType(declaredType.default) && isErrorType(declaredType.default)) ||
            declaredType.default === null) {
            return declaredType.default;
        }
        return checkNode(node.default, mapper);
    }
    function checkTemplateParameterDefault(nodeDefault, templateParameters, index, constraint) {
        function visit(node) {
            const entity = checkNode(node);
            let hasError = false;
            if (entity !== null && "kind" in entity && entity.kind === "TemplateParameter") {
                for (let i = index; i < templateParameters.length; i++) {
                    if (entity.node.symbol === templateParameters[i].symbol) {
                        reportCheckerDiagnostic(createDiagnostic({ code: "invalid-template-default", target: node }));
                        return undefined;
                    }
                }
                return entity;
            }
            visitChildren(node, (x) => {
                const visited = visit(x);
                if (visited === undefined) {
                    hasError = true;
                }
            });
            return hasError ? undefined : entity;
        }
        const type = visit(nodeDefault) ?? errorType;
        if (!("kind" in type && isErrorType(type)) && constraint) {
            checkTypeAssignable(type, constraint, nodeDefault);
        }
        return type;
    }
    /**
     * Check and resolve a type for the given type reference node.
     * @param node Node.
     * @param mapper Type mapper for template instantiation context.
     * @param instantiateTemplate If templated type should be instantiated if they haven't yet.
     * @returns Resolved type.
     */
    function checkTypeReference(node, mapper, instantiateTemplate = true) {
        const sym = resolveTypeReferenceSym(node, mapper);
        if (!sym) {
            return errorType;
        }
        const type = checkTypeReferenceSymbol(sym, node, mapper, instantiateTemplate);
        return type;
    }
    /**
     * Check and resolve a type for the given type reference node.
     * @param node Node.
     * @param mapper Type mapper for template instantiation context.
     * @param instantiateTemplate If templated type should be instantiated if they haven't yet.
     * @returns Resolved type.
     */
    function checkTypeOrValueReference(node, mapper, instantiateTemplate = true) {
        const sym = resolveTypeReferenceSym(node, mapper);
        if (!sym) {
            return errorType;
        }
        return checkTypeOrValueReferenceSymbol(sym, node, mapper, instantiateTemplate) ?? errorType;
    }
    function checkTemplateArgument(node, mapper) {
        return checkNode(node.argument, mapper);
    }
    function resolveTypeReference(node) {
        const oldDiagnosticHook = onCheckerDiagnostic;
        const diagnostics = [];
        onCheckerDiagnostic = (x) => diagnostics.push(x);
        const type = checkTypeReference(node, undefined, false);
        onCheckerDiagnostic = oldDiagnosticHook;
        return [type === errorType ? undefined : type, diagnostics];
    }
    function copyDeprecation(sourceType, destType) {
        const deprecationDetails = getDeprecationDetails(program, sourceType);
        if (deprecationDetails) {
            markDeprecated(program, destType, deprecationDetails);
        }
    }
    function checkDeprecated(type, node, target) {
        if (node) {
            const deprecationDetails = getDeprecationDetails(program, node);
            if (deprecationDetails) {
                reportDeprecation(program, target, deprecationDetails.message, reportCheckerDiagnostic);
                return;
            }
        }
        const deprecationDetails = getDeprecationDetails(program, type);
        if (deprecationDetails) {
            reportDeprecation(program, target, deprecationDetails.message, reportCheckerDiagnostic);
        }
    }
    function isTypeReferenceContextDeprecated(node) {
        function checkDeprecatedNode(node) {
            // Perform a simple check if the parent node is deprecated.  We do this
            // out of band because `checkDirectives` usually gets called on the parent
            // type after child types have already been checked (including their
            // deprecations).
            if (!nodeDeprecationMap.has(node)) {
                nodeDeprecationMap.set(node, (node.directives ?? []).findIndex((d) => d.target.sv === "deprecated") >= 0);
            }
            return nodeDeprecationMap.get(node);
        }
        // Walk the parent hierarchy up to a node which might have a
        // deprecation which would mitigate the deprecation warning of the original
        // type reference. This is done to prevent multiple deprecation notices from
        // being raised when a parent context is already being deprecated.
        switch (node.kind) {
            case SyntaxKind.ModelStatement:
                return checkDeprecatedNode(node);
            case SyntaxKind.OperationStatement:
                return (checkDeprecatedNode(node) ||
                    (node.parent.kind === SyntaxKind.InterfaceStatement &&
                        isTypeReferenceContextDeprecated(node.parent)));
            case SyntaxKind.InterfaceStatement:
                return checkDeprecatedNode(node);
            case SyntaxKind.IntersectionExpression:
            case SyntaxKind.UnionExpression:
            case SyntaxKind.ModelProperty:
            case SyntaxKind.OperationSignatureDeclaration:
            case SyntaxKind.OperationSignatureReference:
                return isTypeReferenceContextDeprecated(node.parent);
            default:
                return false;
        }
    }
    function checkTemplateInstantiationArgs(node, args, decls, mapper) {
        const params = new Map();
        const positional = [];
        const initMap = new Map(decls.map((decl) => {
            const declaredType = checkTemplateParameterDeclaration(decl, undefined);
            positional.push(declaredType);
            params.set(decl.id.sv, declaredType);
            return [
                declaredType,
                {
                    decl,
                    checkArgument: null,
                },
            ];
        }));
        let named = false;
        for (const [arg, idx] of args.map((v, i) => [v, i])) {
            function deferredCheck() {
                return [arg, checkNode(arg.argument, mapper)];
            }
            if (arg.name) {
                named = true;
                const param = params.get(arg.name.sv);
                if (!param) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "unknownName",
                        format: {
                            name: arg.name.sv,
                        },
                        target: arg,
                    }));
                    continue;
                }
                if (initMap.get(param).checkArgument !== null) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "specifiedAgain",
                        format: {
                            name: arg.name.sv,
                        },
                        target: arg,
                    }));
                    continue;
                }
                initMap.get(param).checkArgument = deferredCheck;
            }
            else {
                if (named) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "positionalAfterNamed",
                        target: arg,
                    }));
                    // we just throw this arg away. any missing args will be filled with ErrorType
                }
                if (idx >= positional.length) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "tooMany",
                        target: node,
                    }));
                    continue;
                }
                const param = positional[idx];
                initMap.get(param).checkArgument ??= deferredCheck;
            }
        }
        const finalMap = initMap;
        const mapperParams = [];
        const mapperArgs = [];
        for (const [param, { decl, checkArgument: init }] of [...initMap]) {
            function commit(param, type) {
                finalMap.set(param, type);
                mapperParams.push(param);
                mapperArgs.push(type);
            }
            if (init === null) {
                const argumentMapper = createTypeMapper(mapperParams, mapperArgs, { node, mapper });
                const defaultValue = getResolvedTypeParameterDefault(param, decl, argumentMapper);
                if (defaultValue) {
                    commit(param, defaultValue);
                }
                else {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "missing",
                        format: {
                            name: decl.id.sv,
                        },
                        target: node,
                    }));
                    // TODO-TIM check if we expose this below
                    commit(param, param.constraint?.type ?? unknownType);
                }
                continue;
            }
            const [argNode, type] = init();
            if (type === null) {
                commit(param, unknownType);
                continue;
            }
            if (param.constraint) {
                const constraint = param.constraint.type?.kind === "TemplateParameter"
                    ? finalMap.get(param.constraint.type)
                    : param.constraint;
                if (isType(type) && shouldTryInterceptTypeUsedAsValue(param.constraint)) {
                    const converted = interceptTypesUsedAsValue(type);
                    // If we manage to convert it means this might be convertable so we skip type checking.
                    // However we still return the original entity
                    if (converted !== type) {
                        commit(param, type);
                        continue;
                    }
                }
                if (param.constraint && !checkArgumentAssignable(type, constraint, argNode)) {
                    const effectiveType = param.constraint.type ?? unknownType;
                    commit(param, effectiveType);
                    continue;
                }
            }
            else if (isErrorType(type)) {
                // If we got an error type we don't want to keep passing it through so we reduce to unknown
                // Similar to the above where if the type is not assignable to the constraint we reduce to the constraint
                commit(param, unknownType);
                continue;
            }
            else if (isValue(type)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "value-in-type",
                    messageId: "noTemplateConstraint",
                    target: argNode,
                }));
                commit(param, unknownType);
                continue;
            }
            commit(param, type);
        }
        return finalMap;
    }
    /**
     * Check and resolve the type for the given symbol + node.
     * @param sym Symbol
     * @param node Node
     * @param mapper Type mapper for template instantiation context.
     * @param instantiateTemplates If a templated type should be instantiated if not yet @default true
     * @returns resolved type.
     */
    function checkTypeReferenceSymbol(sym, node, mapper, instantiateTemplates = true) {
        const result = checkTypeOrValueReferenceSymbol(sym, node, mapper, instantiateTemplates);
        if (result === null || isValue(result)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "value-in-type", target: node }));
            return errorType;
        }
        if (result.entityKind === "Indeterminate") {
            return result.type;
        }
        return result;
    }
    function checkTypeOrValueReferenceSymbol(sym, node, mapper, instantiateTemplates = true) {
        const entity = checkTypeOrValueReferenceSymbolWorker(sym, node, mapper, instantiateTemplates);
        if (entity !== null && isType(entity) && entity.kind === "TemplateParameter") {
            templateParameterUsageMap.set(entity.node, true);
        }
        return entity;
    }
    function checkTypeOrValueReferenceSymbolWorker(sym, node, mapper, instantiateTemplates = true) {
        if (sym.flags & 131072 /* SymbolFlags.Const */) {
            return getValueForNode(sym.declarations[0], mapper);
        }
        if (sym.flags & 512 /* SymbolFlags.Decorator */) {
            reportCheckerDiagnostic(createDiagnostic({ code: "invalid-type-ref", messageId: "decorator", target: sym }));
            return errorType;
        }
        if (sym.flags & 2048 /* SymbolFlags.Function */) {
            reportCheckerDiagnostic(createDiagnostic({ code: "invalid-type-ref", messageId: "function", target: sym }));
            return errorType;
        }
        const argumentNodes = node.kind === SyntaxKind.TypeReference ? node.arguments : [];
        const symbolLinks = getSymbolLinks(sym);
        let baseType;
        if (sym.flags & 1048576 /* SymbolFlags.Declaration */ &&
            sym.flags &
                (2 /* SymbolFlags.Model */ |
                    4 /* SymbolFlags.Scalar */ |
                    128 /* SymbolFlags.Alias */ |
                    32 /* SymbolFlags.Interface */ |
                    8 /* SymbolFlags.Operation */ |
                    64 /* SymbolFlags.Union */)) {
            const decl = sym.declarations[0];
            if (!isTemplatedNode(decl)) {
                if (argumentNodes.length > 0) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "notTemplate",
                        target: node,
                    }));
                }
                if (sym.flags & 4194304 /* SymbolFlags.LateBound */) {
                    compilerAssert(sym.type, "Expected late bound symbol to have type");
                    return sym.type;
                }
                else if (symbolLinks.declaredType) {
                    baseType = symbolLinks.declaredType;
                }
                else if (sym.flags & 65536 /* SymbolFlags.Member */) {
                    baseType = checkMemberSym(sym, mapper);
                }
                else {
                    baseType = checkDeclaredTypeOrIndeterminate(sym, decl, mapper);
                }
            }
            else {
                const declaredType = getOrCheckDeclaredType(sym, decl, mapper);
                const templateParameters = decl.templateParameters;
                const instantiation = checkTemplateInstantiationArgs(node, argumentNodes, templateParameters, mapper);
                baseType = getOrInstantiateTemplate(decl, [...instantiation.keys()], [...instantiation.values()], { node, mapper }, declaredType.templateMapper, instantiateTemplates);
            }
        }
        else {
            const symNode = getSymNode(sym);
            // some other kind of reference
            if (argumentNodes.length > 0) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-template-args",
                    messageId: "notTemplate",
                    target: node,
                }));
            }
            if (sym.flags & 4194304 /* SymbolFlags.LateBound */) {
                compilerAssert(sym.type, `Expected late bound symbol to have type`);
                return sym.type;
            }
            else if (sym.flags & 1024 /* SymbolFlags.TemplateParameter */) {
                const mapped = checkTemplateParameterDeclaration(symNode, mapper);
                baseType = mapped;
            }
            else if (symbolLinks.type) {
                // Have a cached type for non-declarations
                baseType = symbolLinks.type;
            }
            else if (symbolLinks.declaredType) {
                baseType = symbolLinks.declaredType;
            }
            else {
                if (sym.flags & 65536 /* SymbolFlags.Member */) {
                    baseType = checkMemberSym(sym, mapper);
                }
                else {
                    // don't have a cached type for this symbol, so go grab it and cache it
                    baseType = getTypeForNode(symNode, mapper);
                    symbolLinks.type = baseType;
                }
            }
        }
        // Check for deprecations here, first on symbol, then on type.  However,
        // don't raise deprecation when the usage site is also a deprecated
        // declaration.
        const declarationNode = getSymNode(sym);
        if (declarationNode && mapper === undefined && isType(baseType)) {
            if (!isTypeReferenceContextDeprecated(node.parent)) {
                checkDeprecated(baseType, declarationNode, node);
            }
        }
        // Elements that could be used as type or values depending on the context
        if ((isType(baseType) && (baseType.kind === "EnumMember" || baseType.kind === "UnionVariant")) ||
            isNullType(baseType)) {
            return createIndeterminateEntity(baseType);
        }
        return baseType;
    }
    /**
     * Get or check the declared type of a templatable node.
     * @param node Declaration node
     * @param sym Node Symbol
     * @param mapper Type mapper for template resolution
     * @returns The declared type for the given node.
     */
    function getOrCheckDeclaredType(sym, decl, mapper) {
        const symbolLinks = getSymbolLinks(sym);
        if (symbolLinks.declaredType) {
            return symbolLinks.declaredType;
        }
        if (sym.flags & 4194304 /* SymbolFlags.LateBound */) {
            compilerAssert(sym.type, "Expected late bound symbol to have type");
            return sym.type;
        }
        if (sym.flags & 65536 /* SymbolFlags.Member */) {
            return checkMemberSym(sym, mapper);
        }
        else {
            return checkDeclaredType(sym, decl, mapper);
        }
    }
    /**
     * Check the declared type of a templatable node.
     * @param node Declaration node
     * @param sym Node Symbol
     * @param mapper Type mapper for template resolution
     * @returns The declared type for the given node.
     */
    function checkDeclaredTypeOrIndeterminate(sym, node, mapper) {
        const type = sym.flags & 2 /* SymbolFlags.Model */
            ? checkModelStatement(node, mapper)
            : sym.flags & 4 /* SymbolFlags.Scalar */
                ? checkScalar(node, mapper)
                : sym.flags & 128 /* SymbolFlags.Alias */
                    ? checkAlias(node, mapper)
                    : sym.flags & 32 /* SymbolFlags.Interface */
                        ? checkInterface(node, mapper)
                        : sym.flags & 8 /* SymbolFlags.Operation */
                            ? checkOperation(node, mapper)
                            : checkUnion(node, mapper);
        return type;
    }
    function checkDeclaredType(sym, node, mapper) {
        return getTypeForTypeOrIndeterminate(checkDeclaredTypeOrIndeterminate(sym, node, mapper));
    }
    function getOrInstantiateTemplate(templateNode, params, args, source, parentMapper, instantiateTempalates = true) {
        const symbolLinks = templateNode.kind === SyntaxKind.OperationStatement &&
            templateNode.parent.kind === SyntaxKind.InterfaceStatement
            ? getSymbolLinksForMember(templateNode)
            : getSymbolLinks(templateNode.symbol);
        compilerAssert(symbolLinks, `Unexpected checker error. symbolLinks was not defined for ${SyntaxKind[templateNode.kind]}`);
        if (symbolLinks.instantiations === undefined) {
            const type = getTypeForNode(templateNode);
            if (isErrorType(type)) {
                return errorType;
            }
            else {
                compilerAssert(false, `Unexpected checker error. symbolLinks.instantiations was not defined for ${SyntaxKind[templateNode.kind]}`);
            }
        }
        const mapper = createTypeMapper(params, args, source, parentMapper);
        const cached = symbolLinks.instantiations?.get(mapper.args);
        if (cached) {
            return cached;
        }
        if (instantiateTempalates) {
            return instantiateTemplate(symbolLinks.instantiations, templateNode, params, mapper);
        }
        else {
            return errorType;
        }
    }
    /**
     * Builds a model type from a template and its template arguments.
     * Adds the template node to a set we can check when we bind template
     * parameters to access type type arguments.
     *
     * This will fall over if the same template is ever being instantiated
     * twice at the same time, or if template parameters from more than one template
     * are ever in scope at once.
     */
    function instantiateTemplate(instantiations, templateNode, params, mapper) {
        const type = getTypeForNode(templateNode, mapper);
        if (!instantiations.get(mapper.args)) {
            instantiations.set(mapper.args, type);
        }
        if (type.kind === "Model") {
            type.templateNode = templateNode;
        }
        return type;
    }
    /** Check a union expresion used in a parameter constraint, those allow the use of `valueof` as a variant. */
    function checkMixedParameterConstraintUnion(node, mapper) {
        const values = [];
        const types = [];
        for (const option of node.options) {
            const [kind, type] = getTypeOrValueOfTypeForNode(option, mapper);
            if (kind === "value") {
                values.push(type);
            }
            else {
                types.push(type);
            }
        }
        return {
            entityKind: "MixedParameterConstraint",
            node,
            valueType: values.length === 0
                ? undefined
                : values.length === 1
                    ? values[0]
                    : createConstraintUnion(node, values),
            type: types.length === 0
                ? undefined
                : types.length === 1
                    ? types[0]
                    : createConstraintUnion(node, types),
        };
    }
    function createConstraintUnion(node, options) {
        const variants = createRekeyableMap();
        const union = createAndFinishType({
            kind: "Union",
            node,
            options,
            decorators: [],
            variants,
            expression: true,
        });
        for (const option of options) {
            const name = Symbol("indexer-union-variant");
            variants.set(name, createAndFinishType({
                kind: "UnionVariant",
                node: undefined,
                type: option,
                name,
                union,
                decorators: [],
            }));
        }
        return union;
    }
    function checkUnionExpression(node, mapper) {
        const unionType = createAndFinishType({
            kind: "Union",
            node,
            get options() {
                return Array.from(this.variants.values()).map((v) => v.type);
            },
            expression: true,
            variants: createRekeyableMap(),
            decorators: [],
        });
        for (const o of node.options) {
            const type = getTypeForNode(o, mapper);
            // The type `A | never` is just `A`
            if (type === neverType) {
                continue;
            }
            if (type.kind === "Union" && type.expression) {
                for (const [name, variant] of type.variants) {
                    unionType.variants.set(name, variant);
                }
            }
            else {
                const variant = createType({
                    kind: "UnionVariant",
                    type,
                    name: Symbol("name"),
                    decorators: [],
                    node: undefined,
                    union: unionType,
                });
                unionType.variants.set(variant.name, variant);
            }
        }
        linkMapper(unionType, mapper);
        return unionType;
    }
    /**
     * Intersection produces a model type from the properties of its operands.
     * So this doesn't work if we don't have a known set of properties (e.g.
     * with unions). The resulting model is anonymous.
     */
    function checkIntersectionExpression(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        const options = node.options.map((o) => [o, getTypeForNode(o, mapper)]);
        const type = mergeModelTypes(node.symbol, node, options, mapper);
        linkType(links, type, mapper);
        return type;
    }
    function checkDecoratorDeclaration(node, mapper) {
        const symbol = getMergedSymbol(node.symbol);
        const links = getSymbolLinks(symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this operation and we've already checked it
            return links.declaredType;
        }
        const namespace = getParentNamespaceType(node);
        compilerAssert(namespace, `Decorator ${node.id.sv} should have resolved a namespace or found the global namespace.`);
        const name = node.id.sv;
        if (!(node.modifierFlags & 2 /* ModifierFlags.Extern */)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "decorator-extern", target: node }));
        }
        const implementation = symbol.value;
        if (implementation === undefined) {
            reportCheckerDiagnostic(createDiagnostic({ code: "missing-implementation", target: node }));
        }
        const decoratorType = createType({
            kind: "Decorator",
            name: `@${name}`,
            namespace,
            node,
            target: checkFunctionParameter(node.target, mapper, true),
            parameters: node.parameters.map((x) => checkFunctionParameter(x, mapper, true)),
            implementation: implementation ?? (() => { }),
        });
        namespace.decoratorDeclarations.set(name, decoratorType);
        linkType(links, decoratorType, mapper);
        return decoratorType;
    }
    function checkFunctionDeclaration(node, mapper) {
        reportCheckerDiagnostic(createDiagnostic({ code: "function-unsupported", target: node }));
        return errorType;
    }
    function checkFunctionParameter(node, mapper, mixed) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType) {
            return links.declaredType;
        }
        if (node.rest &&
            node.type &&
            !(node.type.kind === SyntaxKind.ArrayExpression ||
                (node.type.kind === SyntaxKind.ValueOfExpression &&
                    node.type.target.kind === SyntaxKind.ArrayExpression))) {
            reportCheckerDiagnostic(createDiagnostic({ code: "rest-parameter-array", target: node.type }));
        }
        const base = {
            kind: "FunctionParameter",
            node,
            name: node.id.sv,
            optional: node.optional,
            rest: node.rest,
            implementation: node.symbol.value,
        };
        let parameterType;
        if (mixed) {
            const type = node.type
                ? getParamConstraintEntityForNode(node.type)
                : {
                    entityKind: "MixedParameterConstraint",
                    type: unknownType,
                };
            parameterType = createType({
                ...base,
                type,
                mixed: true,
                implementation: node.symbol.value,
            });
        }
        else {
            parameterType = createType({
                ...base,
                mixed: false,
                type: node.type ? getTypeForNode(node.type) : unknownType,
                implementation: node.symbol.value,
            });
        }
        linkType(links, parameterType, mapper);
        return parameterType;
    }
    function getTypeOrValueOfTypeForNode(node, mapper) {
        switch (node.kind) {
            case SyntaxKind.ValueOfExpression:
                const target = getTypeForNode(node.target, mapper);
                return ["value", target];
            default:
                return ["type", getTypeForNode(node, mapper)];
        }
    }
    function getParamConstraintEntityForNode(node, mapper) {
        switch (node.kind) {
            case SyntaxKind.UnionExpression:
                return checkMixedParameterConstraintUnion(node, mapper);
            default:
                const [kind, entity] = getTypeOrValueOfTypeForNode(node, mapper);
                return {
                    entityKind: "MixedParameterConstraint",
                    node: node,
                    type: kind === "value" ? undefined : entity,
                    valueType: kind === "value" ? entity : undefined,
                };
        }
    }
    function mergeModelTypes(parentModelSym, node, options, mapper) {
        const properties = createRekeyableMap();
        const intersection = createType({
            kind: "Model",
            node,
            name: "",
            namespace: getParentNamespaceType(node),
            properties: properties,
            decorators: [],
            derivedModels: [],
            sourceModels: [],
        });
        const indexers = [];
        const modelOptions = options.filter((entry) => {
            const [optionNode, option] = entry;
            if (option.kind === "TemplateParameter") {
                return false;
            }
            if (option.kind !== "Model") {
                reportCheckerDiagnostic(createDiagnostic({ code: "intersect-non-model", target: optionNode }));
                return false;
            }
            return true;
        });
        for (const [optionNode, option] of modelOptions) {
            if (option.indexer) {
                if (option.indexer.key.name === "integer") {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "intersect-invalid-index",
                        messageId: "array",
                        target: optionNode,
                    }));
                }
                else {
                    indexers.push(option.indexer);
                }
            }
        }
        for (const [_, option] of modelOptions) {
            intersection.sourceModels.push({ usage: "intersection", model: option });
            const allProps = walkPropertiesInherited(option);
            for (const prop of allProps) {
                if (properties.has(prop.name)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "intersect-duplicate-property",
                        format: { propName: prop.name },
                        target: node,
                    }));
                    continue;
                }
                const memberSym = parentModelSym && getMemberSymbol(parentModelSym, prop.name);
                const overrides = {
                    sourceProperty: prop,
                    model: intersection,
                };
                const newPropType = memberSym
                    ? cloneTypeForSymbol(memberSym, prop, overrides)
                    : cloneType(prop, overrides);
                properties.set(prop.name, newPropType);
                linkIndirectMember(node, newPropType, mapper);
                for (const indexer of indexers.filter((x) => x !== option.indexer)) {
                    checkPropertyCompatibleWithIndexer(indexer, prop, node);
                }
            }
        }
        if (indexers.length === 1) {
            intersection.indexer = indexers[0];
        }
        else if (indexers.length > 1) {
            intersection.indexer = {
                key: indexers[0].key,
                value: mergeModelTypes(undefined, node, indexers.map((x) => [x.value.node, x.value]), mapper),
            };
        }
        linkMapper(intersection, mapper);
        return finishType(intersection);
    }
    function checkArrayExpression(node, mapper) {
        const elementType = getTypeForNode(node.elementType, mapper);
        const arrayType = getStdType("Array");
        const arrayNode = arrayType.node;
        const param = getTypeForNode(arrayNode.templateParameters[0]);
        return getOrInstantiateTemplate(arrayNode, [param], [elementType], { node, mapper }, undefined);
    }
    function checkNamespace(node) {
        const links = getSymbolLinks(getMergedSymbol(node.symbol));
        let type = links.type;
        if (!type) {
            type = initializeTypeForNamespace(node);
        }
        if (node.kind === SyntaxKind.NamespaceStatement) {
            if (isArray(node.statements)) {
                node.statements.forEach((x) => checkNode(x));
            }
            else if (node.statements) {
                const subNs = checkNamespace(node.statements);
                type.namespaces.set(subNs.name, subNs);
            }
        }
        return type;
    }
    function initializeTypeForNamespace(node) {
        compilerAssert(node.symbol, "Namespace is unbound.", node);
        const mergedSymbol = getMergedSymbol(node.symbol);
        const symbolLinks = getSymbolLinks(mergedSymbol);
        if (!symbolLinks.type) {
            // haven't seen this namespace before
            const namespace = getParentNamespaceType(node);
            const name = node.id.sv;
            const type = createType({
                kind: "Namespace",
                name,
                namespace,
                node: node,
                models: new Map(),
                scalars: new Map(),
                operations: new Map(),
                namespaces: new Map(),
                interfaces: new Map(),
                unions: new Map(),
                enums: new Map(),
                decoratorDeclarations: new Map(),
                functionDeclarations: new Map(),
                decorators: [],
            });
            symbolLinks.type = type;
            for (const sourceNode of mergedSymbol.declarations) {
                // namespaces created from TypeSpec scripts don't have decorators
                if (sourceNode.kind !== SyntaxKind.NamespaceStatement)
                    continue;
                type.decorators = type.decorators.concat(checkDecorators(type, sourceNode, undefined));
            }
            finishType(type);
            namespace?.namespaces.set(name, type);
        }
        return symbolLinks.type;
    }
    function getParentNamespaceType(node) {
        if (node === globalNamespaceType.node)
            return undefined;
        if (node.kind === SyntaxKind.ModelExpression ||
            node.kind === SyntaxKind.IntersectionExpression) {
            let parent = node.parent;
            while (parent !== undefined) {
                if (parent.kind === SyntaxKind.AliasStatement ||
                    parent.kind === SyntaxKind.ModelStatement ||
                    parent.kind === SyntaxKind.ScalarStatement ||
                    parent.kind === SyntaxKind.OperationStatement ||
                    parent.kind === SyntaxKind.EnumStatement ||
                    parent.kind === SyntaxKind.InterfaceStatement ||
                    parent.kind === SyntaxKind.UnionStatement ||
                    parent.kind === SyntaxKind.ModelExpression ||
                    parent.kind === SyntaxKind.IntersectionExpression) {
                    return getParentNamespaceType(parent);
                }
                else {
                    parent = parent.parent;
                }
            }
            return undefined;
        }
        if (node.kind === SyntaxKind.OperationStatement &&
            node.parent &&
            node.parent.kind === SyntaxKind.InterfaceStatement) {
            return getParentNamespaceType(node.parent);
        }
        if (!node.symbol.parent) {
            return globalNamespaceType;
        }
        if (node.symbol.parent.declarations[0].kind === SyntaxKind.TypeSpecScript ||
            node.symbol.parent.declarations[0].kind === SyntaxKind.JsSourceFile) {
            return globalNamespaceType;
        }
        const mergedSymbol = getMergedSymbol(node.symbol.parent);
        const symbolLinks = getSymbolLinks(mergedSymbol);
        if (!symbolLinks.type) {
            // in general namespaces should be typed before anything calls this function.
            // However, one case where this is not true is when a decorator on a namespace
            // refers to a model in another namespace. In this case, we need to evaluate
            // the namespace here.
            const namespaceNode = mergedSymbol.declarations.find((x) => x.kind === SyntaxKind.NamespaceStatement || x.kind === SyntaxKind.JsNamespaceDeclaration);
            compilerAssert(namespaceNode, "Can't find namespace declaration node.", node);
            symbolLinks.type = initializeTypeForNamespace(namespaceNode);
        }
        return symbolLinks.type;
    }
    function checkOperation(node, mapper, parentInterface) {
        const inInterface = node.parent?.kind === SyntaxKind.InterfaceStatement;
        const symbol = inInterface ? getSymbolForMember(node) : node.symbol;
        const links = symbol && getSymbolLinks(symbol);
        if (links) {
            if (links.declaredType && mapper === undefined) {
                // we're not instantiating this operation and we've already checked it
                return links.declaredType;
            }
        }
        if (mapper === undefined && inInterface) {
            compilerAssert(parentInterface, "Operation in interface should already have been checked.", node.parent);
        }
        checkTemplateDeclaration(node, mapper);
        // If we are instantating operation inside of interface
        if (isTemplatedNode(node) && mapper !== undefined && parentInterface) {
            mapper = { ...mapper, partial: true };
        }
        const namespace = getParentNamespaceType(node);
        const name = node.id.sv;
        let decorators = [];
        const { resolvedSymbol: parameterModelSym } = resolver.resolveMetaMemberByName(symbol, "parameters");
        if (parameterModelSym?.members) {
            const members = resolver.getAugmentedSymbolTable(parameterModelSym.members);
            const paramDocs = extractParamDocs(node);
            for (const [name, memberSym] of members) {
                const doc = paramDocs.get(name);
                if (doc) {
                    docFromCommentForSym.set(memberSym, doc);
                }
            }
        }
        // Is this a definition or reference?
        let parameters, returnType, sourceOperation;
        if (node.signature.kind === SyntaxKind.OperationSignatureReference) {
            // Attempt to resolve the operation
            const baseOperation = checkOperationIs(node, node.signature.baseOperation, mapper);
            if (baseOperation) {
                sourceOperation = baseOperation;
                // Reference the same return type and create the parameters type
                const clone = initializeClone(baseOperation.parameters, {
                    properties: createRekeyableMap(),
                });
                clone.properties = createRekeyableMap(Array.from(baseOperation.parameters.properties.entries()).map(([key, prop]) => [
                    key,
                    cloneTypeForSymbol(getMemberSymbol(parameterModelSym, prop.name), prop, {
                        model: clone,
                        sourceProperty: prop,
                    }),
                ]));
                parameters = finishType(clone);
                returnType = baseOperation.returnType;
                // Copy decorators from the base operation, inserting the base decorators first
                decorators = [...baseOperation.decorators];
            }
            else {
                // If we can't resolve the signature we return an empty model.
                parameters = createAndFinishType({
                    kind: "Model",
                    name: "",
                    decorators: [],
                    properties: createRekeyableMap(),
                    derivedModels: [],
                    sourceModels: [],
                });
                returnType = voidType;
            }
        }
        else {
            parameters = getTypeForNode(node.signature.parameters, mapper);
            returnType = getTypeForNode(node.signature.returnType, mapper);
        }
        const operationType = createType({
            kind: "Operation",
            name,
            namespace,
            node,
            parameters,
            returnType,
            decorators,
            sourceOperation,
            interface: parentInterface,
        });
        if (links) {
            linkType(links, operationType, mapper);
        }
        decorators.push(...checkDecorators(operationType, node, mapper));
        operationType.parameters.namespace = namespace;
        const parent = node.parent;
        linkMapper(operationType, mapper);
        if (parent.kind === SyntaxKind.InterfaceStatement) {
            if (shouldCreateTypeForTemplate(parent, mapper) &&
                shouldCreateTypeForTemplate(node, mapper)) {
                finishType(operationType);
            }
        }
        else {
            if (shouldCreateTypeForTemplate(node, mapper)) {
                finishType(operationType);
            }
            if (mapper === undefined) {
                namespace?.operations.set(name, operationType);
            }
        }
        return operationType;
    }
    function checkOperationIs(operation, opReference, mapper) {
        if (!opReference)
            return undefined;
        // Ensure that we don't end up with a circular reference to the same operation
        const opSymId = getNodeSym(operation);
        if (opSymId) {
            pendingResolutions.start(opSymId, ResolutionKind.BaseType);
        }
        const target = resolver.getNodeLinks(opReference).resolvedSymbol;
        // Did we encounter a circular operation reference?
        if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-op-signature",
                    format: { typeName: target.name },
                    target: opReference,
                }));
            }
            return undefined;
        }
        // Resolve the base operation type
        const baseOperation = getTypeForNode(opReference, mapper);
        if (opSymId) {
            pendingResolutions.finish(opSymId, ResolutionKind.BaseType);
        }
        if (isErrorType(baseOperation)) {
            return undefined;
        }
        // Was the wrong type referenced?
        if (baseOperation.kind !== "Operation") {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-operation", target: opReference }));
            return;
        }
        return baseOperation;
    }
    function getGlobalNamespaceType() {
        return globalNamespaceType;
    }
    function getGlobalNamespaceNode() {
        return resolver.symbols.global.declarations[0];
    }
    function checkTupleExpression(node, mapper) {
        return createAndFinishType({
            kind: "Tuple",
            node: node,
            values: node.values.map((v) => getTypeForNode(v, mapper)),
        });
    }
    function getSymbolLinks(s) {
        return resolver.getSymbolLinks(s);
    }
    function resolveRelatedSymbols(id, mapper) {
        let sym;
        const { node, kind } = getIdentifierContext(id);
        switch (kind) {
            case IdentifierKind.ModelExpressionProperty:
            case IdentifierKind.ObjectLiteralProperty:
                const model = getReferencedModel(node);
                return model
                    .map((m) => getMemberSymbol(m.node.symbol, id.sv))
                    .filter((m) => m !== undefined);
            case IdentifierKind.ModelStatementProperty:
            case IdentifierKind.Declaration:
                const links = resolver.getNodeLinks(id);
                return links.resolvedSymbol === undefined ? undefined : [links.resolvedSymbol];
            case IdentifierKind.Other:
                return undefined;
            case IdentifierKind.Decorator:
            case IdentifierKind.Function:
            case IdentifierKind.Using:
            case IdentifierKind.TypeReference:
                let ref = id;
                let resolveDecorator = kind === IdentifierKind.Decorator;
                if (id.parent?.kind === SyntaxKind.MemberExpression) {
                    if (id.parent.id === id) {
                        // If the identifier is Y in X.Y, then resolve (X.Y).
                        ref = id.parent;
                    }
                    else {
                        // If the identifier is X in X.Y then we are resolving a
                        // namespace, which is never a decorator.
                        resolveDecorator = false;
                    }
                }
                sym = resolveTypeReferenceSym(ref, mapper, resolveDecorator);
                break;
            case IdentifierKind.TemplateArgument:
                const templates = getTemplateDeclarationsForArgument(node, mapper);
                const firstMatchingParameter = templates
                    .flatMap((t) => t.templateParameters)
                    .find((p) => p.id.sv === id.sv);
                if (firstMatchingParameter) {
                    sym = getMergedSymbol(firstMatchingParameter.symbol);
                }
                break;
            default:
                compilerAssert(false, "Unreachable");
        }
        if (sym) {
            if (sym.symbolSource) {
                return [sym.symbolSource];
            }
            else {
                return [sym];
            }
        }
        return undefined; //sym?.symbolSource ?? sym;
    }
    function getTemplateDeclarationsForArgument(node, mapper) {
        const ref = node.parent;
        let resolved = resolveTypeReferenceSym(ref, mapper, false);
        // if the reference type can't be resolved and has parse error,
        // it likely means the reference type hasn't been completed yet. i.e. Foo<string,
        // so try to resolve it by it's target directly to see if we can find its sym
        if (!resolved && hasParseError(ref) && ref.target !== undefined) {
            resolved = resolveTypeReferenceSym(ref.target, mapper, false);
        }
        return (resolved?.declarations.filter((n) => isTemplatedNode(n)) ?? []);
    }
    function getReferencedModel(propertyNode) {
        const isModelOrArrayValue = (n) => n?.kind === SyntaxKind.ArrayLiteral || n?.kind === SyntaxKind.ObjectLiteral;
        const isModelOrArrayType = (n) => n?.kind === SyntaxKind.ModelExpression || n?.kind === SyntaxKind.TupleExpression;
        const isModelOrArray = (n) => isModelOrArrayValue(n) || isModelOrArrayType(n);
        const path = [];
        let preNode;
        const foundNode = getFirstAncestor(propertyNode, (n) => {
            pushToModelPath(n, preNode, path);
            preNode = n;
            return ((isModelOrArray(n) &&
                (n.parent?.kind === SyntaxKind.TemplateParameterDeclaration ||
                    n.parent?.kind === SyntaxKind.TemplateArgument ||
                    n.parent?.kind === SyntaxKind.DecoratorExpression)) ||
                (isModelOrArrayValue(n) &&
                    (n.parent?.kind === SyntaxKind.CallExpression ||
                        n.parent?.kind === SyntaxKind.ConstStatement)));
        });
        let refType;
        switch (foundNode?.parent?.kind) {
            case SyntaxKind.TemplateParameterDeclaration:
            case SyntaxKind.TemplateArgument:
                refType = getReferencedTypeFromTemplateDeclaration(foundNode);
                break;
            case SyntaxKind.DecoratorExpression:
                refType = getReferencedTypeFromDecoratorArgument(foundNode);
                break;
            case SyntaxKind.CallExpression:
                refType = getReferencedTypeFromScalarConstructor(foundNode);
                break;
            case SyntaxKind.ConstStatement:
                refType = getReferencedTypeFromConstAssignment(foundNode);
                break;
        }
        return getNestedModel(refType, path);
        function pushToModelPath(node, preNode, path) {
            if (node.kind === SyntaxKind.ArrayLiteral || node.kind === SyntaxKind.TupleExpression) {
                const index = node.values.findIndex((n) => n === preNode);
                if (index >= 0) {
                    path.unshift({ tupleIndex: index });
                }
                else {
                    compilerAssert(false, "not expected, can't find child from the parent?");
                }
            }
            if (node.kind === SyntaxKind.ModelProperty ||
                node.kind === SyntaxKind.ObjectLiteralProperty) {
                path.unshift({ propertyName: node.id.sv });
            }
        }
        function getNestedModel(modelOrTupleOrUnion, path) {
            let cur = modelOrTupleOrUnion;
            if (cur && cur.kind !== "Model" && cur.kind !== "Tuple" && cur.kind !== "Union") {
                return [];
            }
            if (path.length === 0) {
                // Handle union and model type nesting when path is empty
                switch (cur?.kind) {
                    case "Model":
                        return [cur];
                    case "Union":
                        const models = [];
                        for (const variant of cur.variants.values()) {
                            if (variant.type.kind === "Model" ||
                                variant.type.kind === "Tuple" ||
                                variant.type.kind === "Union") {
                                models.push(...(getNestedModel(variant.type, path) ?? []));
                            }
                        }
                        return models;
                    default:
                        return [];
                }
            }
            const seg = path[0];
            switch (cur?.kind) {
                case "Tuple":
                    if (seg.tupleIndex !== undefined &&
                        seg.tupleIndex >= 0 &&
                        seg.tupleIndex < cur.values.length) {
                        return getNestedModel(cur.values[seg.tupleIndex], path.slice(1));
                    }
                    else {
                        return [];
                    }
                case "Model":
                    if (cur.name === "Array" && seg.tupleIndex !== undefined) {
                        cur = cur.templateMapper?.args[0];
                    }
                    else if (cur.name !== "Array" && seg.propertyName) {
                        cur = cur.properties.get(seg.propertyName)?.type;
                    }
                    else {
                        return [];
                    }
                    return getNestedModel(cur, path.slice(1));
                case "Union":
                    // When seg.property name exists, it means that it is in the union model or tuple,
                    // and the corresponding model or tuple needs to be found recursively.
                    const models = [];
                    for (const variant of cur.variants.values()) {
                        if (variant.type.kind === "Model" ||
                            variant.type.kind === "Tuple" ||
                            variant.type.kind === "Union") {
                            models.push(...(getNestedModel(variant.type, path) ?? []));
                        }
                    }
                    return models;
                default:
                    return [];
            }
        }
        function getReferencedTypeFromTemplateDeclaration(node) {
            let templateParmaeterDeclNode = undefined;
            if (node?.parent?.kind === SyntaxKind.TemplateArgument &&
                node?.parent?.parent?.kind === SyntaxKind.TypeReference) {
                const argNode = node.parent;
                const refNode = node.parent.parent;
                const decl = getTemplateDeclarationsForArgument(argNode, 
                // We should be giving the argument so the mapper here should be undefined
                undefined /* mapper */);
                const index = refNode.arguments.findIndex((n) => n === argNode);
                if (decl.length > 0 && decl[0].templateParameters.length > index) {
                    templateParmaeterDeclNode = decl[0].templateParameters[index];
                }
            }
            else if (node.parent?.kind === SyntaxKind.TemplateParameterDeclaration) {
                templateParmaeterDeclNode = node?.parent;
            }
            if (templateParmaeterDeclNode?.kind !== SyntaxKind.TemplateParameterDeclaration ||
                !templateParmaeterDeclNode.constraint) {
                return undefined;
            }
            let constraintType;
            if (isModelOrArrayValue(node) &&
                templateParmaeterDeclNode.constraint.kind === SyntaxKind.ValueOfExpression) {
                constraintType = program.checker.getTypeForNode(templateParmaeterDeclNode.constraint.target);
            }
            else if (isModelOrArrayType(node) &&
                templateParmaeterDeclNode.constraint.kind !== SyntaxKind.ValueOfExpression) {
                constraintType = program.checker.getTypeForNode(templateParmaeterDeclNode.constraint);
            }
            return constraintType;
        }
        function getReferencedTypeFromScalarConstructor(argNode) {
            const callExpNode = argNode?.parent;
            if (callExpNode?.kind !== SyntaxKind.CallExpression) {
                return undefined;
            }
            const ctorType = checkCallExpressionTarget(callExpNode, undefined);
            if (ctorType?.kind !== "ScalarConstructor") {
                return undefined;
            }
            const argIndex = callExpNode.arguments.findIndex((n) => n === argNode);
            if (argIndex < 0 || argIndex >= ctorType.parameters.length) {
                return undefined;
            }
            const arg = ctorType.parameters[argIndex];
            return arg.type;
        }
        function getReferencedTypeFromConstAssignment(valueNode) {
            const constNode = valueNode?.parent;
            if (!constNode ||
                constNode.kind !== SyntaxKind.ConstStatement ||
                !constNode.type ||
                constNode.value !== valueNode) {
                return undefined;
            }
            return program.checker.getTypeForNode(constNode.type);
        }
        function getReferencedTypeFromDecoratorArgument(decArgNode) {
            const decNode = decArgNode?.parent;
            if (decNode?.kind !== SyntaxKind.DecoratorExpression) {
                return undefined;
            }
            const decSym = program.checker.resolveRelatedSymbols(decNode.target.kind === SyntaxKind.MemberExpression ? decNode.target.id : decNode.target);
            if (!decSym || decSym.length <= 0) {
                return undefined;
            }
            const decDecl = decSym[0].declarations.find((x) => x.kind === SyntaxKind.DecoratorDeclarationStatement);
            if (!decDecl) {
                return undefined;
            }
            const decType = program.checker.getTypeForNode(decDecl);
            compilerAssert(decType.kind === "Decorator", "Expected type to be a Decorator.");
            const argIndex = decNode.arguments.findIndex((n) => n === decArgNode);
            if (argIndex < 0 || argIndex >= decType.parameters.length) {
                return undefined;
            }
            const decArg = decType.parameters[argIndex];
            let type;
            if (isModelOrArrayValue(decArgNode)) {
                type = decArg.type.valueType;
            }
            else if (isModelOrArrayType(decArgNode)) {
                type = decArg.type.type ?? decArg.type.valueType;
            }
            else {
                compilerAssert(false, "not expected node type to get reference model from decorator argument");
            }
            return type;
        }
    }
    function resolveCompletions(identifier) {
        const completions = new Map();
        const { kind, node: ancestor } = getIdentifierContext(identifier);
        switch (kind) {
            case IdentifierKind.Using:
            case IdentifierKind.Decorator:
            case IdentifierKind.Function:
            case IdentifierKind.TypeReference:
            case IdentifierKind.ModelExpressionProperty:
            case IdentifierKind.ModelStatementProperty:
            case IdentifierKind.ObjectLiteralProperty:
                break; // supported
            case IdentifierKind.Other:
                return completions; // not implemented
            case IdentifierKind.Declaration:
                return completions; // cannot complete, name can be chosen arbitrarily
            case IdentifierKind.TemplateArgument: {
                const templates = getTemplateDeclarationsForArgument(ancestor, undefined);
                for (const template of templates) {
                    for (const param of template.templateParameters) {
                        addCompletion(param.id.sv, param.symbol);
                    }
                }
                return completions;
            }
            default:
                compilerAssert(false, "Unreachable");
        }
        if (kind === IdentifierKind.ModelStatementProperty) {
            const model = ancestor.parent;
            const modelType = program.checker.getTypeForNode(model);
            const baseType = modelType.baseModel;
            const baseNode = baseType?.node;
            if (!baseNode) {
                return completions;
            }
            for (const prop of baseType.properties.values()) {
                if (identifier.sv === prop.name || !modelType.properties.has(prop.name)) {
                    const sym = getMemberSymbol(baseNode.symbol, prop.name);
                    if (sym) {
                        addCompletion(prop.name, sym);
                    }
                }
            }
        }
        else if (kind === IdentifierKind.ModelExpressionProperty ||
            kind === IdentifierKind.ObjectLiteralProperty) {
            const model = getReferencedModel(ancestor);
            if (model.length <= 0) {
                return completions;
            }
            const curModelNode = ancestor.parent;
            for (const curModel of model) {
                for (const prop of walkPropertiesInherited(curModel)) {
                    if (identifier.sv === prop.name ||
                        !curModelNode.properties.find((p) => (p.kind === SyntaxKind.ModelProperty ||
                            p.kind === SyntaxKind.ObjectLiteralProperty) &&
                            p.id.sv === prop.name)) {
                        const sym = getMemberSymbol(curModel.node.symbol, prop.name);
                        if (sym) {
                            addCompletion(prop.name, sym);
                        }
                    }
                }
            }
        }
        else if (identifier.parent && identifier.parent.kind === SyntaxKind.MemberExpression) {
            let base = resolver.getNodeLinks(identifier.parent.base).resolvedSymbol;
            if (base) {
                if (base.flags & 128 /* SymbolFlags.Alias */) {
                    base = getAliasedSymbol(base, undefined);
                }
                if (base) {
                    if (identifier.parent.selector === "::") {
                        if (base?.node === undefined && base?.declarations && base.declarations.length > 0) {
                            // Process meta properties separately, such as `::parameters`, `::returnType`
                            const nodeModels = base?.declarations[0];
                            if (nodeModels.kind === SyntaxKind.OperationStatement) {
                                const operation = nodeModels;
                                addCompletion("parameters", operation.symbol);
                                addCompletion("returnType", operation.symbol);
                            }
                        }
                        else if (base?.node?.kind === SyntaxKind.ModelProperty) {
                            // Process meta properties separately, such as `::type`
                            const metaProperty = base.node;
                            addCompletion("type", metaProperty.symbol);
                        }
                    }
                    else {
                        addCompletions(base.exports ?? base.members);
                    }
                }
            }
        }
        else {
            // We will only add template arguments if the template isn't already named
            // to avoid completing the name of the argument again.
            if (kind === IdentifierKind.TypeReference &&
                exprIsBareIdentifier(ancestor) &&
                ancestor.parent?.kind === SyntaxKind.TemplateArgument &&
                ancestor.parent.name === undefined) {
                const templates = getTemplateDeclarationsForArgument(ancestor.parent, undefined);
                for (const template of templates) {
                    for (const param of template.templateParameters) {
                        addCompletion(param.id.sv, param.symbol, { suffix: " = " });
                    }
                }
            }
            let scope = identifier.parent;
            while (scope && scope.kind !== SyntaxKind.TypeSpecScript) {
                if (scope.symbol && scope.symbol.exports) {
                    const mergedSymbol = getMergedSymbol(scope.symbol);
                    addCompletions(mergedSymbol.exports);
                }
                if ("locals" in scope) {
                    addCompletions(scope.locals);
                }
                scope = scope.parent;
            }
            if (scope && scope.kind === SyntaxKind.TypeSpecScript) {
                // check any blockless namespace decls
                for (const ns of scope.inScopeNamespaces) {
                    const mergedSymbol = getMergedSymbol(ns.symbol);
                    addCompletions(mergedSymbol.exports);
                }
                // check "global scope" declarations
                addCompletions(resolver.symbols.global.exports);
                // check "global scope" usings
                addCompletions(scope.locals);
            }
        }
        return completions;
        function addCompletions(table) {
            if (!table) {
                return;
            }
            table = resolver.getAugmentedSymbolTable(table);
            for (const [key, sym] of table) {
                if (sym.flags & 16384 /* SymbolFlags.DuplicateUsing */) {
                    const duplicates = table.duplicates.get(sym);
                    for (const duplicate of duplicates) {
                        if (duplicate.flags & 8192 /* SymbolFlags.Using */) {
                            const fqn = getFullyQualifiedSymbolName(duplicate.symbolSource);
                            addCompletion(fqn, duplicate);
                        }
                    }
                }
                else {
                    addCompletion(key, sym);
                }
            }
        }
        function addCompletion(key, sym, options = {}) {
            if (sym.symbolSource) {
                sym = sym.symbolSource;
            }
            if (!shouldAddCompletion(sym)) {
                return;
            }
            if (key.startsWith("@")) {
                key = key.slice(1);
            }
            if (!completions.has(key)) {
                completions.set(key, { ...options, sym });
            }
        }
        function shouldAddCompletion(sym) {
            switch (kind) {
                case IdentifierKind.ModelExpressionProperty:
                case IdentifierKind.ModelStatementProperty:
                case IdentifierKind.ObjectLiteralProperty:
                    return !!(sym.flags & 65536 /* SymbolFlags.Member */);
                case IdentifierKind.Decorator:
                    // Only return decorators and namespaces when completing decorator
                    return !!(sym.flags & (512 /* SymbolFlags.Decorator */ | 256 /* SymbolFlags.Namespace */));
                case IdentifierKind.Using:
                    // Only return namespaces when completing using
                    return !!(sym.flags & 256 /* SymbolFlags.Namespace */);
                case IdentifierKind.TypeReference:
                    // Do not return functions or decorators when completing types
                    return !(sym.flags & (2048 /* SymbolFlags.Function */ | 512 /* SymbolFlags.Decorator */));
                case IdentifierKind.TemplateArgument:
                    return !!(sym.flags & 1024 /* SymbolFlags.TemplateParameter */);
                default:
                    compilerAssert(false, "We should have bailed up-front on other kinds.");
            }
        }
    }
    function getCodefixesForUnknownIdentifier(node) {
        switch (node.sv) {
            case "number":
                return [createChangeIdentifierCodeFix(node, "float64")];
            default:
                return undefined;
        }
    }
    function resolveTypeReferenceSym(node, mapper, options) {
        const resolvedOptions = typeof options === "boolean"
            ? { ...defaultSymbolResolutionOptions, resolveDecorators: options }
            : { ...defaultSymbolResolutionOptions, ...(options ?? {}) };
        if (mapper === undefined && resolvedOptions.checkTemplateTypes && referenceSymCache.has(node)) {
            return referenceSymCache.get(node);
        }
        const sym = resolveTypeReferenceSymInternal(node, mapper, resolvedOptions);
        if (resolvedOptions.checkTemplateTypes) {
            referenceSymCache.set(node, sym);
        }
        return sym;
    }
    function resolveTypeReferenceSymInternal(node, mapper, options) {
        if (hasParseError(node)) {
            // Don't report synthetic identifiers used for parser error recovery.
            // The parse error is the root cause and will already have been logged.
            return undefined;
        }
        if (node.kind === SyntaxKind.TypeReference) {
            return resolveTypeReferenceSym(node.target, mapper, options);
        }
        else if (node.kind === SyntaxKind.Identifier) {
            const links = resolver.getNodeLinks(node);
            if (mapper === undefined && links.resolutionResult) {
                if (mapper === undefined && // do not report error when instantiating
                    links.resolutionResult & (ResolutionResultFlags.NotFound | ResolutionResultFlags.Unknown)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-ref",
                        messageId: options.resolveDecorators ? "decorator" : "identifier",
                        format: { id: printTypeReferenceNode(node) },
                        target: node,
                        codefixes: getCodefixesForUnknownIdentifier(node),
                    }));
                }
                else if (links.resolutionResult & ResolutionResultFlags.Ambiguous) {
                    reportAmbiguousIdentifier(node, links.ambiguousSymbols);
                }
            }
            const sym = links.resolvedSymbol;
            return sym?.symbolSource ?? sym;
        }
        else if (node.kind === SyntaxKind.MemberExpression) {
            let base = resolveTypeReferenceSym(node.base, mapper, {
                ...options,
                resolveDecorators: false, // when resolving decorator the base cannot also be one
            });
            if (!base) {
                return undefined;
            }
            // when resolving a type reference based on an alias, unwrap the alias.
            if (base.flags & 128 /* SymbolFlags.Alias */) {
                const aliasedSym = getAliasedSymbol(base, mapper);
                if (!aliasedSym) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-ref",
                        messageId: "node",
                        format: {
                            id: node.id.sv,
                            nodeName: base.declarations[0]
                                ? SyntaxKind[base.declarations[0].kind]
                                : "Unknown node",
                        },
                        target: node,
                    }));
                    return undefined;
                }
                base = aliasedSym;
            }
            return resolveMemberInContainer(base, node, options);
        }
        compilerAssert(false, `Unknown type reference kind "${SyntaxKind[node.kind]}"`, node);
    }
    function reportAmbiguousIdentifier(node, symbols) {
        const duplicateNames = symbols.map((s) => getFullyQualifiedSymbolName(s, { useGlobalPrefixAtTopLevel: true }));
        program.reportDiagnostic(createDiagnostic({
            code: "ambiguous-symbol",
            format: { name: node.sv, duplicateNames: duplicateNames.join(", ") },
            target: node,
        }));
    }
    function resolveMemberInContainer(base, node, options) {
        const { finalSymbol: sym, resolvedSymbol: nextSym } = resolver.resolveMemberExpressionForSym(base, node, options);
        const symbol = nextSym ?? sym;
        if (symbol) {
            return symbol;
        }
        if (base.flags & 256 /* SymbolFlags.Namespace */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "underNamespace",
                format: {
                    namespace: getFullyQualifiedSymbolName(base),
                    id: node.id.sv,
                },
                target: node,
            }));
        }
        else if (base.flags & 512 /* SymbolFlags.Decorator */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "inDecorator",
                format: { id: node.id.sv },
                target: node,
            }));
        }
        else if (base.flags & 2048 /* SymbolFlags.Function */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "node",
                format: { id: node.id.sv, nodeName: "function" },
                target: node,
            }));
        }
        else if (base.flags & 118 /* SymbolFlags.MemberContainer */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: node.selector === "." ? "member" : "metaProperty",
                format: { kind: getMemberKindName(getSymNode(base)), id: node.id.sv },
                target: node,
            }));
        }
        else {
            const symNode = getSymNode(base);
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "node",
                format: {
                    id: node.id.sv,
                    nodeName: symNode ? SyntaxKind[symNode.kind] : "Unknown node",
                },
                target: node,
            }));
        }
        return undefined;
    }
    function getMemberKindName(node) {
        switch (node.kind) {
            case SyntaxKind.ModelStatement:
            case SyntaxKind.ModelExpression:
                return "Model";
            case SyntaxKind.ModelProperty:
                return "ModelProperty";
            case SyntaxKind.EnumStatement:
                return "Enum";
            case SyntaxKind.InterfaceStatement:
                return "Interface";
            case SyntaxKind.UnionStatement:
                return "Union";
            default:
                return "Type";
        }
    }
    /**
     * Return the symbol that is aliased by this alias declaration. If no such symbol is aliased,
     * return the symbol for the alias instead. For member containers which need to be late bound
     * (i.e. they contain symbols we don't know until we've instantiated the type and the type is an
     * instantiation) we late bind the container which creates the symbol that will hold its members.
     */
    function getAliasedSymbol(aliasSymbol, mapper) {
        const node = getSymNode(aliasSymbol);
        const links = resolver.getSymbolLinks(aliasSymbol);
        if (!links.aliasResolutionIsTemplate) {
            return links.aliasedSymbol ?? resolver.getNodeLinks(node).resolvedSymbol;
        }
        // Otherwise for templates we need to get the type and retrieve the late bound symbol.
        const aliasType = getTypeForNode(node, mapper);
        if (isErrorType(aliasType)) {
            return undefined;
        }
        switch (aliasType.kind) {
            case "Model":
            case "Interface":
            case "Union":
                if (isTemplateInstance(aliasType)) {
                    // this is an alias for some instantiation, so late-bind the instantiation
                    lateBindMemberContainer(aliasType);
                    return aliasType.symbol;
                }
            // fallthrough
            default:
                // get the symbol from the node aliased type's node, or just return the base
                // if it doesn't have a symbol (which will likely result in an error later on)
                return getMergedSymbol(aliasType.node.symbol) ?? aliasSymbol;
        }
    }
    function checkStringTemplateExpresion(node, mapper) {
        let hasType = false;
        let hasValue = false;
        const spanTypeOrValues = node.spans.map((span) => [span, checkNode(span.expression, mapper)]);
        for (const [_, typeOrValue] of spanTypeOrValues) {
            if (typeOrValue !== null) {
                if (isValue(typeOrValue)) {
                    hasValue = true;
                }
                else if ("kind" in typeOrValue && typeOrValue.kind === "TemplateParameter") {
                    if (typeOrValue.constraint) {
                        if (typeOrValue.constraint.valueType) {
                            hasValue = true;
                        }
                        if (typeOrValue.constraint.type) {
                            hasType = true;
                        }
                    }
                    else {
                        hasType = true;
                    }
                }
                else {
                    hasType = true;
                }
            }
        }
        if (hasType && hasValue) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "mixed-string-template",
                target: node,
            }));
            return null;
        }
        if (hasValue) {
            let str = node.head.value;
            for (const [span, typeOrValue] of spanTypeOrValues) {
                if (typeOrValue !== null &&
                    (!("kind" in typeOrValue) || typeOrValue.kind !== "TemplateParameter")) {
                    compilerAssert(typeOrValue !== null && isValue(typeOrValue), "Expected value.");
                    str += stringifyValueForTemplate(typeOrValue);
                }
                str += span.literal.value;
            }
            return checkStringValue(createLiteralType(str), undefined, node);
        }
        else {
            let hasNonStringElement = false;
            let stringValue = node.head.value;
            const spans = [createTemplateSpanLiteral(node.head)];
            for (const [span, typeOrValue] of spanTypeOrValues) {
                compilerAssert(typeOrValue !== null && !isValue(typeOrValue), "Expected type.");
                const type = typeOrValue.entityKind === "Indeterminate" ? typeOrValue.type : typeOrValue;
                const spanValue = createTemplateSpanValue(span.expression, type);
                spans.push(spanValue);
                const spanValueAsString = stringifyTypeForTemplate(type);
                if (spanValueAsString) {
                    stringValue += spanValueAsString;
                }
                else {
                    hasNonStringElement = true;
                }
                spans.push(createTemplateSpanLiteral(span.literal));
                stringValue += span.literal.value;
            }
            return createIndeterminateEntity(createType({
                kind: "StringTemplate",
                node,
                spans,
                stringValue: hasNonStringElement ? undefined : stringValue,
            }));
        }
    }
    function createIndeterminateEntity(type) {
        const existing = indeterminateEntities.get(type);
        if (existing) {
            return existing;
        }
        const entity = {
            entityKind: "Indeterminate",
            type,
        };
        indeterminateEntities.set(type, entity);
        return entity;
    }
    function stringifyTypeForTemplate(type) {
        switch (type.kind) {
            case "String":
            case "Number":
            case "Boolean":
                return String(type.value);
            case "StringTemplate":
                if (type.stringValue !== undefined) {
                    return type.stringValue;
                }
                return undefined;
            default:
                return undefined;
        }
    }
    function stringifyValueForTemplate(value) {
        switch (value.valueKind) {
            case "StringValue":
            case "NumericValue":
            case "BooleanValue":
                return value.value.toString();
            default:
                reportCheckerDiagnostic(createDiagnostic({
                    code: "non-literal-string-template",
                    target: value,
                }));
                return `[${value.valueKind}]`;
        }
    }
    function createTemplateSpanLiteral(node) {
        return createType({
            kind: "StringTemplateSpan",
            node: node,
            isInterpolated: false,
            type: getLiteralType(node),
        });
    }
    function createTemplateSpanValue(node, type) {
        return createType({
            kind: "StringTemplateSpan",
            node: node,
            isInterpolated: true,
            type: type,
        });
    }
    function checkStringLiteral(str) {
        return createIndeterminateEntity(getLiteralType(str));
    }
    function checkNumericLiteral(num) {
        return createIndeterminateEntity(getLiteralType(num));
    }
    function checkBooleanLiteral(bool) {
        return createIndeterminateEntity(getLiteralType(bool));
    }
    function checkProgram() {
        checkDuplicateSymbols();
        for (const file of program.sourceFiles.values()) {
            checkDuplicateUsings(file);
            for (const ns of file.namespaces) {
                initializeTypeForNamespace(ns);
            }
        }
        for (const file of program.sourceFiles.values()) {
            checkSourceFile(file);
        }
        internalDecoratorValidation();
    }
    function checkDuplicateSymbols() {
        program.reportDuplicateSymbols(resolver.symbols.global.exports);
        for (const file of program.sourceFiles.values()) {
            for (const ns of file.namespaces) {
                const exports = getMergedSymbol(ns.symbol).exports ?? ns.symbol.exports;
                program.reportDuplicateSymbols(exports);
            }
        }
    }
    /** Report error with duplicate using in the same scope. */
    function checkDuplicateUsings(file) {
        const duplicateTrackers = new Map();
        function getTracker(sym) {
            const existing = duplicateTrackers.get(sym);
            if (existing)
                return existing;
            const newTacker = new DuplicateTracker();
            duplicateTrackers.set(sym, newTacker);
            return newTacker;
        }
        for (const using of file.usings) {
            const ns = using.parent;
            const sym = getMergedSymbol(ns.symbol);
            const tracker = getTracker(sym);
            const targetSym = resolver.getNodeLinks(using.name).resolvedSymbol;
            if (!targetSym)
                continue;
            tracker.track(targetSym, using);
        }
        for (const tracker of duplicateTrackers.values()) {
            for (const [_, nodes] of tracker.entries()) {
                for (const node of nodes) {
                    program.reportDiagnostic(createDiagnostic({
                        code: "duplicate-using",
                        format: { usingName: typeReferenceToString(node.name) },
                        target: node,
                    }));
                }
            }
        }
    }
    /**
     * Post checking validation for internal decorators.
     */
    function internalDecoratorValidation() {
        validateInheritanceDiscriminatedUnions(program);
    }
    function checkSourceFile(file) {
        for (const statement of file.statements) {
            checkNode(statement, undefined);
        }
    }
    /**
     * Check that the given node template parameters are valid if applicable.
     * @param node Node with template parameters
     * @param mapper Type mapper, set if instantiating the template, undefined otherwise.
     */
    function checkTemplateDeclaration(node, mapper) {
        // If mapper is undefined it means we are checking the declaration of the template.
        if (mapper === undefined) {
            for (const templateParameter of node.templateParameters) {
                checkTemplateParameterDeclaration(templateParameter, undefined);
            }
        }
    }
    function checkModel(node, mapper) {
        if (node.kind === SyntaxKind.ModelStatement) {
            return checkModelStatement(node, mapper);
        }
        else {
            return checkModelExpression(node, mapper);
        }
    }
    function checkModelStatement(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const decorators = [];
        const type = createType({
            kind: "Model",
            name: node.id.sv,
            node: node,
            properties: createRekeyableMap(),
            namespace: getParentNamespaceType(node),
            decorators,
            sourceModels: [],
            derivedModels: [],
        });
        linkType(links, type, mapper);
        if (node.symbol.members) {
            const members = resolver.getAugmentedSymbolTable(node.symbol.members);
            const propDocs = extractPropDocs(node);
            for (const [name, memberSym] of members) {
                const doc = propDocs.get(name);
                if (doc) {
                    docFromCommentForSym.set(memberSym, doc);
                }
            }
        }
        const isBase = checkModelIs(node, node.is, mapper);
        if (isBase) {
            type.sourceModel = isBase;
            type.sourceModels.push({ usage: "is", model: isBase });
            // copy decorators
            decorators.push(...isBase.decorators);
            if (isBase.indexer) {
                type.indexer = isBase.indexer;
            }
        }
        if (isBase) {
            for (const prop of isBase.properties.values()) {
                const memberSym = getMemberSymbol(node.symbol, prop.name);
                const newProp = cloneTypeForSymbol(memberSym, prop, {
                    sourceProperty: prop,
                    model: type,
                });
                linkIndirectMember(node, newProp, mapper);
                type.properties.set(prop.name, newProp);
            }
        }
        if (isBase) {
            type.baseModel = isBase.baseModel;
        }
        else if (node.extends) {
            type.baseModel = checkClassHeritage(node, node.extends, mapper);
            if (type.baseModel) {
                copyDeprecation(type.baseModel, type);
            }
        }
        if (type.baseModel) {
            type.baseModel.derivedModels.push(type);
        }
        // Hold on to the model type that's being defined so that it
        // can be referenced
        if (mapper === undefined) {
            type.namespace?.models.set(type.name, type);
        }
        // Evaluate the properties after
        checkModelProperties(node, type.properties, type, mapper);
        decorators.push(...checkDecorators(type, node, mapper));
        linkMapper(type, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(type);
        }
        const indexer = getIndexer(program, type);
        if (type.name === "Array" && isInTypeSpecNamespace(type)) {
            stdTypes.Array = type;
        }
        else if (type.name === "Record" && isInTypeSpecNamespace(type)) {
            stdTypes.Record = type;
        }
        if (indexer) {
            type.indexer = indexer;
        }
        lateBindMemberContainer(type);
        lateBindMembers(type);
        return type;
    }
    function shouldCreateTypeForTemplate(node, mapper) {
        // Node is not a template we should create the type.
        if (node.templateParameters.length === 0) {
            return true;
        }
        // There is no mapper so we shouldn't be instantiating the template.
        if (mapper === undefined) {
            return false;
        }
        // Some of the mapper args are still template parameter so we shouldn't create the type.
        return (!mapper.partial &&
            mapper.args.every((t) => isValue(t) || t.entityKind === "Indeterminate" || t.kind !== "TemplateParameter"));
    }
    function checkModelExpression(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        const properties = createRekeyableMap();
        const type = createType({
            kind: "Model",
            name: "",
            node: node,
            properties,
            indexer: undefined,
            namespace: getParentNamespaceType(node),
            decorators: [],
            derivedModels: [],
            sourceModels: [],
        });
        linkType(links, type, mapper);
        linkMapper(type, mapper);
        checkModelProperties(node, properties, type, mapper);
        return finishType(type);
    }
    /** Find the indexer that applies to this model. Either defined on itself or from a base model */
    function findIndexer(model) {
        let current = model;
        while (current) {
            if (current.indexer) {
                return current.indexer;
            }
            current = current.baseModel;
        }
        return undefined;
    }
    function checkPropertyCompatibleWithModelIndexer(parentModel, property, diagnosticTarget) {
        const indexer = findIndexer(parentModel);
        if (indexer === undefined) {
            return;
        }
        return checkPropertyCompatibleWithIndexer(indexer, property, diagnosticTarget);
    }
    function checkPropertyCompatibleWithIndexer(indexer, property, diagnosticTarget) {
        if (indexer.key.name === "integer") {
            reportCheckerDiagnostics([
                createDiagnostic({
                    code: "no-array-properties",
                    target: diagnosticTarget,
                }),
            ]);
            return;
        }
        const [valid, diagnostics] = relation.isTypeAssignableTo(property.type, indexer.value, diagnosticTarget);
        if (!valid)
            reportCheckerDiagnostic(createDiagnostic({
                code: "incompatible-indexer",
                format: { message: diagnostics.map((x) => `  ${x.message}`).join("\n") },
                target: diagnosticTarget.kind === SyntaxKind.ModelProperty
                    ? diagnosticTarget.value
                    : diagnosticTarget,
            }));
    }
    function checkModelProperties(node, properties, parentModel, mapper) {
        let spreadIndexers;
        for (const prop of node.properties) {
            if ("id" in prop) {
                const newProp = checkModelProperty(prop, mapper);
                newProp.model = parentModel;
                checkPropertyCompatibleWithModelIndexer(parentModel, newProp, prop);
                defineProperty(properties, newProp);
            }
            else {
                // spread property
                const [newProperties, additionalIndexer] = checkSpreadProperty(node.symbol, prop.target, parentModel, mapper);
                if (additionalIndexer) {
                    if (spreadIndexers) {
                        spreadIndexers.push(additionalIndexer);
                    }
                    else {
                        spreadIndexers = [additionalIndexer];
                    }
                }
                for (const newProp of newProperties) {
                    linkIndirectMember(node, newProp, mapper);
                    checkPropertyCompatibleWithModelIndexer(parentModel, newProp, prop);
                    defineProperty(properties, newProp, prop);
                }
            }
        }
        if (spreadIndexers) {
            const value = spreadIndexers.length === 1
                ? spreadIndexers[0].value
                : createUnion(spreadIndexers.map((i) => i.value));
            parentModel.indexer = {
                key: spreadIndexers[0].key,
                value: value,
            };
        }
    }
    function checkObjectValue(node, mapper, constraint) {
        const properties = checkObjectLiteralProperties(node, mapper);
        if (properties === null) {
            return null;
        }
        const preciseType = createTypeForObjectValue(node, properties);
        if (constraint && !checkTypeOfValueMatchConstraint(preciseType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "ObjectValue",
            node: node,
            properties,
            type: constraint ? constraint.type : preciseType,
        }, preciseType);
    }
    function createTypeForObjectValue(node, properties) {
        const model = createType({
            kind: "Model",
            name: "",
            node,
            properties: createRekeyableMap(),
            decorators: [],
            derivedModels: [],
            sourceModels: [],
        });
        for (const prop of properties.values()) {
            model.properties.set(prop.name, createModelPropertyForObjectPropertyDescriptor(prop, model));
        }
        return finishType(model);
    }
    function createModelPropertyForObjectPropertyDescriptor(prop, parentModel) {
        return createAndFinishType({
            kind: "ModelProperty",
            node: prop.node,
            model: parentModel,
            optional: false,
            name: prop.name,
            type: prop.value.type,
            decorators: [],
        });
    }
    function checkObjectLiteralProperties(node, mapper) {
        const properties = new Map();
        let hasError = false;
        for (const prop of node.properties) {
            if ("id" in prop) {
                const value = getValueForNode(prop.value, mapper);
                if (value === null) {
                    hasError = true;
                }
                else {
                    properties.set(prop.id.sv, { name: prop.id.sv, value: value, node: prop });
                }
            }
            else {
                const targetType = checkObjectSpreadProperty(prop.target, mapper);
                if (targetType) {
                    for (const [name, value] of targetType.properties) {
                        properties.set(name, { ...value });
                    }
                }
            }
        }
        return hasError ? null : properties;
    }
    function checkObjectSpreadProperty(targetNode, mapper) {
        const value = getValueForNode(targetNode, mapper);
        if (value === null) {
            return null;
        }
        if (value.valueKind !== "ObjectValue") {
            reportCheckerDiagnostic(createDiagnostic({ code: "spread-object", target: targetNode }));
            return null;
        }
        return value;
    }
    function checkArrayValue(node, mapper, constraint) {
        let hasError = false;
        const values = node.values.map((itemNode) => {
            const value = getValueForNode(itemNode, mapper);
            if (value === null) {
                hasError = true;
            }
            return value;
        });
        if (hasError) {
            return null;
        }
        const preciseType = createTypeForArrayValue(node, values);
        if (constraint && !checkTypeOfValueMatchConstraint(preciseType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "ArrayValue",
            node: node,
            values: values,
            type: constraint ? constraint.type : preciseType,
        }, preciseType);
    }
    function createValue(value, preciseType) {
        valueExactTypes.set(value, preciseType);
        return value;
    }
    function copyValue(value, overrides) {
        const newValue = { ...value, ...overrides };
        const preciseType = valueExactTypes.get(value);
        if (preciseType) {
            valueExactTypes.set(newValue, preciseType);
        }
        return newValue;
    }
    function createTypeForArrayValue(node, values) {
        return createAndFinishType({
            kind: "Tuple",
            node,
            values: values.map((x) => x.type),
        });
    }
    function inferScalarForPrimitiveValue(type, literalType) {
        if (type === undefined) {
            return undefined;
        }
        switch (type.kind) {
            case "Scalar":
                if (ignoreDiagnostics(checker.isTypeAssignableTo(literalType, type, literalType))) {
                    return type;
                }
                return undefined;
            case "Union":
                let found = undefined;
                for (const variant of type.variants.values()) {
                    const scalar = inferScalarForPrimitiveValue(variant.type, literalType);
                    if (scalar) {
                        if (found) {
                            reportCheckerDiagnostic(createDiagnostic({
                                code: "ambiguous-scalar-type",
                                format: {
                                    value: getTypeName(literalType),
                                    types: [found, scalar].map((x) => x.name).join(", "),
                                    example: found.name,
                                },
                                target: literalType,
                            }));
                            return undefined;
                        }
                        else {
                            found = scalar;
                        }
                    }
                }
                return found;
            default:
                return undefined;
        }
    }
    function checkStringValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        let value;
        if (literalType.kind === "StringTemplate") {
            if (literalType.stringValue) {
                value = literalType.stringValue;
            }
            else {
                reportCheckerDiagnostics(explainStringTemplateNotSerializable(literalType));
                return null;
            }
        }
        else {
            value = literalType.value;
        }
        const scalar = inferScalarForPrimitiveValue(constraint?.type, literalType);
        return createValue({
            entityKind: "Value",
            valueKind: "StringValue",
            value,
            type: constraint ? constraint.type : literalType,
            scalar,
        }, literalType);
    }
    function checkNumericValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        const scalar = inferScalarForPrimitiveValue(constraint?.type, literalType);
        return createValue({
            entityKind: "Value",
            valueKind: "NumericValue",
            value: Numeric(literalType.valueAsString),
            type: constraint ? constraint.type : literalType,
            scalar,
        }, literalType);
    }
    function checkBooleanValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        const scalar = inferScalarForPrimitiveValue(constraint?.type, literalType);
        return createValue({
            entityKind: "Value",
            valueKind: "BooleanValue",
            value: literalType.value,
            type: constraint ? constraint.type : literalType,
            scalar,
        }, literalType);
    }
    function checkNullValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "NullValue",
            type: constraint ? constraint.type : literalType,
            value: null,
        }, literalType);
    }
    function checkEnumValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "EnumValue",
            type: constraint ? constraint.type : literalType,
            value: literalType,
        }, literalType);
    }
    function checkCallExpressionTarget(node, mapper) {
        const target = checkTypeReference(node.target, mapper);
        if (target.kind === "Scalar" || target.kind === "ScalarConstructor") {
            return target;
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({
                code: "non-callable",
                format: { type: target.kind },
                target: node.target,
            }));
            return null;
        }
    }
    /** Check the arguments of the call expression are a single value of the given syntax. */
    function checkPrimitiveArg(node, scalar, valueKind) {
        if (node.arguments.length !== 1) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-primitive-init",
                target: node.target,
            }));
            return null;
        }
        const argNode = node.arguments[0];
        const value = getValueForNode(argNode, undefined);
        if (value === null) {
            return null; // error should already have been reported above.
        }
        if (value.valueKind !== valueKind) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-primitive-init",
                messageId: "invalidArg",
                format: { actual: value.valueKind, expected: valueKind },
                target: argNode,
            }));
            return null;
        }
        if (!checkValueOfType(value, scalar, argNode)) {
            return null;
        }
        return copyValue(value, { scalar, type: scalar });
    }
    function createScalarValue(node, mapper, declaration) {
        let hasError = false;
        const minArgs = declaration.parameters.filter((x) => !x.optional && !x.rest).length ?? 0;
        const maxArgs = declaration.parameters[declaration.parameters.length - 1]?.rest
            ? undefined
            : declaration.parameters.length;
        if (node.arguments.length < minArgs ||
            (maxArgs !== undefined && node.arguments.length > maxArgs)) {
            // In the case we have too little args then this decorator is not applicable.
            // If there is too many args then we can still run the decorator as long as the args are valid.
            if (node.arguments.length < minArgs) {
                hasError = true;
            }
            if (maxArgs === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    messageId: "atLeast",
                    format: { actual: node.arguments.length.toString(), expected: minArgs.toString() },
                    target: node,
                }));
            }
            else {
                const expected = minArgs === maxArgs ? minArgs.toString() : `${minArgs}-${maxArgs}`;
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    format: { actual: node.arguments.length.toString(), expected },
                    target: node,
                }));
            }
        }
        const resolvedArgs = [];
        for (const [index, parameter] of declaration.parameters.entries()) {
            if (parameter.rest) {
                const restType = getIndexType(parameter.type);
                if (restType) {
                    for (let i = index; i < node.arguments.length; i++) {
                        const argNode = node.arguments[i];
                        if (argNode) {
                            const arg = getValueForNode(argNode, mapper, { kind: "argument", type: restType });
                            if (arg === null) {
                                hasError = true;
                                continue;
                            }
                            if (checkValueOfType(arg, restType, argNode)) {
                                resolvedArgs.push(arg);
                            }
                            else {
                                hasError = true;
                            }
                        }
                    }
                }
                break;
            }
            const argNode = node.arguments[index];
            if (argNode) {
                const arg = getValueForNode(argNode, mapper, {
                    kind: "argument",
                    type: parameter.type,
                });
                if (arg === null) {
                    hasError = true;
                    continue;
                }
                if (checkValueOfType(arg, parameter.type, argNode)) {
                    resolvedArgs.push(arg);
                }
                else {
                    hasError = true;
                }
            }
        }
        if (hasError) {
            return null;
        }
        return {
            entityKind: "Value",
            valueKind: "ScalarValue",
            value: {
                name: declaration.name,
                args: resolvedArgs,
            },
            scalar: declaration.scalar,
            type: declaration.scalar,
        };
    }
    function checkCallExpression(node, mapper) {
        const target = checkCallExpressionTarget(node, mapper);
        if (target === null) {
            return null;
        }
        if (target.kind === "ScalarConstructor") {
            return createScalarValue(node, mapper, target);
        }
        if (relation.areScalarsRelated(target, getStdType("string"))) {
            return checkPrimitiveArg(node, target, "StringValue");
        }
        else if (relation.areScalarsRelated(target, getStdType("numeric"))) {
            return checkPrimitiveArg(node, target, "NumericValue");
        }
        else if (relation.areScalarsRelated(target, getStdType("boolean"))) {
            return checkPrimitiveArg(node, target, "BooleanValue");
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({
                code: "named-init-required",
                format: { typeKind: target.kind },
                target: node.target,
            }));
            return null;
        }
    }
    function checkTypeOfExpression(node, mapper) {
        const entity = checkNode(node.target, mapper, undefined);
        if (entity === null) {
            // Shouldn't need to emit error as we assume null value already emitted error when produced
            return errorType;
        }
        if (entity.entityKind === "Indeterminate") {
            return entity.type;
        }
        if (isType(entity)) {
            if (entity.kind === "TemplateParameter") {
                if (entity.constraint === undefined || entity.constraint.type !== undefined) {
                    // means this template constraint will accept values
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "expect-value",
                        messageId: "templateConstraint",
                        format: { name: getTypeName(entity) },
                        target: node.target,
                    }));
                    return errorType;
                }
                else if (entity.constraint.valueType) {
                    return entity.constraint.valueType;
                }
            }
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                format: { name: getTypeName(entity) },
                target: node.target,
            }));
            return entity;
        }
        return entity.type;
    }
    function createUnion(options) {
        const variants = createRekeyableMap();
        const union = createAndFinishType({
            kind: "Union",
            node: undefined,
            options,
            decorators: [],
            variants,
            expression: true,
        });
        for (const option of options) {
            const name = Symbol("indexer-union-variant");
            variants.set(name, createAndFinishType({
                kind: "UnionVariant",
                node: undefined,
                type: option,
                name,
                union,
                decorators: [],
            }));
        }
        return union;
    }
    function defineProperty(properties, newProp, diagnosticTarget) {
        if (properties.has(newProp.name)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "duplicate-property",
                format: { propName: newProp.name },
                target: diagnosticTarget ?? newProp,
            }));
            return;
        }
        const overriddenProp = getOverriddenProperty(newProp);
        if (overriddenProp) {
            const [isAssignable, _] = relation.isTypeAssignableTo(newProp.type, overriddenProp.type, newProp);
            const parentType = getTypeName(overriddenProp.type);
            const newPropType = getTypeName(newProp.type);
            let invalid = false;
            if (!isAssignable) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "override-property-mismatch",
                    format: { propName: newProp.name, propType: newPropType, parentType: parentType },
                    target: diagnosticTarget ?? newProp,
                }));
                invalid = true;
            }
            if (!overriddenProp.optional && newProp.optional) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "override-property-mismatch",
                    messageId: "disallowedOptionalOverride",
                    format: { propName: overriddenProp.name },
                    target: diagnosticTarget ?? newProp,
                }));
                invalid = true;
            }
            if (invalid)
                return;
        }
        properties.set(newProp.name, newProp);
    }
    /**
     * Initializes a late bound symbol for the type. This is generally necessary when attempting to
     * access a symbol for a type that is created during the check phase.
     */
    function lateBindMemberContainer(type) {
        if (type.symbol)
            return;
        switch (type.kind) {
            case "Model":
                type.symbol = createSymbol(type.node, type.name, 2 /* SymbolFlags.Model */ | 4194304 /* SymbolFlags.LateBound */);
                mutate(type.symbol).type = type;
                break;
            case "Interface":
                type.symbol = createSymbol(type.node, type.name, 32 /* SymbolFlags.Interface */ | 4194304 /* SymbolFlags.LateBound */);
                if (isTemplateInstance(type) && type.name === "Foo") {
                    getSymbolLinks(type.symbol);
                }
                mutate(type.symbol).type = type;
                break;
            case "Union":
                if (!type.name)
                    return; // don't make a symbol for anonymous unions
                type.symbol = createSymbol(type.node, type.name, 64 /* SymbolFlags.Union */ | 4194304 /* SymbolFlags.LateBound */);
                mutate(type.symbol).type = type;
                break;
        }
    }
    function lateBindMembers(type) {
        compilerAssert(type.symbol, "Type must have a symbol to late bind members");
        const containerSym = type.symbol;
        compilerAssert(containerSym.members, "Container symbol didn't have members at late-bind");
        const containerMembers = resolver.getAugmentedSymbolTable(containerSym.members);
        switch (type.kind) {
            case "Model":
                for (const prop of walkPropertiesInherited(type)) {
                    lateBindMember(prop, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Scalar":
                for (const member of type.constructors.values()) {
                    lateBindMember(member, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Enum":
                for (const member of type.members.values()) {
                    lateBindMember(member, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Interface":
                for (const member of type.operations.values()) {
                    lateBindMember(member, 65536 /* SymbolFlags.Member */ | 8 /* SymbolFlags.Operation */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Union":
                for (const variant of type.variants.values()) {
                    lateBindMember(variant, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
        }
        function lateBindMember(member, kind) {
            if (!member.node || typeof member.name !== "string") {
                // don't bind anything for union expressions
                return;
            }
            const sym = createSymbol(member.node, member.name, kind | 4194304 /* SymbolFlags.LateBound */, containerSym);
            mutate(sym).type = member;
            compilerAssert(containerSym.members, "containerSym.members is undefined");
            containerMembers.set(member.name, sym);
        }
    }
    function checkClassHeritage(model, heritageRef, mapper) {
        if (heritageRef.kind === SyntaxKind.ModelExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "extend-model",
                messageId: "modelExpression",
                target: heritageRef,
            }));
            return undefined;
        }
        if (heritageRef.kind !== SyntaxKind.TypeReference) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "extend-model",
                target: heritageRef,
            }));
            return undefined;
        }
        const modelSymId = getNodeSym(model);
        pendingResolutions.start(modelSymId, ResolutionKind.BaseType);
        const target = resolver.getNodeLinks(heritageRef).resolvedSymbol;
        if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-base-type",
                    format: { typeName: target.name },
                    target: target,
                }));
            }
            return undefined;
        }
        const heritageType = getTypeForNode(heritageRef, mapper);
        pendingResolutions.finish(modelSymId, ResolutionKind.BaseType);
        if (isErrorType(heritageType)) {
            compilerAssert(program.hasError(), "Should already have reported an error.", heritageRef);
            return undefined;
        }
        if (heritageType.kind !== "Model") {
            reportCheckerDiagnostic(createDiagnostic({ code: "extend-model", target: heritageRef }));
            return undefined;
        }
        if (heritageType.name === "") {
            reportCheckerDiagnostic(createDiagnostic({
                code: "extend-model",
                messageId: "modelExpression",
                target: heritageRef,
            }));
        }
        return heritageType;
    }
    function checkModelIs(model, isExpr, mapper) {
        if (!isExpr)
            return undefined;
        const modelSymId = getNodeSym(model);
        pendingResolutions.start(modelSymId, ResolutionKind.BaseType);
        let isType;
        if (isExpr.kind === SyntaxKind.ModelExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "is-model",
                messageId: "modelExpression",
                target: isExpr,
            }));
            return undefined;
        }
        else if (isExpr.kind === SyntaxKind.ArrayExpression) {
            isType = checkArrayExpression(isExpr, mapper);
        }
        else if (isExpr.kind === SyntaxKind.TypeReference) {
            const target = resolver.getNodeLinks(isExpr).resolvedSymbol;
            if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
                if (mapper === undefined) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "circular-base-type",
                        format: { typeName: target.name },
                        target: target,
                    }));
                }
                return undefined;
            }
            isType = getTypeForNode(isExpr, mapper);
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-model", target: isExpr }));
            return undefined;
        }
        pendingResolutions.finish(modelSymId, ResolutionKind.BaseType);
        if (isType.kind !== "Model") {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-model", target: isExpr }));
            return;
        }
        if (isType.name === "") {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-model", messageId: "modelExpression", target: isExpr }));
            return undefined;
        }
        return isType;
    }
    function checkSpreadProperty(parentModelSym, targetNode, parentModel, mapper) {
        const targetType = getTypeForNode(targetNode, mapper);
        if (targetType.kind === "TemplateParameter" || isErrorType(targetType)) {
            return [[], undefined];
        }
        if (targetType.kind !== "Model") {
            reportCheckerDiagnostic(createDiagnostic({ code: "spread-model", target: targetNode }), mapper);
            return [[], undefined];
        }
        if (isArrayModelType(program, targetType)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "spread-model", target: targetNode }), mapper);
            return [[], undefined];
        }
        if (parentModel === targetType) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "spread-model",
                messageId: "selfSpread",
                target: targetNode,
            }));
        }
        parentModel.sourceModels.push({ usage: "spread", model: targetType });
        const props = [];
        // copy each property
        for (const prop of walkPropertiesInherited(targetType)) {
            const memberSym = getMemberSymbol(parentModelSym, prop.name);
            props.push(cloneTypeForSymbol(memberSym, prop, {
                sourceProperty: prop,
                model: parentModel,
            }));
        }
        return [props, targetType.indexer];
    }
    /**
     * Link an indirect model property(included via spread or model is) to its model member symbols.
     * @param containerNode Model Node
     * @param member New Property
     * @param mapper Type Mapper.
     */
    function linkIndirectMember(containerNode, member, mapper) {
        if (mapper !== undefined) {
            return;
        }
        compilerAssert(typeof member.name === "string", "Cannot link unmapped unions");
        if (containerNode.symbol === undefined) {
            return;
        }
        const memberSym = getMemberSymbol(containerNode.symbol, member.name);
        if (memberSym) {
            const links = resolver.getSymbolLinks(memberSym);
            linkMemberType(links, member, mapper);
        }
    }
    function checkModelProperty(prop, mapper) {
        const sym = getSymbolForMember(prop);
        const links = getSymbolLinksForMember(prop);
        if (links && links.declaredType && mapper === undefined) {
            return links.declaredType;
        }
        const name = prop.id.sv;
        const type = createType({
            kind: "ModelProperty",
            name,
            node: prop,
            optional: prop.optional,
            type: undefined,
            decorators: [],
        });
        if (pendingResolutions.has(sym, ResolutionKind.Type) && mapper === undefined) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "circular-prop",
                format: { propName: name },
                target: prop,
            }));
            type.type = errorType;
        }
        else {
            pendingResolutions.start(sym, ResolutionKind.Type);
            type.type = getTypeForNode(prop.value, mapper);
            if (prop.default) {
                const defaultValue = checkDefaultValue(prop.default, type.type);
                if (defaultValue !== null) {
                    type.defaultValue = defaultValue;
                }
            }
            if (links) {
                linkType(links, type, mapper);
            }
        }
        type.decorators = checkDecorators(type, prop, mapper);
        const parentTemplate = getParentTemplateNode(prop);
        linkMapper(type, mapper);
        if (!parentTemplate || shouldCreateTypeForTemplate(parentTemplate, mapper)) {
            const docComment = docFromCommentForSym.get(sym);
            if (docComment) {
                type.decorators.unshift(createDocFromCommentDecorator("self", docComment));
            }
            finishType(type);
        }
        pendingResolutions.finish(sym, ResolutionKind.Type);
        return type;
    }
    function createDocFromCommentDecorator(key, doc) {
        return {
            decorator: docFromCommentDecorator,
            args: [
                { value: createLiteralType(key), jsValue: key },
                { value: createLiteralType(doc), jsValue: doc },
            ],
        };
    }
    function checkDefaultValue(defaultNode, type) {
        if (isErrorType(type)) {
            // if the prop type is an error we don't need to validate again.
            return null;
        }
        const defaultValue = getValueForNode(defaultNode, undefined, {
            kind: "assignment",
            type,
        }, { });
        if (defaultValue === null) {
            return null;
        }
        const [related, diagnostics] = relation.isValueOfType(defaultValue, type, defaultNode);
        if (!related) {
            reportCheckerDiagnostics(diagnostics);
            return null;
        }
        else {
            return { ...defaultValue, type };
        }
    }
    function checkDecoratorApplication(targetType, decNode, mapper) {
        const sym = resolveTypeReferenceSym(decNode.target, undefined, true);
        if (!sym) {
            // Error should already have been reported above
            return undefined;
        }
        if (!(sym.flags & 512 /* SymbolFlags.Decorator */)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-decorator",
                format: { id: sym.name },
                target: decNode,
            }));
            return undefined;
        }
        const symbolLinks = getSymbolLinks(sym);
        let hasError = false;
        if (symbolLinks.declaredType === undefined) {
            const decoratorDeclNode = sym.declarations.find((x) => x.kind === SyntaxKind.DecoratorDeclarationStatement);
            if (decoratorDeclNode) {
                checkDecoratorDeclaration(decoratorDeclNode, undefined);
            }
        }
        if (symbolLinks.declaredType) {
            compilerAssert(symbolLinks.declaredType.kind === "Decorator", `Expected to find a decorator type but got ${symbolLinks.declaredType.kind}`);
            if (!checkDecoratorTarget(targetType, symbolLinks.declaredType, decNode)) {
                hasError = true;
            }
        }
        const [argsHaveError, args] = checkDecoratorArguments(decNode, mapper, symbolLinks.declaredType);
        if (hasError || argsHaveError) {
            return undefined;
        }
        return {
            definition: symbolLinks.declaredType,
            decorator: sym.value ?? ((...args) => { }),
            node: decNode,
            args,
        };
    }
    /** Check the decorator target is valid */
    function checkDecoratorTarget(targetType, declaration, decoratorNode) {
        const [targetValid] = relation.isTypeAssignableTo(targetType, declaration.target.type, decoratorNode);
        if (!targetValid) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "decorator-wrong-target",
                messageId: "withExpected",
                format: {
                    decorator: declaration.name,
                    to: getTypeName(targetType),
                    expected: getEntityName(declaration.target.type),
                },
                target: decoratorNode,
            }));
        }
        return targetValid;
    }
    function checkDecoratorArguments(node, mapper, declaration) {
        // if we don't have a declaration we can just return the types or values if
        if (declaration === undefined) {
            return [
                false,
                node.arguments.map((argNode) => {
                    let type = checkNode(argNode, mapper) ?? errorType;
                    if (type.entityKind === "Indeterminate") {
                        type = type.type;
                    }
                    return {
                        value: type,
                        jsValue: type,
                        node: argNode,
                    };
                }),
            ];
        }
        let hasError = false;
        const minArgs = declaration.parameters.filter((x) => !x.optional && !x.rest).length ?? 0;
        const maxArgs = declaration.parameters[declaration.parameters.length - 1]?.rest
            ? undefined
            : declaration.parameters.length;
        if (node.arguments.length < minArgs ||
            (maxArgs !== undefined && node.arguments.length > maxArgs)) {
            // In the case we have too little args then this decorator is not applicable.
            // If there is too many args then we can still run the decorator as long as the args are valid.
            if (node.arguments.length < minArgs) {
                hasError = true;
            }
            if (maxArgs === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    messageId: "atLeast",
                    format: { actual: node.arguments.length.toString(), expected: minArgs.toString() },
                    target: node,
                }));
            }
            else {
                const expected = minArgs === maxArgs ? minArgs.toString() : `${minArgs}-${maxArgs}`;
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    format: { actual: node.arguments.length.toString(), expected },
                    target: node,
                }));
            }
        }
        const resolvedArgs = [];
        function resolveArg(argNode, perParamType) {
            const arg = getTypeOrValueForNode(argNode, mapper, {
                kind: "argument",
                constraint: perParamType,
            });
            if (arg !== null &&
                !(isType(arg) && isErrorType(arg)) &&
                checkArgumentAssignable(arg, perParamType, argNode)) {
                return {
                    value: arg,
                    node: argNode,
                    jsValue: resolveDecoratorArgJsValue(arg, extractValueOfConstraints({
                        kind: "argument",
                        constraint: perParamType,
                    })),
                };
            }
            else {
                return undefined;
            }
        }
        for (const [index, parameter] of declaration.parameters.entries()) {
            if (parameter.rest) {
                const restType = extractRestParamConstraint(parameter.type);
                if (restType) {
                    for (let i = index; i < node.arguments.length; i++) {
                        const argNode = node.arguments[i];
                        if (argNode) {
                            const arg = resolveArg(argNode, restType);
                            if (arg) {
                                resolvedArgs.push(arg);
                            }
                            else {
                                hasError = true;
                            }
                        }
                    }
                }
                break;
            }
            const argNode = node.arguments[index];
            if (argNode) {
                const arg = resolveArg(argNode, parameter.type);
                if (arg) {
                    resolvedArgs.push(arg);
                }
                else {
                    hasError = true;
                }
            }
        }
        return [hasError, resolvedArgs];
    }
    /** For a rest param of constraint T[] or valueof T[] return the T or valueof T */
    function extractRestParamConstraint(constraint) {
        let valueType;
        let type;
        if (constraint.valueType) {
            if (constraint.valueType.kind === "Model" &&
                isArrayModelType(program, constraint.valueType)) {
                valueType = constraint.valueType.indexer.value;
            }
            else {
                return undefined;
            }
        }
        if (constraint.type) {
            if (constraint.type.kind === "Model" && isArrayModelType(program, constraint.type)) {
                type = constraint.type.indexer.value;
            }
            else {
                return undefined;
            }
        }
        return {
            entityKind: "MixedParameterConstraint",
            type,
            valueType,
        };
    }
    function getIndexType(type) {
        return type.kind === "Model" ? type.indexer?.value : undefined;
    }
    function resolveDecoratorArgJsValue(value, valueConstraint) {
        if (valueConstraint !== undefined) {
            if (isValue(value)) {
                return marshallTypeForJS(value, valueConstraint.type);
            }
            else {
                return value;
            }
        }
        return value;
    }
    function checkArgumentAssignable(argumentType, parameterType, diagnosticTarget) {
        const [valid] = relation.isTypeAssignableTo(argumentType, parameterType, diagnosticTarget);
        if (!valid) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-argument",
                format: {
                    value: getEntityName(argumentType),
                    expected: getEntityName(parameterType),
                },
                target: diagnosticTarget,
            }));
        }
        return valid;
    }
    function checkAugmentDecorators(sym, targetType, mapper) {
        const augmentDecoratorNodes = resolver.getAugmentDecoratorsForSym(sym);
        const decorators = [];
        for (const decNode of augmentDecoratorNodes) {
            const decorator = checkDecoratorApplication(targetType, decNode, mapper);
            if (decorator) {
                decorators.unshift(decorator);
            }
        }
        return decorators;
    }
    /**
     * Check that augment decorator are targeting valid symbols.
     */
    function checkAugmentDecorator(node) {
        // This will validate the target type is pointing to a valid ref.
        resolveTypeReferenceSym(node.targetType, undefined);
        const links = resolver.getNodeLinks(node.targetType);
        if (links.isTemplateInstantiation) {
            program.reportDiagnostic(createDiagnostic({
                code: "augment-decorator-target",
                messageId: "noInstance",
                target: node.targetType,
            }));
        }
        else if (links.finalSymbol?.flags &&
            ~links.finalSymbol.flags & 1048576 /* SymbolFlags.Declaration */ &&
            ~links.finalSymbol.flags & 65536 /* SymbolFlags.Member */) {
            program.reportDiagnostic(createDiagnostic({
                code: "augment-decorator-target",
                messageId: links.finalSymbol.flags & 2 /* SymbolFlags.Model */
                    ? "noModelExpression"
                    : links.finalSymbol.flags & 64 /* SymbolFlags.Union */
                        ? "noUnionExpression"
                        : "default",
                target: node.targetType,
            }));
        }
        else if (links.finalSymbol?.flags && links.finalSymbol.flags & 128 /* SymbolFlags.Alias */) {
            const aliasNode = getSymNode(links.finalSymbol);
            program.reportDiagnostic(createDiagnostic({
                code: "augment-decorator-target",
                messageId: aliasNode.value.kind === SyntaxKind.UnionExpression ? "noUnionExpression" : "default",
                target: node.targetType,
            }));
        }
        // If this was used to get a type this is invalid, only used for validation.
        return errorType;
    }
    /**
     * Check that using statements are targeting valid symbols.
     */
    function checkUsings(node) {
        const usedSym = resolveTypeReferenceSym(node.name, undefined);
        if (usedSym) {
            if (~usedSym.flags & 256 /* SymbolFlags.Namespace */) {
                reportCheckerDiagnostic(createDiagnostic({ code: "using-invalid-ref", target: node.name }));
            }
        }
        // If this was used to get a type this is invalid, only used for validation.
        return errorType;
    }
    function checkDecorators(targetType, node, mapper) {
        const sym = isMemberNode(node)
            ? (getSymbolForMember(node) ?? node.symbol)
            : getMergedSymbol(node.symbol);
        const decorators = [];
        const augmentDecoratorNodes = resolver.getAugmentDecoratorsForSym(sym);
        const decoratorNodes = [
            ...augmentDecoratorNodes, // the first decorator will be executed at last, so augmented decorator should be placed at first.
            ...node.decorators,
        ];
        for (const decNode of decoratorNodes) {
            const decorator = checkDecoratorApplication(targetType, decNode, mapper);
            if (decorator) {
                decorators.unshift(decorator);
            }
        }
        // Doc comment should always be the first decorator in case an explicit @doc must override it.
        const docComment = extractMainDoc(targetType);
        if (docComment) {
            decorators.unshift(createDocFromCommentDecorator("self", docComment));
        }
        if (targetType.kind === "Operation") {
            const returnTypesDocs = extractReturnsDocs(targetType);
            if (returnTypesDocs.returns) {
                decorators.unshift(createDocFromCommentDecorator("returns", returnTypesDocs.returns));
            }
            if (returnTypesDocs.errors) {
                decorators.unshift(createDocFromCommentDecorator("errors", returnTypesDocs.errors));
            }
        }
        else if (targetType.kind === "ModelProperty") ;
        return decorators;
    }
    function checkScalar(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const decorators = [];
        const type = createType({
            kind: "Scalar",
            name: node.id.sv,
            node: node,
            constructors: new Map(),
            namespace: getParentNamespaceType(node),
            decorators,
            derivedScalars: [],
        });
        linkType(links, type, mapper);
        if (node.extends) {
            type.baseScalar = checkScalarExtends(node, node.extends, mapper);
            if (type.baseScalar) {
                copyDeprecation(type.baseScalar, type);
                type.baseScalar.derivedScalars.push(type);
            }
        }
        checkScalarConstructors(type, node, type.constructors, mapper);
        decorators.push(...checkDecorators(type, node, mapper));
        if (mapper === undefined) {
            type.namespace?.scalars.set(type.name, type);
        }
        linkMapper(type, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(type);
        }
        if (isInTypeSpecNamespace(type)) {
            stdTypes[type.name] = type;
        }
        return type;
    }
    function checkScalarExtends(scalar, extendsRef, mapper) {
        const symId = getNodeSym(scalar);
        pendingResolutions.start(symId, ResolutionKind.BaseType);
        const target = resolver.getNodeLinks(extendsRef).resolvedSymbol;
        if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-base-type",
                    format: { typeName: target.declarations[0].id.sv },
                    target: target,
                }));
            }
            return undefined;
        }
        const extendsType = getTypeForNode(extendsRef, mapper);
        pendingResolutions.finish(symId, ResolutionKind.BaseType);
        if (isErrorType(extendsType)) {
            compilerAssert(program.hasError(), "Should already have reported an error.", extendsRef);
            return undefined;
        }
        if (extendsType.kind !== "Scalar") {
            reportCheckerDiagnostic(createDiagnostic({ code: "extend-model", target: extendsRef }));
            return undefined;
        }
        return extendsType;
    }
    function checkScalarConstructors(parentScalar, node, constructors, mapper) {
        if (parentScalar.baseScalar) {
            for (const member of parentScalar.baseScalar.constructors.values()) {
                const newConstructor = cloneTypeForSymbol(getMemberSymbol(node.symbol, member.name), {
                    ...member,
                    scalar: parentScalar,
                });
                linkIndirectMember(node, newConstructor, mapper);
                constructors.set(member.name, newConstructor);
            }
        }
        for (const member of node.members) {
            const constructor = checkScalarConstructor(member, mapper, parentScalar);
            if (constructors.has(constructor.name)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "constructor-duplicate",
                    format: { name: constructor.name.toString() },
                    target: member,
                }));
                continue;
            }
            constructors.set(constructor.name, constructor);
        }
    }
    function checkScalarConstructor(node, mapper, parentScalar) {
        const name = node.id.sv;
        const links = getSymbolLinksForMember(node);
        if (links && links.declaredType && mapper === undefined) {
            // we're not instantiating this scalar constructor and we've already checked it
            return links.declaredType;
        }
        const member = createType({
            kind: "ScalarConstructor",
            scalar: parentScalar,
            name,
            node,
            parameters: node.parameters.map((x) => checkFunctionParameter(x, mapper, false)),
        });
        linkMapper(member, mapper);
        if (shouldCreateTypeForTemplate(node.parent, mapper)) {
            finishType(member);
        }
        if (links) {
            linkType(links, member, mapper);
        }
        return finishType(member);
    }
    function checkAlias(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const aliasSymId = getNodeSym(node);
        if (pendingResolutions.has(aliasSymId, ResolutionKind.Type)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-alias-type",
                    format: { typeName: node.id.sv },
                    target: node,
                }));
            }
            links.declaredType = errorType;
            return errorType;
        }
        pendingResolutions.start(aliasSymId, ResolutionKind.Type);
        const type = checkNode(node.value, mapper);
        if (type === null) {
            links.declaredType = errorType;
            return errorType;
        }
        if (isValue(type)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "value-in-type", target: node.value }));
            links.declaredType = errorType;
            return errorType;
        }
        linkType(links, type, mapper);
        pendingResolutions.finish(aliasSymId, ResolutionKind.Type);
        return type;
    }
    function checkConst(node) {
        const links = getSymbolLinks(node.symbol);
        if (links.value !== undefined) {
            return links.value;
        }
        const type = node.type ? getTypeForNode(node.type, undefined) : undefined;
        if (pendingResolutions.has(node.symbol, ResolutionKind.Value)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "circular-const",
                format: { name: node.id.sv },
                target: node,
            }));
            return null;
        }
        pendingResolutions.start(node.symbol, ResolutionKind.Value);
        const value = getValueForNode(node.value, undefined, type && { kind: "assignment", type });
        pendingResolutions.finish(node.symbol, ResolutionKind.Value);
        if (value === null || (type && !checkValueOfType(value, type, node.id))) {
            links.value = null;
            return links.value;
        }
        links.value = type ? copyValue(value, { type }) : copyValue(value);
        return links.value;
    }
    function inferScalarsFromConstraints(value, type) {
        switch (value.valueKind) {
            case "BooleanValue":
            case "StringValue":
            case "NumericValue":
                if (value.scalar === undefined) {
                    const scalar = inferScalarForPrimitiveValue(type, value.type);
                    return copyValue(value, { scalar });
                }
                return value;
            case "ArrayValue":
            case "ObjectValue":
            case "EnumValue":
            case "NullValue":
            case "ScalarValue":
                return value;
        }
    }
    function checkEnum(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (!links.type) {
            const enumType = (links.type = createType({
                kind: "Enum",
                name: node.id.sv,
                node,
                members: createRekeyableMap(),
                decorators: [],
            }));
            const memberNames = new Set();
            for (const member of node.members) {
                if (member.kind === SyntaxKind.EnumMember) {
                    const memberType = checkEnumMember(member, mapper, enumType);
                    if (memberNames.has(memberType.name)) {
                        reportCheckerDiagnostic(createDiagnostic({
                            code: "enum-member-duplicate",
                            format: { name: memberType.name },
                            target: node,
                        }));
                        continue;
                    }
                    memberNames.add(memberType.name);
                    enumType.members.set(memberType.name, memberType);
                }
                else {
                    const members = checkEnumSpreadMember(node.symbol, enumType, member.target, mapper, memberNames);
                    for (const memberType of members) {
                        linkIndirectMember(node, memberType, mapper);
                        enumType.members.set(memberType.name, memberType);
                    }
                }
            }
            const namespace = getParentNamespaceType(node);
            enumType.namespace = namespace;
            enumType.namespace?.enums.set(enumType.name, enumType);
            enumType.decorators = checkDecorators(enumType, node, mapper);
            linkMapper(enumType, mapper);
            finishType(enumType);
        }
        return links.type;
    }
    function checkInterface(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this interface and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const interfaceType = createType({
            kind: "Interface",
            decorators: [],
            node,
            namespace: getParentNamespaceType(node),
            sourceInterfaces: [],
            operations: createRekeyableMap(),
            name: node.id.sv,
        });
        linkType(links, interfaceType, mapper);
        interfaceType.decorators = checkDecorators(interfaceType, node, mapper);
        const ownMembers = checkInterfaceMembers(node, mapper, interfaceType);
        for (const extendsNode of node.extends) {
            const extendsType = getTypeForNode(extendsNode, mapper);
            if (extendsType.kind !== "Interface") {
                reportCheckerDiagnostic(createDiagnostic({ code: "extends-interface", target: extendsNode }));
                continue;
            }
            for (const member of extendsType.operations.values()) {
                if (interfaceType.operations.has(member.name)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "extends-interface-duplicate",
                        format: { name: member.name },
                        target: extendsNode,
                    }));
                }
                const newMember = cloneTypeForSymbol(getMemberSymbol(node.symbol, member.name), member, {
                    interface: interfaceType,
                });
                // Don't link it it is overritten
                if (!ownMembers.has(member.name)) {
                    linkIndirectMember(node, newMember, mapper);
                }
                // Clone deprecation information
                copyDeprecation(member, newMember);
                interfaceType.operations.set(newMember.name, newMember);
            }
            interfaceType.sourceInterfaces.push(extendsType);
        }
        for (const [key, value] of ownMembers) {
            interfaceType.operations.set(key, value);
        }
        linkMapper(interfaceType, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(interfaceType);
        }
        if (mapper === undefined) {
            interfaceType.namespace?.interfaces.set(interfaceType.name, interfaceType);
        }
        lateBindMemberContainer(interfaceType);
        lateBindMembers(interfaceType);
        return interfaceType;
    }
    function checkInterfaceMembers(node, mapper, interfaceType) {
        const ownMembers = new Map();
        for (const opNode of node.operations) {
            const opType = checkOperation(opNode, mapper, interfaceType);
            if (ownMembers.has(opType.name)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "interface-duplicate",
                    format: { name: opType.name },
                    target: opNode,
                }));
                continue;
            }
            ownMembers.set(opType.name, opType);
        }
        return ownMembers;
    }
    function checkUnion(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this union and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const variants = createRekeyableMap();
        const unionType = createType({
            kind: "Union",
            decorators: [],
            node,
            namespace: getParentNamespaceType(node),
            name: node.id.sv,
            variants,
            get options() {
                return Array.from(this.variants.values()).map((v) => v.type);
            },
            expression: false,
        });
        linkType(links, unionType, mapper);
        unionType.decorators = checkDecorators(unionType, node, mapper);
        checkUnionVariants(unionType, node, variants, mapper);
        linkMapper(unionType, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(unionType);
        }
        if (mapper === undefined) {
            unionType.namespace?.unions.set(unionType.name, unionType);
        }
        lateBindMemberContainer(unionType);
        lateBindMembers(unionType);
        return unionType;
    }
    function checkUnionVariants(parentUnion, node, variants, mapper) {
        for (const variantNode of node.options) {
            const variantType = checkUnionVariant(variantNode, mapper);
            variantType.union = parentUnion;
            if (variants.has(variantType.name)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "union-duplicate",
                    format: { name: variantType.name.toString() },
                    target: variantNode,
                }));
                continue;
            }
            variants.set(variantType.name, variantType);
        }
    }
    function checkUnionVariant(variantNode, mapper) {
        const links = getSymbolLinksForMember(variantNode);
        if (links && links.declaredType && mapper === undefined) {
            // we're not instantiating this union variant and we've already checked it
            return links.declaredType;
        }
        const name = variantNode.id ? variantNode.id.sv : Symbol("name");
        const type = getTypeForNode(variantNode.value, mapper);
        const variantType = createType({
            kind: "UnionVariant",
            name,
            node: variantNode,
            decorators: [],
            type,
            union: undefined,
        });
        variantType.decorators = checkDecorators(variantType, variantNode, mapper);
        linkMapper(variantType, mapper);
        if (shouldCreateTypeForTemplate(variantNode.parent, mapper)) {
            finishType(variantType);
        }
        if (links) {
            linkType(links, variantType, mapper);
        }
        return variantType;
    }
    function isMemberNode(node) {
        return node.symbol && !!(node.symbol.flags & 65536 /* SymbolFlags.Member */);
    }
    function getMemberSymbol(parentSym, name) {
        return parentSym ? resolver.getAugmentedSymbolTable(parentSym.members).get(name) : undefined;
    }
    function getSymbolForMember(node) {
        if (!node.id) {
            return undefined;
        }
        const name = node.id.sv;
        const parentSym = node.parent?.symbol;
        return parentSym ? getMemberSymbol(parentSym, name) : undefined;
    }
    function getSymbolLinksForMember(node) {
        const sym = getSymbolForMember(node);
        return sym ? (getSymNode(sym) === node ? getSymbolLinks(sym) : undefined) : undefined;
    }
    function checkEnumMember(node, mapper, parentEnum) {
        const name = node.id.sv;
        const links = getSymbolLinksForMember(node);
        if (links?.type) {
            return links.type;
        }
        compilerAssert(parentEnum, "Enum member should already have been checked.");
        const value = node.value ? node.value.value : undefined;
        const member = createType({
            kind: "EnumMember",
            enum: parentEnum,
            name,
            node,
            value,
            decorators: [],
        });
        if (links) {
            links.type = member;
        }
        member.decorators = checkDecorators(member, node, mapper);
        return finishType(member);
    }
    function checkEnumSpreadMember(parentEnumSym, parentEnum, targetNode, mapper, existingMemberNames) {
        const members = [];
        const targetType = getTypeForNode(targetNode, mapper);
        if (!isErrorType(targetType)) {
            if (targetType.kind !== "Enum") {
                reportCheckerDiagnostic(createDiagnostic({ code: "spread-enum", target: targetNode }));
                return members;
            }
            for (const member of targetType.members.values()) {
                if (existingMemberNames.has(member.name)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "enum-member-duplicate",
                        format: { name: member.name },
                        target: targetNode,
                    }));
                }
                else {
                    existingMemberNames.add(member.name);
                    const memberSym = getMemberSymbol(parentEnumSym, member.name);
                    const clonedMember = cloneTypeForSymbol(memberSym, member, {
                        enum: parentEnum,
                        sourceMember: member,
                    });
                    if (clonedMember) {
                        members.push(clonedMember);
                    }
                }
            }
        }
        return members;
    }
    function checkDirectives(node, type) {
        let hasDeprecation = false;
        for (const directive of node.directives ?? []) {
            if (directive.target.sv === "deprecated") {
                const message = directive.arguments[0];
                if (message === undefined) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-deprecation-argument",
                        messageId: "missing",
                        target: directive,
                    }));
                    continue;
                }
                let messageStr;
                if (message.kind !== SyntaxKind.StringLiteral) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-deprecation-argument",
                        format: { kind: SyntaxKind[message.kind] },
                        target: directive.arguments[0],
                    }));
                    messageStr = "<missing message>";
                }
                else {
                    messageStr = message.value;
                }
                if (hasDeprecation === true) {
                    reportCheckerDiagnostic(createDiagnostic({ code: "duplicate-deprecation", target: node }));
                }
                else {
                    hasDeprecation = true;
                    markDeprecated(program, type, {
                        message: messageStr,
                    });
                }
            }
        }
    }
    // the types here aren't ideal and could probably be refactored.
    function createAndFinishType(typeDef) {
        createType(typeDef);
        return finishType(typeDef);
    }
    /**
     * Given the own-properties of a type, returns a fully-initialized type.
     */
    function createType(typeDef) {
        Object.setPrototypeOf(typeDef, typePrototype);
        typeDef.isFinished = false;
        // If the type has an associated syntax node, check any directives that
        // might be attached.
        const createdType = typeDef;
        createdType.entityKind = "Type";
        if (createdType.node) {
            checkDirectives(createdType.node, createdType);
        }
        return createdType;
    }
    function finishType(typeDef) {
        return finishTypeForProgramAndChecker(program, typePrototype, typeDef);
    }
    function getLiteralType(node) {
        return createLiteralType(node.value, node);
    }
    function getMergedSymbol(sym) {
        // if (!sym) return sym;
        // return mergedSymbols.get(sym) || sym;
        return resolver.getMergedSymbol(sym);
    }
    function createGlobalNamespaceType() {
        const sym = resolver.symbols.global;
        const type = createType({
            kind: "Namespace",
            name: "",
            node: getGlobalNamespaceNode(),
            models: new Map(),
            scalars: new Map(),
            operations: new Map(),
            namespaces: new Map(),
            interfaces: new Map(),
            unions: new Map(),
            enums: new Map(),
            decoratorDeclarations: new Map(),
            functionDeclarations: new Map(),
            decorators: [],
        });
        getSymbolLinks(sym).type = type;
        type.decorators = checkAugmentDecorators(sym, type, undefined);
        return finishType(type);
    }
    function initializeClone(type, additionalProps) {
        let clone;
        switch (type.kind) {
            case "Model":
                const newModel = createType({
                    ...type,
                    decorators: [...type.decorators],
                    properties: undefined,
                    ...additionalProps,
                });
                if (!("properties" in additionalProps)) {
                    newModel.properties = createRekeyableMap(Array.from(type.properties.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { model: newModel }),
                    ]));
                }
                clone = newModel;
                break;
            case "Union":
                const newUnion = createType({
                    ...type,
                    decorators: [...type.decorators],
                    variants: undefined,
                    get options() {
                        return Array.from(this.variants.values()).map((v) => v.type);
                    },
                    ...additionalProps,
                });
                if (!("variants" in additionalProps)) {
                    newUnion.variants = createRekeyableMap(Array.from(type.variants.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { union: newUnion }),
                    ]));
                }
                clone = newUnion;
                break;
            case "Interface":
                const newInterface = createType({
                    ...type,
                    decorators: [...type.decorators],
                    operations: undefined,
                    ...additionalProps,
                });
                if (!("operations" in additionalProps)) {
                    newInterface.operations = createRekeyableMap(Array.from(type.operations.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { interface: newInterface }),
                    ]));
                }
                clone = newInterface;
                break;
            case "Enum":
                const newEnum = createType({
                    ...type,
                    decorators: [...type.decorators],
                    members: undefined,
                    ...additionalProps,
                });
                if (!("members" in additionalProps)) {
                    newEnum.members = createRekeyableMap(Array.from(type.members.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { enum: newEnum }),
                    ]));
                }
                clone = newEnum;
                break;
            default:
                clone = createType({
                    ...type,
                    ...("decorators" in type ? { decorators: [...type.decorators] } : {}),
                    ...additionalProps,
                });
                break;
        }
        return clone;
    }
    /**
     * Clone a type, resulting in an identical type with all the same decorators
     * applied. Decorators are re-run on the clone to achieve this.
     *
     * Care is taken to clone nested data structures that are part of the type.
     * Any type with e.g. a map or an array property must recreate the map or array
     * so that clones don't share the same object.
     *
     * For types which have sub-types that are part of it, e.g. enums with members,
     * unions with variants, or models with properties, the sub-types are cloned
     * as well.
     *
     * If the entire type graph needs to be cloned, then cloneType must be called
     * recursively by the caller.
     */
    function cloneType(type, additionalProps = {}) {
        let clone = initializeClone(type, additionalProps);
        if (type.isFinished) {
            clone = finishType(clone);
        }
        compilerAssert(clone.kind === type.kind, "cloneType must not change type kind");
        return clone;
    }
    /**
     * Clone a type linking to the given symbol.
     * @param sym Symbol which to associate the clone
     * @param type Type to clone
     * @param additionalProps Additional properties to set/override on the clone
     * @returns cloned type
     */
    function cloneTypeForSymbol(sym, type, additionalProps = {}) {
        let clone = initializeClone(type, additionalProps);
        if ("decorators" in clone) {
            const docComment = docFromCommentForSym.get(sym);
            if (docComment) {
                clone.decorators.push(createDocFromCommentDecorator("self", docComment));
            }
            for (const dec of checkAugmentDecorators(sym, clone, undefined)) {
                clone.decorators.push(dec);
            }
        }
        if (type.isFinished) {
            clone = finishType(clone);
        }
        compilerAssert(clone.kind === type.kind, "cloneType must not change type kind");
        return clone;
    }
    function createLiteralType(value, node) {
        if (program.literalTypes.has(value)) {
            return program.literalTypes.get(value);
        }
        let type;
        switch (typeof value) {
            case "string":
                type = createType({ kind: "String", value });
                break;
            case "boolean":
                type = createType({ kind: "Boolean", value });
                break;
            case "number":
                let valueAsString;
                if (node) {
                    compilerAssert(node.kind === SyntaxKind.NumericLiteral, "Must pass numeric literal node or undefined when creating a numeric literal type");
                    valueAsString = node.valueAsString;
                }
                else {
                    valueAsString = String(value);
                }
                type = createType({
                    kind: "Number",
                    value,
                    valueAsString,
                    numericValue: Numeric(valueAsString),
                });
                break;
        }
        program.literalTypes.set(value, type);
        return type;
    }
    /**
     * Check if the source type can be assigned to the target type and emit diagnostics
     * @param source Type of a value
     * @param constraint
     * @param diagnosticTarget Target for the diagnostic, unless something better can be inferred.
     */
    function checkTypeOfValueMatchConstraint(source, constraint, diagnosticTarget) {
        const [related, diagnostics] = relation.isTypeAssignableTo(source, constraint.type, diagnosticTarget);
        if (!related) {
            if (constraint.kind === "argument") {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument",
                    format: {
                        value: getEntityName(source),
                        expected: getEntityName(constraint.type),
                    },
                    target: diagnosticTarget,
                }));
            }
            else {
                reportCheckerDiagnostics(diagnostics);
            }
        }
        return related;
    }
    /**
     * Check if the source type can be assigned to the target type and emit diagnostics
     * @param source Source type
     * @param target Target type
     * @param diagnosticTarget Target for the diagnostic, unless something better can be inferred.
     */
    function checkTypeAssignable(source, target, diagnosticTarget) {
        const [related, diagnostics] = relation.isTypeAssignableTo(source, target, diagnosticTarget);
        if (!related) {
            reportCheckerDiagnostics(diagnostics);
        }
        return related;
    }
    function checkValueOfType(source, target, diagnosticTarget) {
        const [related, diagnostics] = relation.isValueOfType(source, target, diagnosticTarget);
        if (!related) {
            reportCheckerDiagnostics(diagnostics);
        }
        return related;
    }
    function isStdType(type, stdType) {
        if ((type.kind !== "Model" && type.kind !== "Scalar") ||
            type.namespace === undefined ||
            !isTypeSpecNamespace(type.namespace))
            return false;
        if (type.kind === "Scalar")
            return stdType === undefined || stdType === type.name;
        if (stdType === "Array" && type === stdTypes["Array"])
            return true;
        if (stdType === "Record" && type === stdTypes["Record"])
            return true;
        if (type.kind === "Model")
            return stdType === undefined || stdType === type.name;
        return false;
    }
    function getValueExactType(value) {
        return valueExactTypes.get(value);
    }
}
/**
 * Find all named models that could have been the source of the given
 * property. This includes the named parents of all property sources in a
 * chain.
 */
function getNamedSourceModels(property) {
    if (!property.sourceProperty) {
        return undefined;
    }
    const set = new Set();
    for (let p = property; p; p = p.sourceProperty) {
        if (p.model?.name) {
            set.add(p.model);
        }
    }
    return set;
}
/**
 * Find derived types of `models` in `possiblyDerivedModels` and add them to
 * `models`.
 */
function addDerivedModels(models, possiblyDerivedModels) {
    for (const element of possiblyDerivedModels) {
        if (!models.has(element)) {
            for (let t = element.baseModel; t; t = t.baseModel) {
                if (models.has(t)) {
                    models.add(element);
                    break;
                }
            }
        }
    }
}
function createTypeMapper(parameters, args, source, parentMapper) {
    const map = new Map(parentMapper?.map ?? []);
    for (const [index, param] of parameters.entries()) {
        map.set(param, args[index]);
    }
    return {
        partial: false,
        args: [...(parentMapper?.args ?? []), ...args],
        getMappedType: (type) => {
            return map.get(type) ?? type;
        },
        source,
        map,
    };
}
/**
 * If the input is anonymous (or the provided filter removes properties)
 * and there exists a named model with the same set of properties
 * (ignoring filtered properties), then return that named model.
 * Otherwise, return the input unchanged.
 *
 * This can be used by emitters to find a better name for a set of
 * properties after filtering. For example, given `{ @metadata prop:
 * string} & SomeName`, and an emitter that wishes to discard properties
 * marked with `@metadata`, the emitter can use this to recover that the
 * best name for the remaining properties is `SomeName`.
 *
 * @param model The input model
 * @param filter An optional filter to apply to the input model's
 * properties.
 */
function getEffectiveModelType(program, model, filter) {
    if (filter) {
        model = filterModelProperties(program, model, filter);
    }
    if (model.name) {
        // named model
        return model;
    }
    // We would need to change the algorithm if this doesn't hold. We
    // assume model has no inherited properties below.
    compilerAssert(!model.baseModel, "Anonymous model with base model.");
    if (model.properties.size === 0) {
        // empty model
        return model;
    }
    // Find the candidate set of named model types that could have been the
    // source of every property in the model.
    let candidates;
    for (const property of model.properties.values()) {
        const sources = getNamedSourceModels(property);
        if (!sources) {
            // unsourced property: no possible match
            return model;
        }
        if (!candidates) {
            // first sourced property: initialize candidates to its sources
            candidates = sources;
            continue;
        }
        // Add derived sources as we encounter them. If a model is sourced from
        // a base property, then it can also be sourced from a derived model.
        //
        // (Unless it is overridden, but then the presence of the overridden
        // property will still cause the the base model to be excluded from the
        // candidates.)
        //
        // Note: We depend on the order of that spread and intersect source
        // properties here, which is that we see properties sourced from derived
        // types before properties sourced from their base types.
        addDerivedModels(sources, candidates);
        // remove candidates that are not common to this property.
        for (const candidate of candidates) {
            if (!sources.has(candidate)) {
                candidates.delete(candidate);
            }
        }
    }
    // Search for a candidate that has no additional properties (ignoring
    // filtered properties). If so, it is effectively the same type as the
    // input model. Consider a candidate that meets this test without
    // ignoring filtering as a better match than one that requires filtering
    // to meet this test.
    let match;
    for (const candidate of candidates ?? []) {
        if (model.properties.size === countPropertiesInherited(candidate)) {
            match = candidate;
            break; // exact match
        }
        if (filter && !match && model.properties.size === countPropertiesInherited(candidate, filter)) {
            match = candidate;
            continue; // match with filter: keep searching for exact match
        }
    }
    return match ?? model;
}
/**
 * Applies a filter to the properties of a given type. If no properties
 * are filtered out, then return the input unchanged. Otherwise, return
 * a new anonymous model with only the filtered properties.
 *
 * @param model The input model to filter.
 * @param filter The filter to apply. Properties are kept when this returns true.
 */
function filterModelProperties(program, model, filter) {
    let filtered = false;
    for (const property of walkPropertiesInherited(model)) {
        if (!filter(property)) {
            filtered = true;
            break;
        }
    }
    if (!filtered) {
        return model;
    }
    const realm = Realm.realmForType.get(model);
    const typekit = realm ? $(realm) : $;
    const newModel = typekit.model.create({
        name: "",
        indexer: undefined,
        properties: {},
        decorators: [],
        derivedModels: [],
        sourceModels: [{ usage: "spread", model }],
    });
    for (const property of walkPropertiesInherited(model)) {
        if (filter(property)) {
            const newProperty = typekit.type.clone(property);
            Object.assign(newProperty, {
                sourceProperty: property,
                model: newModel,
            });
            newModel.properties.set(property.name, newProperty);
            typekit.type.finishType(newProperty);
        }
    }
    return finishTypeForProgram(program, newModel);
}
/**
 * Gets the property from the nearest base type that is overridden by the
 * given property, if any.
 */
function getOverriddenProperty(property) {
    compilerAssert(property.model, "Parent model must be set before overridden property can be found.");
    for (let current = property.model.baseModel; current; current = current.baseModel) {
        const overridden = current.properties.get(property.name);
        if (overridden) {
            return overridden;
        }
    }
    return undefined;
}
/**
 * Enumerates the properties declared by model or inherited from its base.
 *
 * Properties declared by more derived types are enumerated before properties
 * of less derived types.
 *
 * Properties that are overridden are not enumerated.
 */
function* walkPropertiesInherited(model) {
    const returned = new Set();
    for (let current = model; current; current = current.baseModel) {
        for (const property of current.properties.values()) {
            if (returned.has(property.name)) {
                // skip properties that have been overridden
                continue;
            }
            returned.add(property.name);
            yield property;
        }
    }
}
function countPropertiesInherited(model, filter) {
    let count = 0;
    for (const property of walkPropertiesInherited(model)) {
        if (!filter || filter(property)) {
            count++;
        }
    }
    return count;
}
function finishTypeForProgram(program, typeDef) {
    return finishTypeForProgramAndChecker(program, program.checker.typePrototype, typeDef);
}
function linkMapper(typeDef, mapper) {
    if (mapper) {
        compilerAssert(!typeDef.templateArguments, "Mapper provided but template arguments already set.");
        typeDef.templateMapper = mapper;
        typeDef.templateArguments = mapper.args;
    }
}
function extractMainDoc(type) {
    if (type.node?.docs === undefined) {
        return undefined;
    }
    let mainDoc = "";
    for (const doc of type.node.docs) {
        mainDoc += getDocContent(doc.content);
    }
    const trimmed = mainDoc.trim();
    return trimmed === "" ? undefined : trimmed;
}
function extractReturnsDocs(type) {
    const result = {
        returns: undefined,
        errors: undefined,
    };
    if (type.node?.docs === undefined) {
        return result;
    }
    for (const doc of type.node.docs) {
        for (const tag of doc.tags) {
            if (tag.kind === SyntaxKind.DocReturnsTag) {
                result.returns = getDocContent(tag.content);
            }
            if (tag.kind === SyntaxKind.DocErrorsTag) {
                result.errors = getDocContent(tag.content);
            }
        }
    }
    return result;
}
function extractParamDocs(node) {
    if (node.docs === undefined) {
        return new Map();
    }
    const paramDocs = new Map();
    for (const doc of node.docs) {
        for (const tag of doc.tags) {
            if (tag.kind === SyntaxKind.DocParamTag) {
                paramDocs.set(tag.paramName.sv, getDocContent(tag.content));
            }
        }
    }
    return paramDocs;
}
function extractPropDocs(node) {
    if (node.docs === undefined) {
        return new Map();
    }
    const propDocs = new Map();
    for (const doc of node.docs) {
        for (const tag of doc.tags) {
            if (tag.kind === SyntaxKind.DocPropTag) {
                propDocs.set(tag.propName.sv, getDocContent(tag.content));
            }
        }
    }
    return propDocs;
}
function getDocContent(content) {
    const docs = [];
    for (const node of content) {
        compilerAssert(node.kind === SyntaxKind.DocText, "No other doc content node kinds exist yet. Update this code appropriately when more are added.");
        docs.push(node.text);
    }
    return docs.join("");
}
function finishTypeForProgramAndChecker(program, typePrototype, typeDef) {
    if ("decorators" in typeDef) {
        for (const decApp of typeDef.decorators) {
            applyDecoratorToType(program, decApp, typeDef);
        }
    }
    Object.setPrototypeOf(typeDef, typePrototype);
    typeDef.isFinished = true;
    return typeDef;
}
function reportDeprecation(program, target, message, reportFunc) {
    if (program.compilerOptions.ignoreDeprecated !== true) {
        reportFunc(createDiagnostic({
            code: "deprecated",
            format: {
                message,
            },
            target,
        }));
    }
}
function applyDecoratorToType(program, decApp, target) {
    compilerAssert("decorators" in target, "Cannot apply decorator to non-decoratable type", target);
    for (const arg of decApp.args) {
        if (isType(arg.value) && isErrorType(arg.value)) {
            // If one of the decorator argument is an error don't run it.
            return;
        }
    }
    // Is the decorator definition deprecated?
    if (decApp.definition) {
        const deprecation = getDeprecationDetails(program, decApp.definition);
        if (deprecation !== undefined) {
            reportDeprecation(program, decApp.node ?? target, deprecation.message, program.reportDiagnostic);
        }
    }
    // peel `fn` off to avoid setting `this`.
    try {
        const args = decApp.args.map((x) => x.jsValue);
        const fn = decApp.decorator;
        const context = createDecoratorContext(program, decApp);
        fn(context, target, ...args);
    }
    catch (error) {
        // do not fail the language server for exceptions in decorators
        if (program.compilerOptions.designTimeBuild) {
            program.reportDiagnostic(createDiagnostic({
                code: "decorator-fail",
                format: { decoratorName: decApp.decorator.name, error: error.stack },
                target: decApp.node ?? target,
            }));
        }
        else {
            throw error;
        }
    }
}
function createDecoratorContext(program, decApp) {
    function createPassThruContext(program, decApp) {
        return {
            program,
            decoratorTarget: decApp.node,
            getArgumentTarget: () => decApp.node,
            call: (decorator, target, ...args) => {
                return decorator(createPassThruContext(program, decApp), target, ...args);
            },
        };
    }
    return {
        program,
        decoratorTarget: decApp.node,
        getArgumentTarget: (index) => {
            return decApp.args[index]?.node;
        },
        call: (decorator, target, ...args) => {
            return decorator(createPassThruContext(program, decApp), target, ...args);
        },
    };
}
function isTemplatedNode(node) {
    return "templateParameters" in node && node.templateParameters.length > 0;
}
var ResolutionKind;
(function (ResolutionKind) {
    ResolutionKind[ResolutionKind["Value"] = 0] = "Value";
    ResolutionKind[ResolutionKind["Type"] = 1] = "Type";
    ResolutionKind[ResolutionKind["BaseType"] = 2] = "BaseType";
    ResolutionKind[ResolutionKind["Constraint"] = 3] = "Constraint";
})(ResolutionKind || (ResolutionKind = {}));
class PendingResolutions {
    #data = new Map();
    start(symId, kind) {
        let existing = this.#data.get(symId);
        if (existing === undefined) {
            existing = new Set();
            this.#data.set(symId, existing);
        }
        existing.add(kind);
    }
    has(symId, kind) {
        return this.#data.get(symId)?.has(kind) ?? false;
    }
    finish(symId, kind) {
        const existing = this.#data.get(symId);
        if (existing === undefined) {
            return;
        }
        existing?.delete(kind);
        if (existing.size === 0) {
            this.#data.delete(symId);
        }
    }
}
const defaultSymbolResolutionOptions = {
    resolveDecorators: false,
    checkTemplateTypes: true,
};
function printTypeReferenceNode(node) {
    switch (node.kind) {
        case SyntaxKind.MemberExpression:
            return `${printTypeReferenceNode(node.base)}.${printTypeReferenceNode(node.id)}`;
        case SyntaxKind.TypeReference:
            return printTypeReferenceNode(node.target);
        case SyntaxKind.Identifier:
            return node.sv;
    }
}

export { $returnTypeVisibility as $, pageIndexDecorator as A, offsetDecorator as B, listDecorator as C, $parameterVisibility as D, EventEmitter as E, $withDefaultKeyVisibility as F, $withLifecycleUpdate as G, $withVisibilityFilter as H, $withVisibility as I, $defaultVisibility as J, $invisible as K, $removeVisibility as L, $visibility as M, $inspectTypeName as N, $inspectType as O, $opExample as P, $example as Q, $discriminator as R, discriminatedDecorator as S, $encodedName as T, $overload as U, $key as V, $friendlyName as W, $tag as X, $secret as Y, $maxValueExclusive as Z, $minValueExclusive as _, getProperty as a, getOverloads as a$, $maxValue as a0, $minValue as a1, $maxItems as a2, $minItems as a3, $maxLength as a4, $minLength as a5, $pattern as a6, $format as a7, $error as a8, $service as a9, getEffectiveModelType as aA, getErrorsDoc as aB, getErrorsDocData as aC, getExamples as aD, getFormat as aE, getFriendlyName as aF, getKey as aG, getLifecycleVisibilityEnum as aH, getMaxItems as aI, getMaxItemsAsNumeric as aJ, getMaxLength as aK, getMaxLengthAsNumeric as aL, getMaxValue as aM, getMaxValueAsNumeric as aN, getMaxValueExclusive as aO, getMaxValueExclusiveAsNumeric as aP, getMediaTypeHint as aQ, getMinItems as aR, getMinItemsAsNumeric as aS, getMinLength as aT, getMinLengthAsNumeric as aU, getMinValue as aV, getMinValueAsNumeric as aW, getMinValueExclusive as aX, getMinValueExclusiveAsNumeric as aY, getOpExamples as aZ, getOverloadedOperation as a_, $errorsDoc as aa, $returnsDoc as ab, $summary as ac, $withoutDefaultValues as ad, $withPickedProperties as ae, $withoutOmittedProperties as af, $withUpdateableProperties as ag, $withOptionalProperties as ah, $doc as ai, $mediaTypeHint as aj, $encode as ak, EmptyVisibilityProvider as al, Numeric as am, VisibilityFilter as an, addService as ao, addVisibilityModifiers as ap, clearVisibilityModifiersForClass as aq, explainStringTemplateNotSerializable as ar, filterModelProperties as as, getAllTags as at, getDeprecated as au, getDiscriminatedTypes as av, getDiscriminatedUnion as aw, getDiscriminatedUnionFromInheritance as ax, getDiscriminator as ay, getDoc as az, getLocationContext as b, getPagingOperation as b0, getParameterVisibilityFilter as b1, getPattern as b2, getPatternData as b3, getPropertyType as b4, getReturnTypeVisibilityFilter as b5, getReturnsDoc as b6, getReturnsDocData as b7, getService as b8, getSummary as b9, validateDecoratorUniqueOnNode as bA, walkPropertiesInherited as bB, createTypekit as bC, $ as bD, TypekitPrototype as bE, TypekitNamespaceSymbol as bF, defineKit as bG, UnionKit as bH, MutatorFlow as bI, mutateSubgraph as bJ, mutateSubgraphWithNamespace as bK, Realm as bL, getTags as ba, getVisibilityForClass as bb, hasVisibility as bc, isDeprecated as bd, isErrorModel as be, isKey as bf, isList as bg, isNumeric as bh, isNumericType as bi, isSealed as bj, isSecret as bk, isService as bl, isStringType as bm, isTypeSpecValueTypeOf as bn, isVisible as bo, listServices as bp, markDeprecated as bq, navigateType as br, navigateTypesInNamespace as bs, removeVisibilityModifiers as bt, resetVisibilityModifiersForClass as bu, scopeNavigationToNamespace as bv, sealVisibilityModifiers as bw, sealVisibilityModifiersForProgram as bx, typespecTypeToJson as by, validateDecoratorNotOnType as bz, createSymbol as c, getSymNode as d, createSymbolTable as e, createChecker as f, getEncode as g, createBinder as h, validatePagingOperations as i, getDocData as j, getDeprecationDetails as k, getterDecorator as l, mapEventEmitterToNodeListener as m, navigateProgram as n, docFromCommentDecorator as o, indexerDecorator as p, lastLinkDecorator as q, resolveEncodedName as r, setCurrentProgram as s, firstLinkDecorator as t, prevLinkDecorator as u, validateEncodedNamesConflicts as v, nextLinkDecorator as w, continuationTokenDecorator as x, pageItemsDecorator as y, pageSizeDecorator as z };
